import datetime
import os
import shelve
import shlex
import warnings
import bsddb

import numpy as np
from scipy import stats
import vg.time_series_analysis.seasonal_kde as skde
from vg.time_series_analysis import distributions
from vg import times
from vg import helpers as my
from vg.meteo import meteox2y, avrwind

try:
    import config as conf
except ImportError:
    from vg import config_template as conf
    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]
    warnings.warn('Could not import "config.py".\n' +
                  'Edit "%s" and rename it to "config.py"' % conf_filepath)


def detrend(values):
    """Detrend by subtracting the linear regression function."""
    dummy_time = np.arange(len(values))
    b, a = stats.linregress(dummy_time, values)[:2]
    return values + (values.mean() - (a + b * dummy_time))


def daily_wind(met, sum_interval=24):
    """Aggregates daily wind direction and speed from hourly values."""
    u_daily = my.sumup(met["u"], sum_interval, sum_to_nan=True)
    v_daily = my.sumup(met["v"], sum_interval, sum_to_nan=True)
    return avrwind.component2angle(u_daily, v_daily)


def met_as_array(met, p=0, T=None, var_names=None):
    """Returns the met - dictionary as an array. Rows are sorted
    alpha-numerically according to the variable names."""
    var_names = sorted(met.keys()) if var_names is None else var_names
    K = len(var_names)
    if T is None:
        T = len(met[var_names[0]]) - p
    # this lets some outer code access the following variables without
    # returning them explicitly
    met_as_array.var_names, met_as_array.T = var_names, T
    # the "T + p" is there to include the presample timesteps.
    data = np.nan * np.empty((K, T + p))
    for var_i, key in enumerate(var_names):
        data[var_i] = met[key]
    return data


def met2array(filepath, met_kwds=None, as_array_kwds=None, sumup_kwds=None):
    dict_default = lambda x: {} if x is None else x
    met_kwds, as_array_kwds, sumup_kwds = \
        map(dict_default, (met_kwds, as_array_kwds, sumup_kwds))
    dt_hourly, met = read_met(filepath, **met_kwds)
    data_hourly = met_as_array(met, **as_array_kwds)
    data, dt = my.sumup(data_hourly, times_=dt_hourly, **sumup_kwds)
    try:
        n_sumup = sumup_kwds["width"]
    except KeyError:
        n_sumup = 24
    return data / n_sumup, dt, met_as_array.var_names


def dyresm2array(filepath, *args, **kwds):
    met_kwds = kwds.pop("met_kwds") if "met_kwds" in kwds else {}
    met_kwds["startfrom"] = 5
    return met2array(filepath, met_kwds=met_kwds, *args, **kwds)


def _parse_time(var_dict):
    """Get the time information in var_dict.
    This code is adjusted every time someone throws new data at me."""
    datetimes = None
    if "Date" in var_dict and "hour" in var_dict:
        datetimes = \
            times.str2datetime(np.array(["%s %s" % (date_str, hour_str)
                                         for date_str, hour_str
                                         in zip(var_dict["Date"],
                                                var_dict["hour"])]),
                               "%d. %m %y %H")
        del var_dict["Date"]
        del var_dict["hour"]
    if "Julian" in var_dict:
        datetimes = times.cwr2datetime(var_dict["Julian"])
        del var_dict["Julian"]
    elif "YrDayNum" in var_dict:
        datetimes = times.cwr2datetime(var_dict["YrDayNum"])
        del var_dict["YrDayNum"]
    elif "jd" in var_dict:
        datetimes = times.cwr2datetime(var_dict["jd"])
        del var_dict["jd"]
    if "chron" in var_dict:
        datetimes = times.str2datetime(var_dict["chron"],
                                       "(%m/%d/%y %H:%M:%S)")
        del var_dict["chron"]
    elif "date" in var_dict:
        datetimes = times.str2datetime(var_dict["date"])
        del var_dict["date"]
    elif "time" in var_dict:
        datetimes = times.iso2datetime(var_dict["time"])
        del var_dict["time"]
    elif "times" in var_dict:
        datetimes = times.str2datetime(var_dict["times"], "%Y-%m-%d %H:%M:%S")
        del var_dict["times"]
    if "year" in var_dict and "doy" in var_dict:
        datetimes = times.str2datetime(var_dict["year"], "%Y")
        datetimes += np.array([datetime.timedelta(int(doy))
                               for doy in var_dict["doy"]])
        del var_dict["year"], var_dict["doy"]
    if "Date" in var_dict and "Time" in var_dict:
        # as seen in Excel
        datetimes = times.xls2datetime([int(day) for day in var_dict["Date"]])
        datetimes += \
            np.array([datetime.timedelta(hours=int(hour.split(":")[0]))
                      for hour in var_dict["Time"]])
        del var_dict["Date"], var_dict["Time"]
    if "ta" in var_dict and "mo" in var_dict and "jahr" in var_dict:
        datetimes = \
            np.array([datetime.datetime(int(year), int(month), int(day))
                      for year, month, day in
                      zip(var_dict["jahr"], var_dict["mo"], var_dict["ta"])])
        del var_dict["ta"]
        del var_dict["mo"]
        del var_dict["jahr"]
    if "index" in var_dict:
        del var_dict["index"]
    if datetimes is None:
        raise TypeError("Could not parse dates.")
    return var_dict, datetimes


# @my.pickle_cache(os.path.join(conf.cache_dir, "met_%s.pkl"))  # , warn=False)
def read_met(filepath=None, minimum_water_temp=.1, delimiter="\t",
             verbose=True, main_diff=None, **kwds):
    if filepath is None:
        filepath = conf.met_file
    var_dict = my.csv2dict(filepath, delimiter=delimiter, **kwds)
    # get rid of quotes around filenames
    var_dict = {shlex.split(var_name)[0]: values
                for var_name, values in var_dict.items()}

    var_dict, datetimes = _parse_time(var_dict)

    met = {key: np.array([val if val != "" else np.nan for val in values],
                         dtype=float)
           for key, values in var_dict.items()}

    met = {key: times.regularize(val, datetimes, nan=True,
                                 main_diff=main_diff)[0]
           for key, val in met.items()}
    # get also the regularized datetimes
    datetimes = times.regularize(np.empty(len(datetimes)), datetimes,
                                 main_diff=main_diff)[1]

    if verbose:
        n_nan = np.sum([np.isnan(val) for val in met.values()])
        print "\tFound %d missing or nan values" % n_nan
        for var_name, val in sorted(met.items()):
            gaps = my.gaps(val)
            if len(gaps) > 0:
                n_gaps = np.sum(np.isnan(val))
                gaps_perc = float(n_gaps) / len(val) * 100
                print "\t%s (%d, %.2f%%)" % (var_name, n_gaps, gaps_perc)
                for start, end in gaps:
                    if start == end:
                        print "\t\t", datetimes[start]
                    else:
                        print ("\t\t%s - %s (%d)" %
                               (datetimes[start], datetimes[end],
                                end - start + 1))

    if "Cloud_Cover" not in var_dict:
        try:
            var_dict["Cloud_Cover"] = \
                meteox2y.lw2clouds(np.array(var_dict["ILWR"], dtype=float),
                                   np.array(var_dict["theta"], dtype=float),
                                   e=np.array(var_dict["e"], dtype=float))
        except KeyError:
            pass

    if "wdir" in met and "U" in met:
        u, v = avrwind.angle2component(met["wdir"], met["U"])
        met.update({"u": u, "v": v})

    if ("rh" not in met) and ("e" in met) and ("theta" in met):
        met["rh"] = meteox2y.vap_p2rel(met["e"], met["theta"])
        too_dry_ii = met["rh"] < 0
        n_too_dry = np.sum(too_dry_ii)
        if n_too_dry > 0:
            if verbose:
                warnings.warn(("Input of e and theta caused %d rel. " +
                               "humidities to be < 0. Capping them at 0.")
                              % n_too_dry)
            met["rh"][too_dry_ii] = np.nan
        too_moist_ii = met["rh"] > 1.01
        n_too_moist = np.sum(too_moist_ii)
        if n_too_moist > 0:
            if verbose:
                warnings.warn(("Input of e and theta caused %d rel. " +
                               "humidities to be > 1. Capping them at 1.")
                              % n_too_moist)
            met["rh"][too_moist_ii] = 1.01  # np.nan
    if "rh" in met and np.nanmax(met["rh"]) > 50:
        met["rh"] /= 100
    for var_name in met.keys():
        if var_name.startswith("wtemp"):
            too_cold_ii = met[var_name] < 0
            n_too_cold = np.sum(too_cold_ii)
            if n_too_cold > 0:
                if verbose:
                    warnings.warn(("%d temperatures < 0 in %s. Capping them " +
                                   "at %.3f")
                                  % (n_too_cold, var_name, minimum_water_temp))
                met[var_name][too_cold_ii] = minimum_water_temp

    return datetimes, met


class VGBase(object):

    """Handle everything except plotting."""

    def __init__(self, var_names, met_file=None, sum_interval=24, plot=False,
                 separator="\t", refit=None, detrend_vars=None, verbose=False,
                 data_dir=None, cache_dir=None, dump_data=True, non_rain=None,
                 **met_kwds):
        # external_var_names=None,
        # external_var_names : sequence of str, optional
        #     Must be present in the met-file.
        self.var_names = var_names
        self.verbose = verbose
        self.met_file = conf.met_file if met_file is None else met_file
        self.data_dir = conf.data_dir if data_dir is None else data_dir
        if cache_dir is not None:
            self.cache_dir = cache_dir
            self.seasonal_cache_file = os.path.join(cache_dir,
                                                    "seasonal_solutions.she")
        else:
            self.cache_dir = conf.cache_dir
            self.seasonal_cache_file = conf.seasonal_cache_file
        self.dump = dump_data
        # will be used in _load_and_prepare_data
        self.met = self.times_orig = None
        # are we given a sequence? then store sum_interval in a way, that
        # data can be divided by it
        try:
            self.sum_interval = np.array(sum_interval)[:, np.newaxis]
        except IndexError:
            # let's make this the same shape, so life gets easier later on
            self.sum_interval = \
                np.array(self.K * [sum_interval])[:, np.newaxis]
        self.plot = plot

        # attributes that will be set with meaningfull values in fit()
        # AM is also used to store the parameters of a pure VAR-process
        self.AM = self.sigma_u = self.residuals = self.means = None
        self.p = self.q = None

        # these attributes will be populated in simulate() and disaggregate()
        self.T = self.mean_arrival = self.disturbance_std = None
        self.theta_incr = self.theta_grad = self.fixed_variables = None
        self.primary_var = self.climate_signal = self.start_date = None
        self.m_t = self.sim = self.sim_sea = self.sim_times = None
        self.outfilepath = self.sim_sea_dis = self.dis_times = None
        self.var_names_dis = self.ex_in = self.ex_out = None

        # cached attributes
        self._data_doys = self._data_doys_raw = None
        self._sim_doys = self._dis_doys = None
        self._sim_doys_len = self._dis_doys_len = np.nan
        self._cov_trans = None

        print "Loading input data."
        self.times, self.data_raw = self._load_and_prepare_data(separator,
                                                                **met_kwds)

        if detrend_vars:
            for var_name in detrend_vars:
                var_ii = var_names.index(var_name)
                self.data_raw[var_ii] = detrend(self.data_raw[var_ii])

        if self.seasonal_fitting:
            print "Fitting seasonal distributions."
            self.data_trans, self.dist_sol = self._fit_seasonal(refit)

        if "R" in self.var_names:
            print "Transforming 'negative rain'"
            rain_i = self.var_names.index("R")
            rain = self.data_raw[rain_i]
            self.data_trans = self._negative_rain(self.data_trans,
                                                  rain,
                                                  self.data_doys,
                                                  non_rain)

    @property
    def sum_interval_dict(self):
        """Maps var_name to sum_interval."""
        return dict((var_name,
                     self.sum_interval[self.var_names.index(var_name)])
                    for var_name in self.var_names)

    @property
    def primary_var_ii(self):
        """The row index of the primary variable."""
        try:
            return [self.var_names.index(prim_var)
                    for prim_var in self.primary_var]
        except ValueError:
            warnings.warn("No %s in input." % self.primary_var)
            return None

    @property
    def output_resolution(self):
        """In hours."""
        delta = self.times[1] - self.times[0]
        # there is timedelta.total_seconds in python 2.7, but lets be kind to
        # the conservative people -- oh, thank you ;-)
        return (24 * delta.days +
                (delta.seconds + 1e-3 * delta.microseconds) / 60 ** 2)

    @property
    def K(self):
        """The number of variables."""
        return len(self.var_names)

    @property
    def T_data(self):
        """Length of the input data time series."""
        return len(self.times_orig)

    @property
    def T_summed(self):
        """Length of aggregated input data."""
        return len(self.times)

    @property
    def data_doys(self):
        """Days of year of aggregated input data."""
        if self._data_doys is None:
            self._data_doys = times.datetime2doy(self.times)
        return self._data_doys

    @property
    def data_doys_raw(self):
        """Days of year of raw (hourly) input data."""
        if self._data_doys_raw is None:
            self._data_doys_raw = times.datetime2doy(self.times_orig)
        return self._data_doys_raw

    @property
    def sim_doys(self):
        """Days of year of aggregated output data."""
        if self._sim_doys is None or len(self.sim_times) != self._sim_doys_len:
            self._sim_doys = times.datetime2doy(self.sim_times)
            self._sim_doys_len = len(self._sim_doys)
        return self._sim_doys

    @property
    def dis_doys(self):
        """Days of year of disaggregated output data."""
        if self._dis_doys is None or len(self._dis_doys) != self._dis_doys_len:
            self._dis_doys = times.datetime2doy(self.dis_times)
            self._dis_doys_len = len(self._dis_doys)
        return self._dis_doys

    def fitted_medians(self, var_name, doys=None):
        """Medians of the fitted seasonal distribution.

        Usefull to construct `climate_signal`s

        Parameter
        ---------
        var_name : str or sequence of strings
        doys : 1d array
            If None, `sim_times` is used. If that is not available,
            `data_doys`.

        """
        if isinstance(var_name, basestring):
            var_names = var_name,
        else:
            var_names = var_name
        medians = []
        for var_name in var_names:
            dist, solution = self.dist_sol[var_name]
            var_i = self.var_names.index(var_name)
            if doys is None:
                if self.sim_times is None:
                    doys = self.data_doys
                else:
                    doys = self.sim_doys
            medians += [(dist.ppf(solution, np.full_like(doys, .5), doys) /
                         self.sum_interval[var_i])]
        return np.squeeze(medians)

    def _shuffle(self, nn, m, tpd=24, autocorr_len=48, doys_in=None,
                 doys_out=None, doy_tolerance=15, nan_mask=None):
        """Draw a clustered sample of size m with elements [0,nn]."""

        # like a normal modulus, but knows how to divide by 0!!1
        mod0 = lambda x, y: x % y if y != 0 else 0
        pool0 = np.where(times.doy_distance(doys_out[0], doys_in) <=
                         doy_tolerance)[0]
        pool_len = len(pool0)

        def choose_chunk(dst_point):
            if doys_in is not None and doys_out is not None:
                # pool = (pool0 + dst_point) % nn
                pool = pool0 + dst_point
                if dst_point > 0:
                    end_pool = pool[-dst_point:]
                    too_big_mask = end_pool >= nn
                    end_pool[too_big_mask] = end_pool[too_big_mask] - nn
                    too_big_mask = end_pool >= nn
                    end_pool[too_big_mask] = end_pool[too_big_mask] - nn
                src_point = pool[np.random.randint(pool_len)]
                while src_point >= (nn - autocorr_len):
                    src_point = pool[np.random.randint(pool_len)]
            else:
                src_point = np.random.randint(nn - autocorr_len)
            # hour of day from 0 to 23 in destination
            hour_of_dst = mod0(dst_point, tpd)
            # insure the same hour of day in source
            src_point += -mod0(src_point, tpd) + hour_of_dst
            src_point = min(src_point, m - autocorr_len)
            chunk_ii = range(src_point, src_point + autocorr_len)
            if np.any(nan_mask[chunk_ii]):
                chunk_ii = choose_chunk(dst_point)
            return chunk_ii

        indices = np.array([choose_chunk(dst_point)
                            for dst_point
                            in xrange(0, m + autocorr_len, autocorr_len)])
        return indices.ravel()[:m]

    def _load_and_prepare_data(self, delimiter="\t", **met_kwds):
        """Loading the data from the met_file and aggregate it according to
        sum_interval (both defined in __init__). Plus plotting if requested."""

        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

        try:
            # get the data
            # this covers the case that a non-default met filename was given
            # that exists in the data_dir but not in the working path
            if ((self.met_file is not None) and
                    (not os.path.exists(self.met_file))):
                self.met_file = os.path.join(self.data_dir, self.met_file)
            self.times_orig, self.met = \
                read_met(self.met_file, delimiter=delimiter,
                         verbose=self.verbose, **met_kwds)
        except TypeError as exc:
            warnings.warn("While reading met-file:\n%s" % exc.message)
            try:
                # could be a pandas DataFrame
                self.times_orig = self.met_file.index.to_pydatetime()
                self.met = self.met_file.to_dict("list")
            except AttributeError:
                # or just a dict
                self.times_orig = self.met_file["datetimes"]
                self.met = {key: val for key, val in self.met_file.items()
                            if key != "datetime"}

        # convert the dictionary into an array.
        # mind the order: alpha-numeric according to var_names
        # will be shorter once summed up
        T_raw_data = len(self.met[self.var_names[0]])
        # self.K corresponds to all variables. the wtemp variables are
        # unfortunately stored someplace else and should be ignored here
        # so for the shape of data, we do not count to K, but something lower
        # if there are water temperatures in var_names
        var_names_part = [var_name for var_name in self.var_names
                          if (not var_name.startswith("wtemp") and
                              var_name != "nao")]
        data = np.empty((len(var_names_part),
                         T_raw_data / self.sum_interval[0, 0]))
        data[:] = np.nan

        for key in var_names_part:
            var_i = var_names_part.index(key)
            if self.sum_interval[var_i] > 1:
                data[var_i], times_ = my.sumup(self.met[key],
                                               self.sum_interval[var_i],
                                               self.times_orig,
                                               middle_time=False,
                                               sum_to_nan=False)
            else:
                data[var_i] = self.met[key]
                times_ = self.times_orig

        if "U" in self.var_names and "U" not in self.met:
            # the wind has to be aggregated taking the direction into account
            data[self.var_names.index("U")] = \
                daily_wind(self.met, self.sum_interval_dict["U"])[1]

#        if "nao" in self.var_names:
#            import nao
#            nao_dates, naoi = nao.read_nao()
#            self.met["nao"] = naoi[np.where(nao_dates == times_[0])[0]:
#                              np.where(nao_dates == times_[-1])[0] + 1]
#            var_row = self.var_names.index("nao")
#            summed_nao = my.sumup(self.met["nao"], self.sum_interval[var_row])
#            if summed_nao.shape[1] < data.shape[1]:
#                shape_diff = data.shape[1] - summed_nao.shape[1]
#                summed_nao = np.concatenate((np.zeros((1, shape_diff)),
#                                             summed_nao), axis=1)
#            data = np.insert(data, var_row, summed_nao, axis=0)

        # check if we are supposed to simulate water temperatures - they are
        # stored someplace else
#        if "wtemp" in "".join(self.var_names):
#            met_rivers = read_met(conf.hyd_file)[1]
#            self.met.update(met_rivers)
#            for var_name in met_rivers.keys():
#                if var_name in self.var_names:
#                    var_row = self.var_names.index(var_name)
#                    new_data = my.sumup(self.met[var_name],
#                                        self.sum_interval[var_row])
# the last measurement of data was dropped in sumup! so
# loose it also here
#                    new_data = new_data[:, :data.shape[1]]
#                    data = np.insert(data, var_row, new_data, axis=0)
        return times_, data

    def _prepare_fixed_data(self):
        """Convert the data to standard-normal an put it into (K,T)-array form.
        """
        if self.fixed_variables:
            sh = shelve.open(self.seasonal_cache_file, "c")
            fixed_data = np.nan * np.empty((self.K, self.T))
            for var_name, values in self.fixed_variables.items():
                var_ii = self.var_names.index(var_name)
                if values is None:
                    # fix the input data the model was fitted on
                    values = self.data_raw[var_ii]
                solution_key = var_name

                # the fitting was originially done for daily sums of hourly
                # values, but we can have different aggregation lengths
                sum_interval = self.sum_interval_dict[var_name]
                if sum_interval != 24:
                    solution_key += "_%d" % sum_interval
                seas_class, dist_class, solution = sh[solution_key]
                dist = seas_class(dist_class, values, self.times,
                                  fixed_pars=conf.par_known[var_name])
                quantiles = np.squeeze(dist.cdf(solution))
                transformed = distributions.norm.ppf(quantiles)
                fixed_data[var_ii] = transformed
            sh.close()
            return fixed_data
        else:
            return None

    def _gen_sim_times(self, T=None, start_str=None, stop_str=None,
                       output_resolution=None):
        """Generates an array of datetimes starting at start_date with T values
        and resolution hours in between.

        Parameters
        ----------
        T : int
            number of timesteps, optional
        resolution : int
            length of timestep in hours
        start_date : datetime object, optional
            Representation of first timestep
        output_resolution : int or float, optional
            Output resolution in hours

        Returns
        -------
        times_out : (T,) ndarray, dtype=object (datetime)

        Examples
        --------
        >>> import datetime
        >>> sim_times(4, 48, datetime.datetime(year=2000, month=1, day=1))
        array([2000-01-01 00:00:00, 2000-01-03 00:00:00, 2000-01-05 00:00:00,
              2000-01-07 00:00:00], dtype=object)
        """
        if T is not None and output_resolution is None:
            self.T = T
        elif T is None:
            T = self.T
        if output_resolution is None:
            output_resolution = self.output_resolution
        # produce sim_doys and times_out:
        if start_str is None:
            if self.sim_times is None:
                self.start_date = self.times_orig[0]
            else:
                self.start_date = self.sim_times[0]
        else:
            self.start_date = times.str2datetime(start_str)
        if stop_str is not None:
            # overwrite T setting
            end_date = times.str2datetime(stop_str)
            t_diff_seconds = (end_date - self.start_date).total_seconds()
            T = int(t_diff_seconds / (60 ** 2 * 24))
#         interval_secs = 60. ** 2 * output_resolution
#
#         times_out = np.cumsum([0] + (T - 1) * [interval_secs])
#         times_out += times.datetime2unix(self.start_date)
#         times_out = times.unix2datetime(times_out)
        resolution_timedelta = \
            datetime.timedelta(hours=float(output_resolution))
        times_out = np.array([self.start_date + t * resolution_timedelta
                              for t in xrange(T)])
        return times_out

    def _fit_seasonal(self, refit=None, values=None, doys=None):
        if refit is None:
            refit = tuple()
        elif refit == "all" or refit is True:
            refit = self.var_names
        sh = shelve.open(self.seasonal_cache_file, 'c')
        try:
            keys = sh.keys()
        except bsddb.db.DBPageNotFoundError:
            print "Cache file corrupted, refitting..."
            os.remove(self.seasonal_cache_file)
            sh = shelve.open(self.seasonal_cache_file, 'c')
            keys = []
        if values is None:
            values = self.data_raw
        if doys is None:
            doys = self.data_doys
        data_trans = np.empty_like(values)
        dist_sol = {}
        for var_ii, var in enumerate(values):
            var_name = self.var_names[var_ii]
            solution_key = var_name

            # the fitting was originially done for daily sums of hourly values
            # we can have different aggregation lengths so we trigger a fitting
            # if we come across one of these
            sum_interval = self.sum_interval_dict[var_name]
            plain_solution_key = solution_key
            if sum_interval != 24:
                solution_key += "_%d" % sum_interval

            kwds = (conf.dists_kwds[var_name]
                    if var_name in conf.dists_kwds
                    else {})

            if solution_key not in keys or plain_solution_key in refit:
                if self.verbose:
                    print "\tFitting a distribution to ", var_name
                seas_class = conf.seasonal_classes[var_name]
                if seas_class is skde.SeasonalKDE:
                    dist = seas_class(var, self.times,
                                      fixed_pars=conf.par_known[var_name],
                                      verbose=self.verbose, **kwds)
                    solution = dist.fit()
                    sh[solution_key] = [seas_class, None, solution]
                else:
                    dist = seas_class(conf.dists[var_name], var, self.times,
                                      fixed_pars=conf.par_known[var_name],
                                      verbose=self.verbose, **kwds)
                    solution = dist.fit()
                    try:
                        sh[solution_key] = \
                            [seas_class, conf.dists[var_name], solution]
                    except TypeError:
                        # we will rely on the distribution set in the config
                        # file later
                        sh[solution_key] = [seas_class, None, solution]

            else:
                if self.verbose:
                    print "\tRecover previous fit from shelve for: ", var_name
                seas_class, dist_class, solution = sh[solution_key]
                if len(self.times) > len(var):
                    # we assume we were called from predict in which case
                    # we have to pass the past data in here and not the
                    # passed data, hehe
                    var_ = self.data_raw[var_ii]
                else:
                    var_ = var
                if seas_class is skde.SeasonalKDE:
                    dist = seas_class(var_, self.times, solution,
                                      fixed_pars=conf.par_known[var_name],
                                      **kwds)
                else:
                    if dist_class is None:
                        dist_class = conf.dists[var_name]
                    dist = seas_class(dist_class, var_, self.times,
                                      solution=solution,
                                      fixed_pars=conf.par_known[var_name],
                                      **kwds)

            if self.verbose:
                print ("\tp-value of chi2 goodness-of-fit %.4f" %
                       dist.chi2_test())
            quantiles = dist.cdf(solution, x=var, doys=doys)
            data_trans[var_ii] = distributions.norm.ppf(quantiles)
            dist_sol[var_name] = dist, solution
        sh.close()
        # we have outrageous outliers from time to time
        data_trans[np.abs(data_trans) >= 1e300] = np.nan
        data_trans = my.interp_nan(data_trans)
        return data_trans, dist_sol

    def dist_sol_hourly(self, var_name):
        return self.dist_sol["%s_hourly" % var_name]

    def _negative_rain(self, data_trans, rain, doys, var_names=None):
        """
        Transform rain-gaps to standard-normal by distance to wet conditions.

        Parameter
        ---------
        data_trans : 2d array
            Transformed variables.
        rain : 1d array
            Untransformed rain.
        doys : 1d array
        var_names : None or sequence of str, optional
            Non-rain variables to use.
        """
        rain_i = self.var_names.index("R")
        if var_names is None:
            var_names = self.var_names
        non_rain_i = [self.var_names.index(var_name)
                      for var_name in var_names
                      if var_name != "R"]
        non_rain = data_trans[non_rain_i]
        threshold = conf.dists_kwds["R"]["threshold"]
        rain_mask = rain >= threshold
        # calculate means of non-rain variables during wet conditions
        wet_means = np.nanmean(non_rain[:, rain_mask], axis=1)[:, None]
        # distance between current vector of non-rain variables and mean of
        # wet conditions
        dists = np.sum(wet_means - non_rain[:, ~rain_mask], axis=0)
        # qq-transform the inverse distances to the lower tail of the gaussian
        dist_ranks = my.rel_ranks(-dists)
        # dryness probability in the standard-normal domain
        if doys[1] - doys[0] < 1:
            rain_dist, sol = self.dist_sol_hourly("R")
        else:
            rain_dist, sol = self.dist_sol["R"]
        rain_prob = rain_dist.all_parameters_dict(sol, doys)["rain_prob"]
        # trying something less deterministic
        # dry_prob = (1 - rain_prob) * np.random.random(size=len(rain_prob))
        dry_prob = 1 - rain_prob
        neg_rain = distributions.norm.ppf(dist_ranks * dry_prob[~rain_mask])
        data_trans[rain_i, ~rain_mask] = neg_rain

        # X = np.matrix(np.array([data_trans[i, rain_mask]
        #                         for i in non_rain_i]).T)
        # y = np.matrix(data_trans[rain_i, rain_mask]).T
        # beta = (X.T * X).I * X.T * y
        # rain_reg = np.array(data_trans[non_rain_i].T * beta)[:, 0]
        # data_trans[rain_i, ~rain_mask] = rain_reg[~rain_mask]

        # if self.verbose:
        #     print "Wet means in std-n:"
        #     non_rain_var_names = (var_name for var_name in var_names
        #                           if var_name != "R")
        #     for var_name, wet_mean in zip(non_rain_var_names, wet_means):
        #         print "\t%s: %.3f" % (var_name, wet_mean)

        #     import matplotlib.pyplot as plt
        #     fig, axs = plt.subplots(nrows=2)
        #     axs[0].plot(self.times, data_trans[rain_i], "-x")
        #     axs[0].plot(self.times[~rain_mask], neg_rain, "-x")
        #     axs[0].plot(self.times, distributions.norm.ppf(rain_prob))
        #     dists_wet = np.sum(wet_means - non_rain[:, rain_mask], axis=0)
        #     dists_ranks_wet = my.rel_ranks(-dists_wet)
        #     neg_rain_wet = distributions.norm.ppf(dry_prob[rain_mask] +
        #                                           rain_prob[rain_mask] *
        #                                           dists_ranks_wet)
        #     axs[0].plot(self.times[rain_mask], neg_rain_wet, "-x")
        #     axs[0].plot(self.times, rain_reg)

        #     # axs[1].scatter(data_trans[rain_i, rain_mask],
        #     #                neg_rain_wet, marker="x")
        #     axs[1].scatter(data_trans[rain_i, rain_mask],
        #                    rain_reg[rain_mask], marker="+", edgecolor="green")
        #     axs[1].set_aspect("equal")
        #     plt.show()

        return data_trans


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from vg.core import (vg, config_konstanz_disag,
                                                 vg_plotting)
    conf = config_konstanz_disag
    vg.conf = vg.vg_base.conf = vg_plotting.conf = conf
    met_vg = vg.VG(("R", "theta", "ILWR", "Qsw", "rh", "u", "v"),
                   verbose=True)
    met_vg.fit(p=3)
    simt, sim = met_vg.simulate()
    met_vg.plot_exceedance_daily()
    plt.show()
