#!/usr/bin/env python
import os
import warnings
import pickle
import datetime
import shutil
import copy

import numpy as np
from scipy import interpolate
# import matplotlib
# matplotlib.use("agg")
import matplotlib.pyplot as plt

from vg import times
from vg.meteo import meteox2y, avrwind
from vg.time_series_analysis import time_series as ts
from vg.time_series_analysis import models
from vg.time_series_analysis import distributions
from vg.time_series_analysis import resample as resampler
from vg.time_series_analysis import cresample
import vg.time_series_analysis.seasonal_distributions as sd
from vg import helpers as my
from vg.core.vg_plotting import VGPlotting
from vg.core import vg_base
read_met = vg_base.read_met

try:
    import config as conf
except ImportError:
    from vg import config_template as conf
    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]
    warnings.warn('Could not import "config.py".' + os.linesep +
                  'Edit "%s" and rename it to "config.py"' % conf_filepath)


def delete_cache():
    shutil.rmtree(conf.cache_dir)


def dump_data(times_, data, var_names, p, q, extra="", random_state=None,
              out_dir=None, conversions=None):
    """Dumps the given datetimes and data in an ascii file and random state in
    a pickle file."""
    if out_dir is None:
        out_dir = conf.out_dir
    if conversions:
        for conversion in list(conversions):
            times_, data, var_names = conversion(times_, data, var_names)
    outfilename = "VARMA_p%d_q%d_sim%s.dat" % (p, q, extra)
    outfilepath = os.path.join(out_dir, outfilename)
    with open(outfilepath, "w") as txt_file:
        txt_file.write("time\t" + "\t".join(var_names) + os.linesep)
        for time, values in zip(times_, data.T):
            txt_file.write("%s\t%s%s" %
                           (time.isoformat(),
                            "\t".join(conf.out_format[var_name] % val
                                      for var_name, val in
                                      zip(var_names, values)),
                            os.linesep))

    if random_state is not None:
        with open(outfilepath + ".random_state", "wb") as pi_file:
            pickle.dump(random_state, pi_file)

    return outfilepath


def metfile2df(met_file):
    import pandas as pd
    try:
        pd_kwds = dict(skiprows=5, index_col=0, parse_dates=True,
                       date_parser=times.cwr2datetime)
        met_df = pd.read_table(met_file, **pd_kwds).astype(float)
    except TypeError:
        def date_hour2datetime(date_str, hour_str):
            return times.str2datetime("%s %s" % (date_str, hour_str),
                                      "%d. %m %y %H")
        pd_kwds = dict(index_col=0, parse_dates=[[0, 1]],
                       date_parser=date_hour2datetime)
        met_df = pd.read_table(met_file, **pd_kwds).astype(float)
    return met_df


def outfile2df(outfile):
    import pandas as pd
    pd_kwds = dict(index_col=0, parse_dates=[0])
    return pd.read_table(outfile, **pd_kwds).astype(float)


def metfile2hdf5(met_file, h5_filename=None, key="met"):
    if h5_filename is None:
        h5_filename = os.path.splitext(met_file)[0] + ".h5"
    met_df = metfile2df(met_file)
    met_df.to_hdf(h5_filename, key)


def seasonal_back(dist_sol, norm_data, var_names, doys=None,
                  solution_template="%s"):
    """Transform variables from normal to real-world marginals.

    Parameter
    ---------
    dist_sol : mapping like dict or shelve object
        Distribution fitting parameters.
    norm_data : (K, T) array
        Normally distributed data.
    var_names : (K,) sequence of str
        Variable names used as dist_sol[solution_template % var_name].
    doys : None or (T,) float array, optional
        Only needed for different doys than the calibration dataset.
    solution_template : str, optional
        Used in the form dist_sol[solution_template % var_name]

    Returns
    -------
    data : (K, T) array
        Re-transformed data
    """
    data = np.empty_like(norm_data)
    for var_ii, (var_name, var) in enumerate(zip(var_names, norm_data)):
        distribution, solution = dist_sol[solution_template % var_name]
        # TODO: distribution is never distributions.norm. however the transform
        # yields wrong results!
        if distribution is distributions.norm:
            # if (hasattr(distribution, "dist") and
            #     type(distribution.dist) is distributions.Normal):
            # in this case, we do not need a qenuine qq-transform, the normal
            # inverse Z transform suffices, AND is immune to producing nans
            # for "extreme" values in the standard-normal world!
            if doys is not None:
                _T = (2 * np.pi / 365 * doys)[np.newaxis, :]
            sigmas = distribution.trig2pars(solution, _T)[1]
            data[var_ii] *= sigmas
        else:
            quantiles = distributions.norm.cdf(var, 0, 1)
            data[var_ii] = distribution.ppf(solution, quantiles, doys=doys)
    return data


def sim_byvariable(values, times, p=2, aggregation="%w"):
    aggr_values, aggr_times = \
        ts.time_part_average(values, times, aggregation, expand=False,
                             return_times=True)
    dist = sd.SeasonalDist(distributions.norm, aggr_values, aggr_times)
    solution = dist.fit(T0=50)
    trans = distributions.norm.ppf(dist.cdf(solution), 0, 1)
    B, sigma_u = models.VAR_LS(trans, p=p)
    T = len(aggr_values)
    sim = models.VAR_LS_sim(B, sigma_u, T - p)
    retrans = dist.ppf(solution, distributions.norm.cdf(sim, 0, 1))
    retrans = np.repeat(retrans, times.time_part_average.repeats)
    return retrans


def interannual_variability(T, mean_arrival=30, disturbance_std=.1):
    """Generate a time series used as a disturbance within the std-norm world.
    Disturbances are simulated as gaussian changes occuring after t timesteps,
    where t is drawn from an exponential distribution.
    """

    def arrival():
        return distributions.expon.ppf(np.random.rand(),
                                       lambd=1. / mean_arrival,
                                       # loc=1 weil wir taegliche Werte
                                       # berechnen und daher 1 Tag die
                                       # kuerzest moegliche Dauer ist
                                       x0=1)

    def disturbance():
        return abs(distributions.norm.ppf(np.random.rand(),
                                          sigma=float(disturbance_std)))
    m_t = np.zeros(T)
    breaks_ii = [t] = [int(round(arrival()))]
    while t < T:
        breaks_ii += [breaks_ii[-1] + int(round(arrival()))]
        t = breaks_ii[-1]
    sign = 1
    for break_, next_break in zip(breaks_ii[:-1], breaks_ii[1:]):
        period_length = min(int(round(next_break - break_)),
                            len(m_t) - break_)
        window = np.bartlett(period_length + 1)[:-1]
        if window.mean() <= 0:
            window = np.ones(period_length) / period_length
        else:
            window /= window.mean()
        m_t[break_:next_break] = sign * disturbance() * window
        sign *= -1
    return m_t


def sw_diurnal(date, daily_sw_data, del_t=3600):
    """ add standard daily cycle to daily mean sw data
    Parameters
    ----------
    date: array
        date in unix timestamps
    daily_sw_data: array
    del_t: int
        timestep of output data in seconds, default = 3600s

    Returns
    -------
    date_: array
        output date in unix timestamps
    data_: array
        sw data in del_t-steps"""
    nn = 86400. / del_t
    date_ = np.arange(date[0], date[-1] + 86400, del_t)
    swmax = meteox2y.pot_s_rad(times.unix2str(date_, "%Y-%m-%dT%H:%M"),
                               lat=conf.latitude, longt=conf.longitude)
    swmax_daily = np.average(swmax.reshape(-1, nn), axis=1).repeat(nn)
    fkt = swmax / swmax_daily
    data_ = daily_sw_data.repeat(nn) * fkt
    return date_, data_


class VG(VGPlotting):

    """A Vector-Autoregressive weather generator.

    >>> my_vg = VG(("theta", "Qsw", "ILWR", "rh", "u", "v"))
    >>> my_vg.fit()
    >>> times_out, sim_data = my_vg.simulate()
    """

    def __init__(self, var_names, met_file=None, sum_interval=24,
                 plot=False, separator="\t", refit=None,
                 detrend_vars=None, verbose=False, data_dir=None,
                 cache_dir=None, dump_data=True, non_rain=None,
                 **met_kwds):
        """A vector autoregressive moving average (VARMA) weather generator.
        Initializing a VG-object does the following:
            - read input data from a met-file
            - aggregate it according to sum_interval
            - convert the data using seasonal distributions

        To fit the VARMA process call fit(). Afterwards
        simulated time series can be obtained with simulate().

        Parameters
        ----------
        var_names : sequence containing strings
        met_file : string, dict, pandas DataFrame or None, optional
            If None, value will be read from config.py.
            Data can also be supplied directly with a dict, mapping variable
            names to ndarrays containing the data. The dictionary has to
            contain an item with the key "datetimes" which has to contain an
            ndarray of datetime objects.
            If data is supplied as a pandas DataFrame, time information will be
            taken from its index.
        sum_inverval : int or sequence of ints, optional
            Number of timesteps to aggregate. If a sequence is given, its
            elements are used variable - specific. In this case the sequence
            has to have the same length as var_names.
            Default 24 corresponds to the case of hourly input data and daily
            simulation
        plot : boolean, optional
            Plot all kinds of information. Warning: a lot of windows will
            appear.
        separator : str or None, optional
            String that separates columns in the met_file. It is fed to the
            split method of str, so None corresponds to any whitespace.
        refit : None, 'all' or sequence of strings from var_names, optional
            Instead of using cached solutions for the seasonal distribution
            fit, run the fit again and cache it.
        detrend_vars : sequence containing strings or None, optional
            Remove a linear trend from the variables.
        verbose : boolean, optional
            Be more verbose if True.
        data_dir : str, optional
            Overwrites the data_dir variable from the config-file.
        cache_dir : str, optional
            Overwrites the cache_dir variable from the config-file.
        dump_data : boolean, optional
            Write simulation and disaggregation results (filename can be
            infered by the outfilepath attribute).
        non_rain : None or sequence of str, optional
            Variables used for rain gap filling.
        """
        # external_var_names=None,
        # "fix" unicode problems
        var_names = [str(var_name) for var_name in var_names]
        super(VG, self).__init__(var_names, met_file=met_file,
                                 sum_interval=sum_interval, plot=plot,
                                 separator=separator, refit=refit,
                                 detrend_vars=detrend_vars,
                                 verbose=verbose, data_dir=data_dir,
                                 cache_dir=cache_dir,
                                 seasonal_fitting=True,
                                 dump_data=dump_data,
                                 non_rain=non_rain, **met_kwds)

    def __getattribute__(self, name):
        # VGPlotting overwrites __getattribute__ to do a seasonal fitting
        # only if it is (implicitly) requested. That behaviour is only usefull
        # when creating a VGPlotting instance
        return vg_base.VGBase.__getattribute__(self, name)

    def fit(self, p=None, q=0, p_max=10, seasonal=False, ex=None,
            ex_kwds=None):
        """p is the order of autoregressive process.
        q the order of the moving average process
        If p is not given, it is selected using the Schwartz Information
        Criterion."""
        self.p, self.q, self.seasonal = p, q, seasonal
        self.ex_in = ex
        if q is 0:
            if p is None:
                print "Perfoming VAR order selection: ",
                try:
                    if ex is None:
                        p = models.VAR_order_selection(self.data_trans, p_max)
                    else:
                        est_kwds = {} if ex_kwds is None else ex_kwds
                        est_kwds["ex"] = ex
                        o_select = models.VAR_order_selection
                        p = o_select(self.data_trans, p_max,
                                     estimator=models.VAREX_LS,
                                     est_kwds=est_kwds)
                    self.p = p
                except ValueError:
                    warnings.warn("Could not fit the VAR process. Trying to "
                                  "remove extreme values")
                    mi, ma = np.min(self.data_trans), np.max(self.data_trans)
                    if ma > -mi:
                        self.data_trans.ravel()[np.argmax(self.data_trans)] = \
                            np.nan
                    else:
                        self.data_trans.ravel()[np.argmin(self.data_trans)] = \
                            np.nan
                    self.data_trans = my.interp_nan(self.data_trans)
                    self.fit(p, q, p_max, seasonal, ex, ex_kwds)
                print "p=%d seems parsimonious to me" % self.p
                if p == 0:
                    print "Wow, you should not do any autoregressive modeling."
            # fit a VAR-model
            if self.seasonal:
                print "Fitting the seasonal VAR model."
                self.Bs, self.sigma_us = \
                    models.SVAR_LS(self.data_trans, self.data_doys, self.p,
                                   var_names=self.var_names,
                                   verbose=self.verbose)
                self.residuals = \
                    models.SVAR_residuals(self.data_trans, self.data_doys,
                                          self.Bs, self.p)
            else:
                print "Fitting the VAR model."
                if ex is None:
                    B, self.sigma_u = models.VAR_LS(self.data_trans, self.p)
                else:
                    B, self.sigma_u = models.VAREX_LS(self.data_trans, p, ex)
                # we call it AM to have one place for VARMA parameters
                self.AM = B
                self.residuals = models.VAR_residuals(self.data_trans, B,
                                                      self.p)
            self.p = p

        else:
            if (p is None) and (q is None):
                print "Perfoming VARMA order selection: ",
                self.p, self.q = p, q = \
                    models.VARMA_order_selection(self.data_trans, p_max,
                                                 plot_table=self.plot)
                print "p=%d and q=%d seems parsimonious to me" % (p, q)
            else:
                self.p, self.q = p, q
            # fit a VARMA-model
            print "Fitting the VARMA model"
            # the VARMA implementation is for mean-adjusted data. the data
            # should have zero mean, asuming the fit for the qq-transformation
            # was perfect.
            # ... and nothing is perfect. so we subtract the mean here
            self.means = self.data_trans.mean(axis=1)[:, np.newaxis]
            self.data_trans -= self.means

            AM, sigma_u, residuals = \
                models.VARMA_LS_prelim(self.data_trans, p, q)

            self.data_trans += self.means
            self.AM, self.sigma_u, self.residuals = AM, sigma_u, residuals

    def simulate(self, T=None, mean_arrival=None,
                 disturbance_std=None, theta_incr=None,
                 theta_grad=None, fixed_variables=None,
                 primary_var="theta", climate_signal=None,
                 start_str=None, stop_str=None, random_state=None,
                 ex=None, ex_kwds=None, seed_before_sim=None,
                 loc_shift=False, resample=False, res_kwds=None):
        """Simulate based on data from __init__ and the fit from calling
        self.fit().
        To choose a specific order of AR and MA call fit(p, q) before calling
        simulate. Otherwise fit will be called without parameters, so p and q
        will be guessed by applying an information criterion.

        Parameters
        ----------
        T : None or int, optional
            Number of timesteps to simulate. T=None produces as many simulation
            timesteps as there were in the data used for fitting the VARMA
            process.
        mean_arrival : float, optional
            average length of warm/cold periods [d]
        disturbance_std : float, optional
            standard deviation of deviation of warm/cold periods from mean
        theta_incr : float, optional
            increase in average of temperature [deg C]
        theta_grad : float, optional
            gradient increase in average of temperature [deg C / (T)]
        fixed_variables : dictionary mapping var_names to None or (T,) ndarrays
            Keeps the provided time - series fixed. Mappings to None indicate
            that the input data should be used as a fixed variable.
            Can be used to simulate hierarchically.
        primary_var : str, one of var_names from __init__ or sequence of those,
                optional
            All disturbances (mean_arrival, distrubance_std, theta_incr,
            theta_grad and climate_signal) correspond to changes in this
            variable.
        climate_signal : (T,) ndarray
            A time series of the 'primary_var'. Differences between
            'climate_signal' and the seasonal mean are used to perturb the
            VAR(MA) process.
        start_str : str of format '%m-%d %H', optional
            String representation of the start date of the simulation.
        stop_str : str of format '%m-%d %H', optional
            String representation of the end date of the simulation.
        random_state : str, tuple as returned by numpy.random.get_state() or
                       None, optional
            Sets the random state before simulating.
            *str* is interpreted as a pickle file containing the random state.
            *tuple* is interpreted as the random state itself.
        ex : (T,) ndarray or function
            External variable. If given as a function, ex_t will be generated
            by calling ex(Y[:t], **ex_kwds), with Y being the simulated values.
        ex_kwds : None or dict
            Keyword arguments to be passed to ex.
        seed_before_sim : None or int, optional
            Seed np.random before simulating. The integer passed will be used
            as seed (also if it is a 0).
            (_scenario_parameters also draws random numbers).
        loc_shift : bool, optional
            Change on the primary variables will be removed after simulation
            and added to the backtransformed variables.
        res_kwds : dict or None, optional
            Keywords for the resampler, if not None, use resampling.
            See `time_series_analysis.resample.resample` for allowed
            parameters.
        # resample : bool, optional
        #     Use the resampler to generate time series.
        # n_candidates : int, optional
        #     Tells the resampler how many time steps to consider.

        Returns
        -------
        sim_times : (T,) ndarray of datetime objects
        sim_sea : (K, T) float ndarray

        """

        if isinstance(random_state, basestring):
            with open(random_state, "rb") as pi_file:
                random_state = pickle.load(pi_file)
        if random_state is not None:
            np.random.set_state(random_state)
        # we record the random state to later dump it
        pre_sim_random_state = np.random.get_state()

        self.T = T
        self.disturbance_std, self.theta_incr, self.theta_grad, \
            self.mean_arrival = \
            map(lambda x: None if x is None else
                np.atleast_1d(x).astype(float),
                (disturbance_std, theta_incr, theta_grad, mean_arrival))
        self.fixed_variables = fixed_variables
        self.climate_signal = np.atleast_2d(climate_signal)
        if self.climate_signal.dtype == object:
            # we have to back up, this has not worked
            self.climate_signal = tuple(self.climate_signal[0])
        if isinstance(primary_var, basestring):
            self.primary_var = primary_var,
        else:
            self.primary_var = primary_var

        if T is None:
            if climate_signal is not None:
                # if we have multiple primary variables, elements of climate
                # signal can be None. we also have to be careful, because the
                # number of primary variables could be interpreted as T
                climate_signal = np.atleast_2d(climate_signal)
                _signal = np.array([_ for _ in climate_signal
                                    if _ is not None])
                self.T = _signal.shape[1]
            else:
                self.T = self.T_summed
        else:
            self.T = T

        # we depend on order selection if self.fit was not called
        if self.p is None:
            self.fit()

        self.sim_times = self._gen_sim_times(start_str=start_str,
                                             stop_str=stop_str)
        # self.T has to be reset when self._gen_sim_times is called, but not
        # inside self.disaggregate
        self.T = len(self.sim_times)

        # converts the given parameters so they are understood in the
        # std-normal world
        # to store effect of scenario parameters by primary variable
        self._m_single, self._m_t_single, self._m_trend_single = [], [], []
        m, m_t, m_trend = self._scenario_parameters()
        # store to be able to play outside with it later
        self.m, self.m_t, self.m_trend = m, m_t, m_trend

        fixed_data = self._prepare_fixed_data()

        if self.verbose:
            print "Simulating a time-series."
        if type(seed_before_sim) is int:
            np.random.seed(seed_before_sim)
        if resample or res_kwds is not None:
            # in contrast to the parametric models, we do not
            # transform anything
            # here we combine any change in theta_incr
            m_resampler = (m + m_t + m_trend[:, None])[self.primary_var_ii]
            try:
                res_kwds = my.ADict(res_kwds)
            except TypeError:
                res_kwds = my.ADict()
            cy = "cy" in res_kwds and res_kwds["cy"]
            if cy:
                sim, self.res_indices, self.candidates = \
                    cresample.resample(self.data_trans, self.times,
                                       self.p,
                                       n_sim_steps=self.T,
                                       theta_incr=m_resampler,
                                       theta_i=self.primary_var_ii,
                                       cache_dir=conf.cache_dir,
                                       verbose=self.verbose,
                                       return_candidates=True,
                                       **(res_kwds - "cy"))
            else:
                sim, self.res_indices, self.candidates = \
                    resampler.resample(self.data_trans, self.times,
                                       self.p,
                                       n_sim_steps=self.T,
                                       theta_incr=m_resampler,
                                       theta_i=self.primary_var_ii,
                                       cache_dir=conf.cache_dir,
                                       verbose=self.verbose,
                                       return_candidates=True,
                                       **res_kwds)
        elif self.q in (0, None):
            # simulate VAR-time-series
            if self.seasonal:
                sim = models.SVAR_LS_sim(self.Bs, self.sigma_us, self.sim_doys,
                                         m, ia=m_t, m_trend=m_trend,
                                         fixed_data=fixed_data)
            else:
                if ex is None:
                    sim = models.VAR_LS_sim(self.AM, self.sigma_u, self.T, m,
                                            ia=m_t, m_trend=m_trend,
                                            fixed_data=fixed_data)
                else:
                    sim, self.ex_out = \
                        models.VAREX_LS_sim(self.AM, self.sigma_u, self.T, ex,
                                            m, ia=m_t, m_trend=m_trend,
                                            ex_kwds=ex_kwds)
        else:
            sim = models.VARMA_LS_sim(self.AM, self.p, self.q, self.sigma_u,
                                      # process means that are used as
                                      # starting values there is some
                                      # confusion here...
                                      self.means,
                                      self.T, m, ia=m_t, m_trend=m_trend,
                                      fixed_data=fixed_data)

        if self.plot:
            ts.matr_img(self.sigma_u, "Noise Covariance matrix",
                        self.var_names, self.var_names)

        # location shifting
        if loc_shift:
            sim = self._location_shift_normal(sim)

        # transform back
        if resample or res_kwds is not None:
            # use the indices to map back, we do not want any new
            # values to occur here!
            sim_sea = np.array([self.data_raw[self.var_names.index(var_name),
                                              self.res_indices]
                                for var_name in self.var_names])
            sim_sea /= self.sum_interval
        else:
            sim_sea = seasonal_back(self.dist_sol, sim, self.var_names,
                                    doys=self.sim_doys)
            sim_sea /= self.sum_interval

        if loc_shift:
            sim_sea = self._location_shift_back(sim_sea)

        if self.verbose:
            # checking the change of moments
            # try:
            # mean_before = my.nanavg(self.data_raw[self.primary_var_ii] /
            #                         self.sum_interval[self.primary_var_ii])
            mean_before = [my.nanavg(self.data_raw[var_i] /
                                     self.sum_interval[var_i])
                           for var_i in self.primary_var_ii]
            mean_after = my.nanavg(sim_sea[self.primary_var_ii],
                                   axis=1)
            print (("\t%s mean (data): %s, %s mean (sim): %s, "
                    "diff: %s")  # , mean(m): %.4f")
                   % (", ".join(self.primary_var),
                      ", ".join("%.3f" % mean for mean in mean_before),
                      ", ".join(self.primary_var),
                      ", ".join("%.3f" % mean for mean in mean_after),
                      ", ".join("%.3f" % diff for diff in
                                mean_after - mean_before),
                      ))  # m[self.primary_var_ii].mean()))
            # except (UnboundLocalError, NameError, ValueError, TypeError):
            #     pass

        # lets store this so we can play with it from the outside
        self.sim, self.sim_sea = sim, sim_sea

        if self.dump:
            self.outfilepath = \
                dump_data(self.sim_times, self.sim_sea, self.var_names, self.p,
                          0 if self.q is None else self.q,
                          random_state=pre_sim_random_state,
                          out_dir=self.data_dir, conversions=conf.conversions)

        return self.sim_times, sim_sea

    def disaggregate(self, var_names_dis=None, event_dt=None, factors=None,
                     doy_tolerance=15):
        """Disaggregate variables to hourly time steps by drawing from the
        residuals of the measured data

        Parameters
        ----------
        var_names_dis : list of str, optional
            names of variables to be disaggregated. if None all variables are
            disaggregated
        event_dt : array of datetimes, optional
            days on which factors will be applied to change daily cycle
            amplitudes.
        factors : sequence of length K
            see event_dt

        Returns
        -------
        times_out : array of datetime objects
        sim_sea_dis : array
            simulated variables in hourly time steps. the last day gets lost
        """
        if self.sim is None:
            raise RuntimeError("Call simulate first.")

        # disaggregate selected variables to hourly values, all the others
        # will be repeated 24 times a day. The last day gets lost
        if var_names_dis is None:
            print "Disaggregating all variables"
            var_names_dis = self.var_names
        elif isinstance(var_names_dis, basestring):
            print "Disaggregating selected variables."
            var_names_dis = var_names_dis,

        # how many timesteps are there per day in the input data?
        # we assume that the data is not finer than hours
        hours_unique = np.unique([dtime.hour for dtime in self.times_orig])
        # read timesteps per day (abbreviation, because it is used a lot as an
        # index modifier below
        tpd = len(hours_unique)

        self.dis_times = self._gen_sim_times((self.T - 1) * tpd,
                                             output_resolution=1.)
        # these are in "hourly" discretization
        doys_in = self.data_doys_raw[:-tpd]

        if event_dt is not None:
            # the event datetimes have to be mapped into the hourly
            # discretization
            year_days_out = times.time_part(self.dis_times, "%Y%j")
            year_days_events = times.time_part(event_dt, "%Y%j")
            event_ii = np.where(year_days_out == year_days_events[:, None])[1]

        if factors is not None:
            factors = np.asarray(factors)[:, None]

        # due to the interpolation with interp1d, we often have to skip the
        # last day (24 hours)
        # size of pool we can draw from:
        nn = self.T_data // tpd * tpd - tpd
        m = self.T * tpd - tpd  # no of hourly timesteps (simulated)
        sim_sea_dis = self.sim_sea.repeat(tpd).reshape(-1, m + tpd)[:, :-tpd]
        deltas_input = np.zeros((self.K, nn))
        sim_interps = np.copy(sim_sea_dis)
        for var_name in var_names_dis:
            var_i = self.var_names.index(var_name)
            # hourly measured values
            var_h = my.interp_nan(self.met[var_name][:nn + tpd], max_interp=3)
            # daily averages of measured values:
            var_d = my.nanavg(var_h.reshape(-1, tpd), axis=1)
            # 'time' array for interp1d
            hourly_input_times = np.arange(var_h.shape[0])
            # interpolate between daily values (measurements)
            f_interp = interpolate.interp1d(hourly_input_times[::tpd], var_d)
            var_interp = f_interp(hourly_input_times[:-tpd])
            # residuals to draw from
            deltas_input[var_i] = var_h[:-tpd] - var_interp
            # if the variable has lower/upper limits, store the percentage of
            # the distance covered between the interpolated line and the limit
            limits = copy.copy(conf.par_known[var_name])
            if var_name.startswith("Qsw"):
                def pot_s(doys):
                    hourly = meteox2y.pot_s_rad(doys, lat=conf.latitude,
                                                longt=conf.longitude)
                    return hourly * self.sum_interval[var_i]
                limits["u"] = pot_s

            pos_mask = deltas_input[var_i] > 0
            neg_mask = ~pos_mask

            # we have to interpret a threshold as a lower limit (no negative
            # rain, please)
            seas_dist, _ = self.dist_sol[var_name]
            if (hasattr(seas_dist, "dist") and
                    hasattr(seas_dist.dist, "thresh")):
                if limits is None:
                    limits = {}
                limits["l"] = conf.array_gen(seas_dist.dist.thresh)
#                 limits["u"] = conf.array_gen(1.e12)

            if limits is not None and ("l" in limits or "lc" in limits):
                lower_func = limits["l"] if "l" in limits else limits["lc"]
                lower = (lower_func(doys_in[neg_mask]) /
                         self.sum_interval[var_i])
                deltas = deltas_input[var_i, neg_mask]
                interps = var_interp[neg_mask]
                div_mask = ~np.isclose(interps, lower)
                lower_perc = np.full_like(interps, 1e-12)
                lower_perc[div_mask] = deltas[div_mask] / (interps[div_mask] -
                                                           lower[div_mask])
#                 lower_perc = deltas / (interps - lower)
#                 lower_perc[np.isnan(lower_perc)] = 1e-12
                lower_perc[interps < lower] = 1.
                lower_perc[lower_perc < -1.] = -1.
                deltas_input[var_i, neg_mask] = lower_perc
            if limits is not None and ("u" in limits or "uc" in limits):
                upper_func = limits["u"] if "u" in limits else limits["uc"]
                upper = (upper_func(doys_in[pos_mask]) /
                         self.sum_interval[var_i])
                deltas = deltas_input[var_i, pos_mask]
                interps = var_interp[pos_mask]
                upper_perc = deltas / (upper - interps)
                upper_perc[interps > upper] = 1
                upper_perc[upper_perc > 1.] = 1.
                deltas_input[var_i, pos_mask] = upper_perc

            # not a real 'time' as above
            hourly_output_times = np.arange(self.T * tpd)
            # interpolate between daily values (simulation)
            f_interp = interpolate.interp1d(hourly_output_times[::tpd],
                                            self.sim_sea[var_i])
            sim_interps[var_i] = f_interp(hourly_output_times[:-tpd])
        # drawing from pool: use same indices for all to preserve dependencies
        nan_mask = np.any(np.isnan(deltas_input), axis=0)
        chosen_chunks = self._shuffle(nn, m, tpd, 2 * tpd, doys_in,
                                      self.dis_doys, nan_mask=nan_mask,
                                      doy_tolerance=doy_tolerance)
        deltas_drawn = deltas_input[:, chosen_chunks]
        if event_dt is not None and factors is not None:
            deltas_drawn[:, event_ii] *= factors
        sim_sea_dis = sim_interps + deltas_drawn

        #  check for variables with limits
        for var_name in var_names_dis:
            var_i = self.var_names.index(var_name)
            limits = copy.copy(conf.par_known[var_name])
            if var_name.startswith("Qsw"):
                def pot_s(doys):
                    hourly = meteox2y.pot_s_rad(doys, lat=conf.latitude,
                                                longt=conf.longitude)
                    return hourly * self.sum_interval[var_i]
                limits["u"] = pot_s

            sim_interp = sim_interps[var_i]
            pos_mask = deltas_drawn[var_i] > 0
            neg_mask = ~pos_mask

            # we have to interpret a threshold as a lower limit (no negative
            # rain, please)
            seas_dist, _ = self.dist_sol[var_name]
            if (hasattr(seas_dist, "dist") and
                    hasattr(seas_dist.dist, "thresh")):
                if limits is None:
                    limits = {}
                limits["l"] = conf.array_gen(seas_dist.dist.thresh)
#                 limits["u"] = conf.array_gen(1.e12)

            if limits is not None and ("u" in limits or "uc" in limits):
                upper_func = limits["u"] if "u" in limits else limits["uc"]
                upper = (upper_func(self.dis_doys[pos_mask]) /
                         self.sum_interval[var_i])
                upper_perc = deltas_drawn[var_i, pos_mask]
                interps = sim_interp[pos_mask]
                upper_dis = interps + upper_perc * (upper - interps)
                sim_sea_dis[var_i, pos_mask] = upper_dis
            if limits is not None and ("l" in limits or "lc" in limits):
                lower_func = limits["l"] if "l" in limits else limits["lc"]
                lower = (lower_func(self.dis_doys[neg_mask]) /
                         self.sum_interval[var_i])
                lower_perc = deltas_drawn[var_i, neg_mask]
                interps = sim_interp[neg_mask]
                lower_dis = interps + lower_perc * (interps - lower)
                sim_sea_dis[var_i, neg_mask] = lower_dis

        self.var_names_dis = var_names_dis
        self.sim_sea_dis = sim_sea_dis

        if self.dump:
            # if varnames_dis==None, daily values are returned and dumped:
            self.outfilepath = \
                dump_data(self.dis_times, sim_sea_dis, self.var_names, self.p,
                          0 if self.q is None else self.q, "_disaggregated",
                          out_dir=self.data_dir, conversions=conf.conversions)

        return self.dis_times, sim_sea_dis

    def _prepare_output(self):
        """Makes the requested conversions defined in the config file and
        returns a dictionary for methods like to_dyresm or to_glm."""
        if self.sim_sea_dis is None:
            raise RuntimeError("Call disaggregate first.")

        dtimes, data, var_names = \
            self.dis_times, self.sim_sea_dis, self.var_names
        if conf.conversions:
            for conversion in list(conf.conversions):
                times, data, var_names = conversion(dtimes, data, var_names)
        met = {var_name: vals
               for var_name, vals in zip(var_names, data)}
        return dtimes, met

    def to_dyresm(self, outfilepath, ts=3600, info=None, wind_fct=1.0,
                  output_rh=False):
        """Converts output to a DYRESM input file.

        Parameters
        ---------
        outfilepath : str
            dyresm meteorological bc file
        ts : int
            timestep in dyresm-met-file in seconds, should be divisor of 86400
        info : information text string in file header
        wind_fct : float
            factor for wind speed. Default: 1.0 (for Konstanz wind data: 1.3)

        """
        times_, met_dict = self._prepare_output()

        # the now string! ready cut, easy to handle, simpson's individual
        # emperor stringettes - just the right length!
        now_str = datetime.datetime.now().isoformat(" ")
        moist_varname = "rh" if output_rh else "e"
        template_filepath = os.path.join(os.path.dirname(__file__),
                                         "dyresm_header_template")
        with open(template_filepath) as template_file:
            template = template_file.read()
        dyresm_header = template.format(now_str=now_str, info=info,
                                        ts=ts, moist_varname=moist_varname)

        if not output_rh:
            if "e" not in met_dict:
                met_dict["e"] = meteox2y.rel2vap_p(met_dict["rh"],
                                                   met_dict["theta"])
        if "U" not in met_dict:
            met_dict["U"] = avrwind.component2angle(met_dict["u"],
                                                    met_dict["v"])[1]
        met_dict["U"] *= wind_fct
        if "R" not in met_dict:
            met_dict["R"] = np.zeros_like(met_dict["theta"])
        data = np.array([times.datetime2cwr(times_)] +
                        [met_dict[var_name] for var_name in
                         ("Qsw", "ILWR", "theta",
                          "rh" if output_rh else "e",
                          "U", "R")]
                        ).T

        with open(outfilepath, "w") as outfile:
            outfile.write(dyresm_header)
            np.savetxt(outfile, data,
                       fmt=('%10.5f\t%7.3f\t%7.3f\t%6.2f\t' +
                            ("%.3f" if output_rh else "%4.1f") +
                            '\t%5.2f\t%.6f'))

    def to_glm(self, outfilepath):
        """Converts output to a GLM input file.

        Parameters
        ----------
        outfilepath : path/filename
            glm meteorological bc file
        """
        dtimes, met = self._prepare_output()

        header = ("time,ShortWave,LongWave, AirTemp,RelHum, WindSpeed,Rain,"
                  "Snow" + os.linesep)
        if "U" not in met:
            try:
                met["U"] = avrwind.component2angle(met["u"], met["v"])[1]
            except KeyError:
                warnings.warn("No wind information found. Filling in 0's!!")
                met["U"] = np.zeros_like(met[met.keys()[0]])
        if "R" not in met:
            warnings.warn("No precipitation information available. Filling "
                          " in with 0's!!")
            met["R"] = np.zeros_like(met[met.keys()[0]])
        times_str = times.datetime2str(dtimes, "%Y-%m-%d %H:%M")
        lines = [",".join([time_str] +
                          ["%.6f" % met[key][i]
                           for key in
                           ("Qsw", "ILWR", "theta", "rh", "U", "R")] +
                          ["0.0" + os.linesep]  # snow
                          )
                 for i, time_str in enumerate(times_str)]
        with open(outfilepath, "w") as glm_file:
            glm_file.write(header)
            glm_file.writelines(lines)

    def to_gotm(self, outfilepath, sw_outfilepath=None):
        """Converts output to a GOTM input file set (met and separate
        short-wave file).

        Parameters
        ----------
        outfilepath : path/filename
            GOTM meteorological bc file
        sw_outfilepath : path/filename
            GOTM short-wave bc file

        """
        if sw_outfilepath is None:
            sw_outfilepath = outfilepath[:-len(".dat")] + "_swr.dat"
        dtimes, met = self._prepare_output()

        if "U" not in met:
            try:
                met["wdir"], met["U"] = \
                    avrwind.component2angle(met["u"], met["v"])
            except KeyError:
                warnings.warn("No wind information found. Filling in 0's!!")
                met["wdir"] = met["U"] = np.zeros_like(met[met.keys()[0]])
        if "R" not in met:
            warnings.warn("No precipitation information available. Filling "
                          " in with 0's!!")
            met["R"] = np.zeros_like(met[met.keys()[0]])
        if "air_pressure" not in met:
            met["air_pressure"] = np.zeros_like(met[met.keys()[0]])
        if "cloud_cover" not in met:
            ilwr = meteox2y.temp2lw(met["theta"])
            met["cloud_cover"] = meteox2y.lw2clouds(ilwr,
                                                    met["theta"],
                                                    rh=met["rh"])

        # relative humidity has to be in %
        rh = met["rh"]
        if np.max(rh) <= 1.5:
            met["rh"] *= 100

        times_str = times.datetime2str(dtimes, "%Y-%m-%d %H:%M:%S")
        var_names = "wdir", "U", "air_pressure", "theta", "rh", "cloud_cover"
        # fmts = "%.2f", "%.6f", "%.0f", "%.3f", "%.6f"
        lines = ["\t".join([time_str] +
                           [conf.out_format[key] % met[key][i]
                            for key in var_names] +
                           [os.linesep])
                 for i, time_str in enumerate(times_str)]

        with open(outfilepath, "w") as gotm_file:
            # gotm_file.write(header)
            gotm_file.writelines(lines)

        # separate short-wave file
        lines = ["%s\t%.3f%s" % (time_str, met["rh"][i], os.linesep)
                 for i, time_str in enumerate(times_str)]
        with open(sw_outfilepath, "w") as gotm_file:
            gotm_file.writelines(lines)

    def to_df(self, kind="hourly input", var_names=None):
        import pandas as pd
        if var_names is None:
            var_names = self.var_names

        if kind == "hourly input":
            if "e" in var_names and "e" not in self.met:
                self.met["e"] = meteox2y.rel2vap_p(self.met["rh"],
                                                   self.met["theta"])
            data = np.array([self.met[var_name] for var_name in var_names]).T
            df_kwds = dict(index=self.times_orig,
                           data=data,
                           columns=var_names)
        elif kind == "hourly output":
            data = np.array([self.sim_sea_dis[self.var_names.index(name)]
                             for name in var_names]).T
            df_kwds = dict(index=self.dis_times,
                           data=data,
                           columns=var_names)
        return pd.DataFrame(**df_kwds)

    def predict(self, dtimes=None, data_past=None, T=1, n_realizations=1):
        """Predict the next T steps given the past data.

        Parameters
        ----------
        dtimes : None or (p,) ndarray, optional
            datetimes of past timesteps, if None, the last p datetimes from
            the input data are used
        data_past : None or (K, p)) ndarray, optional
            Past on which to predict the future. p is the order of the VAR -
            process (it is stored in self.p after self.fit() is called).
            If None, data from the input is used.
        T : int, optional
            number of timesteps to predict.
        n_realizations : int, optional
            number of realizations of prediction

        Returns
        -------
        prediction : (K, T) or (K, T, n_realizations) ndarray
        """
        if self.p is None:
            self.fit()
        if dtimes is None:
            dtimes = self.times[-self.p:]
        if data_past is None:
            data_past = self.data_raw[:, -self.p:]
        past_doys = times.datetime2doy(dtimes)
        data_past_trans = \
            self._fit_seasonal(values=data_past, doys=past_doys)[0]
        prediction = models.VAR_LS_predict(data_past_trans, self.AM,
                                           self.sigma_u, T, n_realizations)
        prediction = prediction.reshape((-1, T, n_realizations))

        # calculate the output resolution in days
        first_doy, second_doy = times.datetime2doy(self.times[:2])
        delta_doy = second_doy - first_doy
        prediction_doys = past_doys[-1] + delta_doy * np.arange(1, T + 1)
        prediction_dtimes = times.doy2datetime(prediction_doys,
                                               year=dtimes[-1].year)
        prediction_doys = np.where(prediction_doys > 366,
                                   prediction_doys - 366, prediction_doys)
        prediction_sea = np.atleast_3d(np.empty_like(prediction))

        for r in range(n_realizations):
            prediction_sea[..., r] = \
                seasonal_back(self.dist_sol, prediction[..., r],
                              self.var_names, doys=prediction_doys)
        prediction_sea /= self.sum_interval[:, np.newaxis, :]
        return prediction_dtimes, np.squeeze(prediction_sea)

    def plot_prediction(self, dtimes=None, data_past=None, T=1, hindcast=True):
        # don't be verbose for this part (we fetch previous fits all the time
        # which is rather boring)
        verbose = self.verbose
        self.verbose = False
        if self.p is None:
            self.fit()
        n_prev_steps = (4 if hindcast else 2) * self.p
        if dtimes is None:
            dtimes = self.times[-n_prev_steps:]
        if data_past is None:
            data_past = self.data_raw[:, -n_prev_steps:]

        prediction_dtimes, prediction = \
            self.predict(dtimes[-self.p:], data_past[:, -self.p:], T)

        fig, axes = plt.subplots(nrows=self.K, sharex=True, squeeze=True)
        fig.suptitle("Prediction made on %s" %
                     times.datetime2str(times.datetime.now()))
        for var_i, (ax, var_name) in enumerate(zip(axes, self.var_names)):
            ax.plot(dtimes[self.p:],
                    data_past[var_i, self.p:] / self.sum_interval[var_i], "k",
                    linewidth=2)
            ax.plot(prediction_dtimes, prediction[var_i], "b-x", linewidth=2)
            ax.grid()
            ax.set_ylabel(r"%s %s" % (var_name, conf.units[var_name]))
            ax.set_title(conf.long_var_names[var_name])

        if hindcast:
            for shift in range(self.p + 1, n_prev_steps):
                hindcast_dtimes, hindcast = \
                    self.predict(dtimes[:shift], data_past[:, :shift],
                                 n_prev_steps - shift, 1)
                col = plt.cm.jet(1 -
                                 float(shift - self.p - 1) /
                                 (n_prev_steps - self.p - 1),
                                 alpha=.5)
                for var_i, ax in enumerate(axes):
                    ax.plot(hindcast_dtimes, hindcast[var_i], "-",
                            color=col, linewidth=1)
                    # connect the hindcast with the end of its supporting data
                    ax.plot([dtimes[shift - 1], hindcast_dtimes[0]],
                            [data_past[var_i, shift - 1] /
                             self.sum_interval[var_i],
                             np.atleast_1d(hindcast[var_i])[0]], "-x",
                            color=col, linewidth=1)
        ti = plt.xticks()[0]
        la_ = [times.datetime2str(times.ordinal2datetime(tii), "%d.%m. %H h")
               for tii in ti]
        plt.xticks(ti, la_, rotation=30)

        self.verbose = verbose

    def _gen_m_trend(self, prim_i, prim_index, prim_is_normal, sigma, scale,
                     scale_nn):
        theta_grad = self.theta_grad[prim_i]
        if np.isnan(theta_grad):
            self._m_trend_single += [np.zeros(self.K)]
            return self._m_trend_single[-1]
        # e.g. gradual temperature change
        primvar_trend = theta_grad * self.sum_interval[prim_index]
        # looking for the mean of the standard deviation
        if prim_is_normal:
            m_trend_primvar = primvar_trend / np.mean(sigma)
        else:
            # TODO: the following is most likely wrong!!!
            m_trend_primvar = np.mean(scale_nn(primvar_trend))
        self._m_trend_single += [m_trend_primvar * scale]
        return self._m_trend_single[-1]

    def _gen_m(self, prim_i, prim_index, prim_is_normal, sigma, scale,
               scale_nn):
        theta_incr = self.theta_incr[prim_i]
        if np.isnan(theta_incr):
            self._m_single += [np.zeros((self.K, self.T))]
            return self._m_single[-1]
        intercept = (self.data_trans.mean(axis=1) -
                     scale * self.data_trans[prim_index].mean()
                     ).reshape((self.K, 1))
        intercept[prim_index] = 0
        # change in mean in the real world
        delta_primvar = theta_incr * self.sum_interval[prim_index]
        if prim_is_normal:
            m_primvar = delta_primvar / sigma
        else:
            m_primvar = scale_nn(delta_primvar)
        self._m_single += [m_primvar * scale.reshape((self.K, 1)) + intercept]
        return self._m_single[-1]

    def _gen_m_t(self, prim_i, prim_index, prim_is_normal, sigma, scale,
                 scale_nn, _T):
        m_t = np.zeros((self.K, self.T))
        # das war an der Tafel im Seminarraum 2:
        # mean_arrival: mittlere Zeit zwischen Aenderungen (kalte/warme
        # Perioden)
        # disturbance_std: Stabw der disturbances
        if self.disturbance_std is not None:
            disturbance_std = self.disturbance_std[prim_i]
            mean_arrival = self.mean_arrival[prim_i]
            disturbance_std = disturbance_std * self.sum_interval[prim_index]
            if prim_is_normal:
                disturbance_std /= np.mean(sigma)
            else:
                disturbance_std = np.mean(scale_nn(disturbance_std))
            m_t_primvar = \
                interannual_variability(self.T, mean_arrival, disturbance_std)
            m_t += m_t_primvar * scale.reshape((self.K, 1))

        if (self.climate_signal is not None and
                np.array(self.climate_signal).ndim > 1 and
                self.climate_signal[prim_i] is not None):
            climate_diff = self._get_climate_diff(_T, prim_i)
            if prim_is_normal:
                climate_diff /= sigma
            else:
                climate_diff = scale_nn(climate_diff)
            m_t += climate_diff * scale.reshape((self.K, 1))
        self._m_t_single += [m_t]
        return m_t

    @property
    def cov_trans(self):
        """The covariance matrix of the transformed data.
        If seasonal is True, this is of (365, K, K)-shape.
        """
        if self._cov_trans is None:
            if self.seasonal:
                covs = []
                for doy_ii, doy in enumerate(self.unique_doys):
                    ii = ((self.data_doys > doy - self.doy_width) &
                          (self.data_doys <= doy + self.doy_width))
                    if (doy - self.doy_width) < 0:
                        ii |= self.doys > (365. - self.doy_width + doy)
                    if (doy + self.doy_width) > 365:
                        ii |= self.doys < (doy + self.doy_width - 365.)
                    covs += np.cov(self.data_trans[ii])
                covs = [[my.fourier_approx(covs[:, i, j], 4)
                         for i in range(self.K)]
                        for j in range(self.K)]
                self._cov_trans = np.array(covs)
            else:
                self._cov_trans = np.cov(self.data_trans)
        return self._cov_trans

    def _scenario_parameters(self):
        """Prepares the scenario parameters, m and m_trend for the
        std.-normal world."""
        m = np.zeros((self.K, self.T))
        m_t = np.zeros_like(m)
        m_trend = np.zeros(self.K)

        # oddly enough, we allow more than one primary variable
        if isinstance(self.primary_var, basestring):
            prim_vars = self.primary_var,
        else:
            prim_vars = self.primary_var

        for prim_i, prim in enumerate(prim_vars):
            # prim_i is index inside sequence of primary variables
            # prim_index is index inside sequence of all variables
            prim_index = self.primary_var_ii[prim_i]
            # scale other variables according to linear regression
            scale = (np.cov(self.data_trans)[prim_index] /
                     np.var(self.data_trans[prim_index]))

            # do not change the primary variable by itself
            # scale[prim_index] = 1
            for index in self.primary_var_ii:
                scale[index] = 1
            if prim in self.var_names:
                seas_dist, trig_params = self.dist_sol[prim]
                if hasattr(seas_dist, "dist"):
                    prim_is_normal = \
                        type(seas_dist.dist) == distributions.Normal
                else:
                    prim_is_normal = False

                if not prim_is_normal:
                    medians = seas_dist.ppf(trig_params,
                                            .5 * np.ones(len(self.sim_doys)),
                                            self.sim_doys)

                    def scale_nn(x):
                        return distributions.norm.ppf(
                            seas_dist.cdf(trig_params, medians + x,
                                          self.sim_doys))
                else:
                    scale_nn = None

                # get the doy-specific distribution parameters
                _T = (2 * np.pi / 365 * self.sim_doys)[np.newaxis, :]
                if prim_is_normal:
                    sigma = seas_dist.trig2pars(trig_params, _T)[1]
                else:
                    sigma = None

                if self.theta_grad is not None:
                    m_trend += self._gen_m_trend(prim_i, prim_index,
                                                 prim_is_normal, sigma, scale,
                                                 scale_nn)
                else:
                    self._m_trend_single += [np.zeros(self.K)]

                if self.theta_incr is not None:
                    m += self._gen_m(prim_i, prim_index, prim_is_normal, sigma,
                                     scale, scale_nn)
                else:
                    self._m_single += [np.zeros((self.K, self.T))]

                m_t += self._gen_m_t(prim_i, prim_index, prim_is_normal, sigma,
                                     scale, scale_nn, _T)

        return m, m_t, m_trend

    def _get_climate_diff(self, _T, prim_i=0):
        """Calculates the difference between the seasonal mean of the primary
        variable and the given climate signal."""
        var_name = self.primary_var[prim_i]
        return ((self.climate_signal[prim_i] -
                 self.fitted_medians(var_name)) *
                self.sum_interval_dict[var_name])

    def _location_shift_normal(self, sim):
        for prim_i, prim in enumerate(self.primary_var):
            prim_index = self.var_names.index(prim)
            dummy_time = np.arange(self.T, dtype=float)
            m = self._m_single[prim_i][prim_index]
            m_t = self._m_t_single[prim_i][prim_index]
            m_trend = self._m_trend_single[prim_i][prim_index]
            sim[prim_index] -= (m + m_t + dummy_time / self.T * m_trend)
        return sim

    def _location_shift_back(self, sim_sea):
        for prim_i, prim in enumerate(self.primary_var):
            prim_index = self.var_names.index(prim)

            if self.theta_incr is not None:
                theta_incr = self.theta_incr[prim_i]
                if not np.isnan(theta_incr):
                    sim_sea[prim_index] += theta_incr

            if self.theta_grad is not None:
                theta_grad = self.theta_grad[prim_i]
                if not np.isnan(theta_grad):
                    dummy_time = np.arange(self.T, dtype=float)
                    sim_sea[prim_index] += dummy_time / self.T * theta_grad

            if self.climate_signal is not None:
                if self.climate_signal[prim_i] is not None:
                    # get the doy-specific distribution parameters
                    _T = (2 * np.pi / 365 * self.sim_doys)[np.newaxis, :]
                    climate_diff = self._get_climate_diff(_T, prim_i)
                    sim_sea[prim_index] += (climate_diff /
                                            self.sum_interval[prim_index])
        return sim_sea

    def random_dryness(self, T=None, start_str=None, stop_str=None,
                       n_events=None, duration_min=2, duration_max=7,
                       month_start=6, month_end=9, event_dryness=.4):
        """This is a very special method designed to generate dry hot events
        for Lake Kinneret.

        Parameters
        ----------
        T : None or int, optional
            Number of timesteps to simulate. T=None produces as many simulation
            timesteps as there were in the data used for fitting the VARMA
            process.
        start_str : str of format '%m-%d %H', optional
            String representation of the start date of the simulation.
        stop_str : str of format '%m-%d %H', optional
            String representation of the end date of the simulation.
        n_events : int or None
            Number of events to generate. If none there will be an average of 3
            events per year.
        duration_min : int
            Minimum length of events
        duration_max : int
            Maximum length of events
        month_start : int
            Month when the events can start to happen (january = 1)
        month_end : int
            Latest month in which the events can happen (january = 1)
        event_dryness : float
            The value of relative humidity that will be set during the events.

        Returns
        -------
        rh_signal : ndarray, dtype=float
            To be used as climate_signal in VG.simulate
        event_dt : ndarray, dtype=object (datetime)
            Datetimes of the events. Can be passed to VG.disaggregate later.
        """
        if T is None:
            self.T = self.T_summed
        else:
            self.T = T
        self.sim_times = self._gen_sim_times(T, start_str, stop_str)
        if n_events is None:
            n_events = 3 * int(len(self.sim_doys) / 365.)

        rh_signal = np.copy(self.fitted_medians("rh", self.sim_doys))
        months = times.time_part(self.sim_times, "%m")
        summer_ii = np.where((months >= month_start) &
                             (months <= month_end))[0]
        durations = np.random.randint(duration_min, duration_max + 1, n_events)

        for event_i in range(n_events):
            i = np.random.choice(summer_ii)
            duration = durations[event_i]
            rh_signal[i:i + duration] = event_dryness

        event_mask = np.where(rh_signal <= .4)[0]
        event_dt = self.sim_times[event_mask]

        return rh_signal, event_dt


if __name__ == "__main__":
    # warnings.simplefilter("error", RuntimeWarning)

    # from vg.core import config_konstanz_disag as conf
    # vg_base.conf = vg_plotting.conf = conf
    met_vg = VG(("theta", "ILWR",
                 "Qsw", "rh", "u", "v"
                 ),
                # non_rain=("theta", "Qsw", "rh"),
                # refit="R",
                verbose=True, dump_data=True)
    met_vg.fit(3)
#     theta_signal = np.load("/tmp/theta_signal.npy")
#     rh_signal, event_dt = \
#         met_vg.random_dryness(start_str="01.01.2010 00:00:00",
#                               stop_str="02.01.2040 00:00:00")
#     simt, sim = met_vg.simulate(primary_var="rh", climate_signal=rh_signal)
    simt, sim = met_vg.simulate(start_str="01.01.2010 00:00:00",
                                stop_str="02.01.2040 00:00:00")
    simt_dis, sim_dis = met_vg.disaggregate()
#     met_vg.plot_meteogramm()
#     met_vg.plot_windrose()
#     met_vg.plot_timehists()
#     met_vg.plot_timehists(sim_dis, 6 * ["Disaggregated %s"], times_=simt_dis)
#     met_vg.plot_episode_hists()
#     met_vg.plot_doy_scatter()
#     met_vg.plot_qq()
#     met_vg.plot_daily_cycles()
#     met_vg.disaggregate(("u", "v"))
#     met_vg.plot_episode_hists()
#     met_vg.plot_autocorr()

#     plt.show()
#    simt, sim = met_vg.simulate(climate_signal=met_vg.fitted_theta_means + 4)
#    met_vg.plot = True
#    met_vg.disaggregate(("theta", "Cloudness", "u", "v"))
#    met_vg.disaggregate(("Qsw", "u", "v"))
#    plt.show()
#    met_vg.fit()
#    sim, stimes = met_vg.simulate()

#     met_vg = VG(("theta", "ILWR", "Qsw", "rh", "u", "v"), verbose=True)
#     met_vg.fit(2, seasonal=True)
#     st, sim = met_vg.simulate()
#     met_vg.plot_doy_scatter()
#     plt.show()
