import collections
import dumbdbm
import os
from tempfile import mkstemp
import warnings
import numpy as np
from scipy import stats
from matplotlib import gridspec, ticker
import matplotlib.pyplot as plt
from vg.meteo.windrose import windrose, seasonal_windroses
from vg.core import vg_base
import vg.time_series_analysis.seasonal_distributions as sd
import vg.time_series_analysis.time_series as ts
from vg.time_series_analysis import rain_stats
from vg.meteo import avrwind
from vg import times
from vg import helpers as my

try:
    import config as conf
except ImportError:
    from vg import config_template as conf
    conf_filepath = conf.__file__
    if conf_filepath.endswith(".pyc"):
        conf_filepath = conf_filepath[:-1]
    warnings.warn('Could not import "config.py".\n' +
                  'Edit "%s" and rename it to "config.py"' % conf_filepath)


def var_names_greek(var_names):
    return [conf.ygreek[var_name] for var_name in var_names]


class VGPlotting(vg_base.VGBase):

    """To be used indirectly (by the VG class) or directly to plot
    meteorological data (met_files, pandas dataframe).

    Caching features will be disabled to not overwrite anything more useful.
    """

    def __init__(self, var_names, met_file=None, sum_interval=24, plot=False,
                 separator="\t", refit=None, detrend_vars=None, verbose=False,
                 data_dir=None, cache_dir=None, seasonal_fitting=False,
                 dump_data=False, non_rain=None, **met_kwds):
        # signal for VGBase to not do the seasonal fitting, which takes time
        # and the reason to initiate VGPlotting might be that you just plot
        # something quickly
        self.seasonal_fitting = seasonal_fitting
        super(VGPlotting, self).__init__(var_names, met_file=met_file,
                                         sum_interval=sum_interval,
                                         plot=plot,
                                         separator=separator,
                                         refit=refit,
                                         detrend_vars=detrend_vars,
                                         verbose=verbose,
                                         data_dir=data_dir,
                                         cache_dir=cache_dir,
                                         dump_data=dump_data,
                                         non_rain=non_rain,
                                         **met_kwds)

    def plot_all(self, *args, **kwds):
        """Executes every method of VGPlotting that starts with "plot_".
        There are quite a number of those. You have been warned.
        """
        methods = {attr: getattr(self, attr)
                   for attr in dir(VGPlotting)
                   if (attr.startswith("plot_")
                       and attr != "plot_all")}
        return_values = []
        for meth_name, meth in methods.items():
            if self.verbose:
                print "Calling %s" % meth_name
            fig_ax = meth(*args, **kwds)
            return_values += [fig_ax]
        return return_values

    def fit_seasonal(self, refit=None, values=None, doys=None):
        """This is here, so that when initialized directly, seasonal
        distributions are fitted, but the existing fit to the data
        defined in the config file is not overwritten.
        """
        f_handle, self.seasonal_cache_file = mkstemp(suffix=".she")
        # need to establish the fact that the temp file is a data base
        db = dumbdbm.open(self.seasonal_cache_file, "n")
        db.close()
        os.close(f_handle)
        self.data_trans, self.dist_sol = \
            super(VGPlotting, self)._fit_seasonal(refit, values, doys)

    def __getattribute__(self, name):
        if name == "dist_sol" and not self.seasonal_fitting:
            # ok, we need the seasonal fitting, so provide it
            self.seasonal_fitting = True
            self.fit_seasonal(refit="all")
            return self.dist_sol
        return vg_base.VGBase.__getattribute__(self, name)

    @property
    def ylabels(self):
        return [conf.units[var] if var in conf.units else "[?]"
                for var in self.var_names]

    def plot_exceedance_daily(self, thresh=None):
        if "R" not in self.var_names:
            warnings.warn("No R in var_names, no plot_exceedance_daily")
            return
        if thresh is None:
            thresh = conf.dists_kwds["R"]["threshold"]
        rain_i = self.var_names.index("R")
        sim = self.sim_sea[rain_i]
        obs = self.data_trans[rain_i]
        return rain_stats.plot_exceedance(obs, sim, kind="all", thresh=thresh)

    def plot_exeedance_hourly(self, thresh=None):
        if "R" not in self.var_names:
            warnings.warn("No R in var_names, no plot_exceedance_hourly")
            return
        if thresh is None:
            thresh = conf.dists_kwds["R"]["threshold"]
        rain_i = self.var_names.index("R")
        sim = self.sim_sea_dis[rain_i]
        obs = self.met["R"]
        return rain_stats.plot_exceedance(obs, sim, kind="all", thresh=thresh)

    def plot_transformed_daily(self, var_names=None):
        return self.plot_meteogramm(var_names=var_names, trans=True)

    def plot_transformed_hourly(self, var_names=None):
        if hasattr(self, "data_trans_hourly"):
            return self.plot_meteogramm(var_names=var_names, trans=True,
                                        hourly=True)
        else:
            warnings.warn("No disaggregation, no plot_transformed_hourly")

    def plot_psd(self, var_names=None, hourly=False, *args, **kwds):
        """Plots power spectral density using matplotlib.pyplot.psd.

        Parameter
        ---------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        hourly : boolean, optional
            Plot hourly and disaggregated values.
        *args, **kwds :
            Passed on to matplotlib.pyplot.psd
        """
        if var_names is None:
            var_names = self.var_names

        fig, axs = plt.subplots(len(var_names))
        if hourly:
            data_obs = [self.met[var_name] for var_name in var_names]
            data_sim = [self.sim_sea_dis[self.var_names.index(var_name)]
                        for var_name in var_names]
            # time discretization
            dt = (self.dis_times[1] - self.dis_times[0]).total_seconds()
            suptitle = "Power spectral density - hourly"
        else:
            data_obs = [self.data_raw[self.var_names.index(var_name)]
                        for var_name in var_names]
            data_sim = [self.sim_sea[self.var_names.index(var_name)]
                        for var_name in var_names]
            dt = (self.sim_times[1] - self.sim_times[0]).total_seconds()
            suptitle = "Power spectral density - daily"

        for var_i, var_name in enumerate(var_names):
            axs[var_i].psd(data_obs[var_i], Fs=1 / dt,
                           scale_by_freq=True, label="observed",
                           *args, **kwds)
            axs[var_i].psd(data_sim[var_i], Fs=1 / dt,
                           scale_by_freq=True, label="simulated",
                           *args, **kwds)
            axs[var_i].set_xscale('log')
            axs[var_i].set_ylabel("%s [dB/Hz]" % var_name)
        axs[0].legend(loc="best")
        fig.suptitle(suptitle)

        return fig, axs

    def plot_doy_scatter(self, var_names=None, opacity=.4, hourly=False):
        """Plot variables over doys. Measurement and backtransformed simulated
        data will be plotted, if it is available.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        opacity : float in the range (0,1], optional
            Opacity of the scatter points.

        .. plot::

            met_vg = VG(("theta",))
            sim_times, sim = met_vg.simulate()
            met_vg.plot_doy_scatter()

        """
        if var_names is None:
            var_names = self.var_names
        # when given a string, this does not separate the letters, as tuple()
        # or list does
        var_names = np.atleast_1d(var_names)
        figs = []
        for var_name in var_names:
            var_ii = self.var_names.index(var_name)
            fig = plt.figure()
            figs.append(fig)
            if hourly:
                plt.scatter(times.datetime2doy(self.times_orig),
                            self.met[var_name],
                            marker="o", edgecolors=(0, 0, 1, opacity),
                            facecolors=(0, 0, 0, 0), label="measured"
                            )
            else:
                plt.scatter(self.data_doys,
                            self.data_raw[var_ii] / self.sum_interval[var_ii],
                            marker="o", edgecolors=(0, 0, 1, opacity),
                            facecolors=(0, 0, 0, 0), label="measured")

            if self.sim_sea is not None:
                if hourly:
                    plt.scatter(self.dis_doys, self.sim_sea_dis[var_ii],
                                marker="o", edgecolors=(1, 0, 0, opacity),
                                facecolors=(0, 0, 0, 0),
                                label="simulated backtransformed")
                else:
                    plt.scatter(self.sim_doys, self.sim_sea[var_ii],
                                marker="o", edgecolors=(1, 0, 0, opacity),
                                facecolors=(0, 0, 0, 0),
                                label="simulated backtransformed")
            plt.xlabel("doy")
            plt.ylabel("%s %s" % (var_names_greek(self.var_names)[var_ii],
                                  conf.units[var_name]))
            plt.legend(loc="best")
            plt.grid()
            title_str = "Seasonal scatter %s" % conf.long_var_names[var_name]
            plt.xlim(0, 366)
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str, fig.canvas.manager.num))
        return figs

    def plot_meteogramm(self, var_names=None, hourly=False,
                        trans=False, **f_kwds):
        """All variables over time in subplots.
        (Not as nice as the meteogramm from vg.meteo)
        """
        def meteogramm(time, data, suptitle, var_names):
            fig = plt.figure(**f_kwds)
            K = data.shape[0]
            gs = gridspec.GridSpec(K, 2, width_ratios=[8, 1])
            axs = np.empty((K, 2), dtype=object)
            for i in range(K):
                if i == 0:
                    axs[0, 0] = plt.subplot(gs[0, 0])
                else:
                    axs[i, 0] = plt.subplot(gs[i, 0], sharex=axs[i - 1, 0])
                axs[i, 1] = plt.subplot(gs[i, 1], sharey=axs[i, 0])

            for var_i, var_name in enumerate(var_names):
                axs[var_i, 0].plot(time, data[var_i], label=var_name)
                axs[var_i, 0].grid()
                axs[var_i, 0].set_ylabel("%s %s" % (conf.ygreek[var_name],
                                                    conf.units[var_name]))
                finite_mask = np.isfinite(data[var_i])
                try:
                    axs[var_i, 1].hist(data[var_i, finite_mask], 40,
                                       normed=True,
                                       orientation="horizontal")
                except TypeError:
                    pass
                axs[var_i, 1].set_xticklabels("")
            plt.suptitle(suptitle)
            return fig, axs

        if self.ex_in is None:
            data = self.data_raw / self.sum_interval
            var_names = self.var_names
        else:
            data = np.vstack((self.data_raw / self.sum_interval, self.ex_in))
            var_names = list(self.var_names) + ["external"]
        if not trans:
            fig, axs = meteogramm(self.times, data, "Measured daily",
                                  var_names)

        if self.sim_sea is not None and not trans:
            if self.ex_out is None:
                data = self.sim_sea
                var_names = self.var_names
            else:
                data = np.vstack((self.sim_sea, self.ex_out))
                var_names = list(self.var_names) + ["external"]
            fig_, axs_ = meteogramm(self.sim_times, data, "Simulated daily",
                                    var_names)
            fig, axs = [fig, fig_], [axs, axs_]

        if trans:
            if not hourly:
                fig, axs = meteogramm(self.times, self.data_trans,
                                      "Observed transformed", var_names)
                if self.sim is not None:
                    fig_, axs_ = meteogramm(self.sim_times, self.sim,
                                            "Simulated std-normal", var_names)
                    fig = [fig, fig_]
                    axs = [axs, axs_]
            if hourly:
                fig, axs = meteogramm(self.times_orig, self.data_trans_hourly,
                                      "Observed transformed hourly",
                                      var_names)

        if hourly or self.sim_sea_dis is not None:
            met_array = \
                vg_base.met_as_array(self.met, var_names=self.var_names)
            fig1, axs1 = meteogramm(self.times_orig, met_array,
                                    "Measured hourly", var_names)
            fig = [fig, fig1]
            axs = [axs, axs1]
            if self.sim_sea_dis is not None:
                fig2, axs2 = meteogramm(self.dis_times, self.sim_sea_dis,
                                        "Simulated hourly", var_names)
                fig += [fig2]
                axs += [axs2]

        return fig, axs

    def plot_doy_scatter_residuals(self, var_names=None, opacity=.4):
        if var_names is None:
            var_names = self.var_names
        # when given a string, this does not separate the letters, as tuple()
        # or list does
        var_names = np.atleast_1d(var_names)
        figs = []
        for var_ii, var_name in enumerate(var_names):
            fig = plt.figure()
            figs.append(fig)
            plt.scatter(self.data_doys,
                        self.data_trans[var_ii], marker="o",
                        edgecolors=(0, 0, 1, opacity),
                        facecolors=(0, 0, 0, 0), label="transformed")

            plt.scatter(self.data_doys, self.residuals[var_ii], marker="o",
                        edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0),
                        label="residuals")
            plt.xlabel("doy")
            plt.ylabel("%s %s" % (var_names_greek(self.var_names)[var_ii],
                                  conf.units[var_name]))
            plt.legend()
            plt.grid()
            title_str = ("Seasonal scatter of residuals %s" %
                         conf.long_var_names[var_name])
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str, fig.canvas.manager.num))
        return figs

    def plot_spaces(self, var_names=None, opacity=.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        for var_ii, var_name1 in enumerate(var_names):
            for var_jj in range(var_ii + 1, len(var_names)):
                var_name2 = var_names[var_jj]
                fig = plt.figure()
                figs.append(fig)
                plt.scatter(self.data_trans[var_ii], self.data_trans[var_jj],
                            marker="o", edgecolors=(0, 0, 1, opacity),
                            facecolors=(0, 0, 0, 0), label="transformed data")
                plt.scatter(self.sim[var_ii], self.sim[var_jj], marker="o",
                            edgecolors=(1, 0, 0, opacity),
                            facecolors=(0, 0, 0, 0), label="simulated")
                plt.legend()
                plt.xlabel(var_name1)
                plt.ylabel(var_name2)
                plt.grid()
                title_str = ("Spaces %s over %s" %
                             (conf.long_var_names[var_name2],
                              conf.long_var_names[var_name1]))
                plt.title(title_str)
                fig.canvas.set_window_title("%s (%d)" %
                                            (title_str,
                                             fig.canvas.manager.num))
        return figs

    def plot_diff_spaces(self, var_names=None, diff=1, opacity=.4):
        if var_names is None:
            var_names = self.var_names

        figs = []
        for var_ii, var_name in enumerate(var_names):
            fig = plt.figure()
            figs.append(fig)
            plt.scatter(self.data_trans[var_ii, diff:],
                        self.data_trans[var_ii, :-diff],
                        marker="o", edgecolors=(0, 0, 1, opacity),
                        facecolors=(0, 0, 0, 0), label="transformed data")
            plt.scatter(self.sim[var_ii, diff:], self.sim[var_ii, :-diff],
                        marker="o", edgecolors=(1, 0, 0, opacity),
                        facecolors=(0, 0, 0, 0), label="simulated")
            plt.legend()
            plt.xlabel("t")
            plt.ylabel("t - %d" % diff)
            plt.grid()
            title_str = ("Spaces %d-lagged %s" %
                         (diff, conf.long_var_names[var_name]))
            plt.title(title_str)
            fig.canvas.set_window_title("%s (%d)" %
                                        (title_str,
                                         fig.canvas.manager.num))
        return figs

    def plot_hourly_fit(self, var_names=None, *args, **kwds):
        """Plots hourly scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        return self._plot_seasonal_fit(var_names=None, hourly=True, *args,
                                       **kwds)

    def plot_daily_fit(self, var_names=None, *args, **kwds):
        """Plots daily scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        return self._plot_seasonal_fit(var_names=None, hourly=False, *args,
                                       **kwds)

    def _plot_seasonal_fit(self, var_names=None, plot_quantiles=True,
                           plot_fourier=True, plot_monthly=False,
                           hourly=False, opacity=.25, n_bins=15,
                           kde=True, s_kwds=None, *args, **kwds):
        """Plots seasonal scatter with fitted densities and a histogram of the
        transformed quantiles.

        Parameters
        ----------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        plot_quantiles : boolean, optional
            Plot histograms of the quantiles?
        plot_fourier : boolean, optional
            Plot the fourier fit to the seasonally changing distribution
            parameters.
        plot_monthly : boolean, optional
            Plot histograms showing how good the fit is on monthly separated
            data.
        opacity : float, optional
            Opacity of scatter.
        n_bins : int or sequence, optional
            Passed on to np.histogram when plotting histograms of quantiles.
        kde : boolean, optional
            Plot a kernel density estimation on the histograms of quantiles.
        s_kwds : None or dict, optional
            Keyword arguments passed on to plt.scatter.
        """
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, basestring):
            var_names = var_names,

        figs, axes = [], []
        for var_ii, var_name in enumerate(var_names):
            if hourly:
                try:
                    dist, solution = self.dist_sol_hourly(var_name)
                except KeyError:
                    warnings.warn("No hourly fit for %s available" % var_name)
                    continue
                n_sumup = 1
                values = self.met[var_name]
                doys = self.data_doys_raw
                discr_str = "hourly"
            else:
                dist, solution = self.dist_sol[var_name]
                n_sumup = self.sum_interval.ravel()[var_ii]
                values = self.data_raw[var_ii]
                doys = self.data_doys
                discr_str = "daily"

            if var_name in conf.long_var_names:
                long_var_name = conf.long_var_names[var_name]
            else:
                long_var_name = var_name

            fig, ax = dist.scatter_pdf(solution, title=long_var_name,
                                       n_sumup=n_sumup, opacity=opacity,
                                       s_kwds=s_kwds, *args, **kwds)
            plt.ylabel(conf.units[var_name]
                       if var_name in conf.units else "[?]")
            fig.canvas.set_window_title("Scatter pdf %s (%d) %s"
                                        % (var_name,
                                           fig.canvas.manager.num,
                                           discr_str))
            figs += [fig]
            axes += [ax]

            if plot_quantiles:
                fig = plt.figure(*args, **kwds)
                var_ii = self.var_names.index(var_name)
                quantiles = dist.cdf(solution, x=values, doys=doys)
                my.hist(quantiles, n_bins, fig=fig, kde=kde)
                plt.title("Quantiles %s" % var_name)
                fig.canvas.set_window_title("Quantiles %s (%d) %s"
                                            % (var_name,
                                               fig.canvas.manager.num,
                                               discr_str))
                figs += [fig]

            if plot_fourier:
                if isinstance(dist, sd.SeasonalDist):
                    fig, _ = dist.plot_fourier_fit()
                    fig.suptitle("%s %s" % (var_name, discr_str))
                    figs += [fig]

            if plot_monthly:
                if isinstance(dist, sd.SlidingDist):
                    fig = dist.plot_monthly_fit()
                    try:
                        for f in fig:
                            f.suptitle("%s %s" % (var_name, discr_str))
                    except TypeError:
                        fig.suptitle("%s %s" % (var_name, discr_str))
                    figs += [fig]
        return figs, axes

    def plot_qq(self, var_names=None, figsize=None, lines=True, *args,
                **kwds):
        if var_names is None:
            var_names = self.var_names
        if isinstance(var_names, basestring):
            var_names = var_names,
        n_axes = len(var_names)
        n_cols = int(np.ceil(n_axes ** .5))
        n_rows = int(np.ceil(float(n_axes) / n_cols))
        fig, axes = plt.subplots(n_rows, n_cols, subplot_kw={"aspect": 1},
                                 figsize=figsize)
        try:
            axes = axes.ravel()
        except AttributeError:
            axes = axes,
        for ax_i, var_name in enumerate(var_names):
            var_ii = self.var_names.index(var_name)
            # we have to limit ourself to the length of the simulated
            # data.
            # TODO: we are biasing this comparision because we take
            # the first values!
            finite_obs = np.isfinite(self.data_raw[var_ii])[:self.T]
            obs = np.sort(np.copy(self.data_raw[var_ii, finite_obs]))
            obs /= self.sum_interval[var_ii]
            sim = np.sort(np.copy(self.sim_sea[var_ii, finite_obs]))
            ax = axes[ax_i]
            if lines:
                ax.plot(sim, obs, color="grey")
            else:
                ax.scatter(sim, obs, marker="x", color="grey")
            global_min = min(obs[0], sim[0])
            global_max = max(np.nanmax(obs), sim[-1])

            if (self.sim_sea_dis is not None and
                self.var_names_dis is not None and
                    var_name in self.var_names_dis):

                finite_obs = np.isfinite(self.met[var_name])
                finite_obs = finite_obs[:self.sim_sea_dis.shape[1]]
                var_i_dis = self.var_names_dis.index(var_name)
                obs_dis = np.sort(np.copy(self.met[var_name][finite_obs]))
                sim_dis = self.sim_sea_dis[var_i_dis, finite_obs]
                sim_dis = np.sort(np.copy(sim_dis))
                if lines:
                    ax.plot(sim_dis, obs_dis, color="black")
                else:
                    ax.scatter(sim_dis, obs_dis, marker="x", color="black")
                global_min = min(global_min, obs_dis[0], sim_dis[0])
                global_max = max(global_max, np.nanmax(obs_dis),
                                 np.nanmax(sim_dis))

            ax.plot([global_min, global_max], [global_min, global_max], "k--")
            ax.grid()
            ax.set_ylabel("observed")
            if ax_i < (n_axes - n_cols):
                ax.set_xticklabels([])
            else:
                ax.set_xlabel("simulated")
            ax.set_title("%s %s" %
                         (conf.ygreek[var_name], conf.units[var_name]))
        # delete axes that we did not use
        if len(axes) > n_axes:
            for ax in axes[n_axes:]:
                fig.delaxes(ax)
            plt.draw()
        fig.canvas.set_window_title("QQ-plots")
        return fig, axes

    def plot_windrose(self, figsize=None, seasonal=False, *args, **kwds):
        try:
            ui, vi = self.var_names.index("u"), self.var_names.index("v")
        except ValueError:
            print "Can only plot wind roses if wind speed is simulated"
            return

        u = self.data_raw[ui] / self.sum_interval[ui]
        v = self.data_raw[vi] / self.sum_interval[vi]
        wind_dirs = avrwind.component2angle(u, v)[0]

        if figsize is not None:
            fig = plt.figure(figsize=figsize)
        else:
            fig = plt.figure()

        def rose_func_gen(times_):
            if seasonal:
                return lambda *args, **kwds: \
                    seasonal_windroses(times_, *args, **kwds)
            else:
                return lambda *args, **kwds: windrose(*args, **kwds)

        rose_func = rose_func_gen(self.times)
        figs = [rose_func(wind_dirs, fig=fig, *args, **kwds)]
        figs[0].suptitle("measured")

        if self.sim_sea is not None:
            u, v = self.sim_sea[ui], self.sim_sea[vi]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("simulated")

        if self.sim_sea_dis is not None:
            rose_func = rose_func_gen(self.times_orig)
            u, v = self.met["u"], self.met["v"]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("measured hourly")

            rose_func = rose_func_gen(self.dis_times)
            ui_dis = self.var_names_dis.index("u")
            vi_dis = self.var_names_dis.index("v")
            u, v = self.sim_sea_dis[ui_dis], self.sim_sea_dis[vi_dis]
            wind_dirs = avrwind.component2angle(u, v)[0]
            if figsize is not None:
                fig = plt.figure(figsize=figsize)
            else:
                fig = plt.figure()
            figs += [rose_func(wind_dirs, fig=fig, *args, **kwds)]
            figs[-1].suptitle("simulated disaggregated")

        return np.squeeze(figs)

    def plot_corr(self, ygreek=None, *args, **kwds):
        if ygreek is None:
            # get rid of the incident for this block
            ygreek = collections.defaultdict(lambda: r"-")
            ygreek.update({"R":      r'$R$',
                           "theta":  r'$\theta$',
                           "rh":     r"$\phi$",
                           "Qsw":    r"$Q_{sw}$",
                           "ILWR":   r"$Q_{lw(i.)}$",
                           "u":      r"$u$",
                           "v":      r"$v$",
                           })

        if self.ex_in is None:
            data = self.data_raw
            var_names = self.var_names
        else:
            data = np.vstack((self.data_raw, self.ex_in))
            var_names = list(self.var_names) + ["external"]

        def my_greek(var_names):
            return [ygreek[var_name] for var_name in var_names]

        greek_short = my_greek(var_names)
        fig = \
            ts.corr_img(data, 0, "Measured daily", greek_short, *args, **kwds)

        if self.sim_sea is not None:
            if self.ex_out is None:
                data = self.sim_sea
                var_names = self.var_names
            else:
                data = np.vstack((self.sim_sea, self.ex_out))
                var_names = list(self.var_names) + ["external"]
            greek_short = my_greek(var_names)
            fig_ = ts.corr_img(data, 0, "Simulated daily", greek_short, *args,
                               **kwds)
            fig = [fig, fig_]
        return fig

    def plot_VAR_par(self, **kwds):
        """Plots the parameters of the VAR-Process."""
        if self.AM is None:
            print "Call fit first"
            return
        return ts.matr_img(np.asarray(self.AM),
                           "AM p=%d q=%d" % (self.p, self.q), None,
                           self.var_names, **kwds)

    def plot_autocorr(self, maxlag=7, title=None, var_names=None, *args,
                      **kwds):
        """Plots autocorrelation of transformed and simulated data (if the
        latter is available).

        Parameters
        ----------
        maxlag : int, optional
            Maximum number of lags to plot
        """
        data = self.data_trans
        figs = []
        if self.sim is not None:
            data = [data, self.sim]
        if title is None:
            title_ = "Autocorrelation of standard-normal distributed variables"
        if var_names is None:
            var_names = var_names_greek(self.var_names)
        figs += [ts.plot_auto_corr(data, maxlag, title_, var_names, *args,
                                   **kwds)]

        data = self.data_raw
        if self.sim_sea is not None:
            data = [data, self.sim_sea]
        if title is None:
            title_ = "Autocorrelation of untransformed variables"
        figs += [ts.plot_auto_corr(data, maxlag, title_, var_names, *args,
                                   **kwds)]
        figs += [ts.plot_auto_corr(self.residuals, maxlag,
                                   "Autocorrelation of the residuals.",
                                   var_names, **kwds)]
        # otherwise we probably have been given two arrays.  then
        # things get more complicated, because we would distinguish
        # between two possibly different sample sizes (data.shape[1])
        if type(data) is np.ndarray:
            # bounds for test on the whiteness of the residuals (see page 160)
            bound = 2. / np.sqrt(data.shape[1]) * np.ones(maxlag)
            plt.plot(range(maxlag), bound, "--", color="grey")
            plt.plot(range(maxlag), -bound, "--", color="grey")

        return figs

    def plot_cross_corr(self, max_lags=7, figsize=None):
        """Plot cross correlations."""
        kwds = dict(var_names=self.var_names, max_lags=max_lags)
        fig, axs = ts.plot_cross_corr(self.data_raw / self.sum_interval,
                                      figsize=figsize, **kwds)
        if self.sim_sea is not None:
            fig, axs = ts.plot_cross_corr(self.sim_sea, linestyle="--",
                                          fig=fig, axs=axs, **kwds)
        plt.suptitle("Daily")

        if self.sim_sea_dis is not None:
            met_array = vg_base.met_as_array(self.met,
                                             var_names=self.var_names)
            kwds["max_lags"] *= max(self.sum_interval)
            fig1, axs1 = ts.plot_cross_corr(met_array, figsize=figsize, **kwds)
            fig1, axs1 = ts.plot_cross_corr(self.sim_sea_dis, linestyle="--",
                                            fig=fig1, axs=axs1, **kwds)
            plt.suptitle("Hourly")
            fig = [fig, fig1]
            axs = [axs, axs1]

        return fig, axs

    def plot_scaling(self, agg_funcs=(np.mean, np.std, stats.skew,
                                      stats.kurtosis),
                     max_agg_len=2 * 365, normalize=False, **fig_kw):
        """Plot mean/std/skew over aggregation length."""
        kwds = dict(var_names=self.var_names, agg_funcs=agg_funcs,
                    max_agg_len=max_agg_len, **fig_kw)
        fig, axs = ts.plot_scaling(self.data_raw / self.sum_interval,
                                   color="black", **kwds)

        if self.sim_sea is not None:
            sim = self.sim_sea
            if normalize:
                # for the sake of argument: make the means equal
                obs_means = np.nanmean(self.data_raw / self.sum_interval,
                                       axis=1)
                sim_means = np.nanmean(self.sim_sea, axis=1)
                sim += (obs_means - sim_means)[:, None]
            fig, axs = ts.plot_scaling(sim, color="blue", fig=fig, axs=axs,
                                       **kwds)
        return fig, axs

    def plot_monthly_hists(self, var_names=None, bins=20):
        """Plot histograms grouped by month and variable."""
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, basestring):
            var_names = var_names,

        figs = []
        data = self.data_raw / self.sum_interval
        month_ii = [times.time_part(self.times, "%m") == month
                    for month in range(1, 13)]
        if self.sim_sea is not None:
            sim_month_ii = [times.time_part(self.sim_times, "%m") == month
                            for month in range(1, 13)]
        for var_name in var_names:
            var_i = var_names.index(var_name)
            fig, axs = plt.subplots(3, 4)
            axs = axs.ravel()
            for month in range(12):
                dat = data[var_i, month_ii[month]]
                if self.sim_sea is not None:
                    dat = [dat, self.sim_sea[var_i, sim_month_ii[month]]]
                axs[month].hist(dat, bins, normed=True)
                axs[month].set_title(month + 1)
            fig.suptitle(var_name)
            figs += [fig]
        if len(figs) == 1:
            figs = figs[0]
        return figs

    def plot_episode_hists(self, var_names=None):
        """Plots histograms of episode duration and amplitude grouped by
        variable."""
        if var_names is None:
            var_names = self.var_names
        elif isinstance(var_names, basestring):
            var_names = var_names,

        figs = []
        for var_i, var_name in enumerate(var_names):
            medians = self.fitted_medians(var_name, self.data_doys)
            devs_obs = \
                self.data_raw[var_i] / self.sum_interval[var_i] - medians
            sign_changes = np.where(np.diff(np.sign(devs_obs)) != 0)[0]
            durations = [np.diff(sign_changes)]
            amplitudes = [np.array([np.mean(epi) for epi in
                                    np.split(devs_obs, sign_changes)
                                    if len(epi) > 0])]

            if self.sim_sea is not None:
                medians = self.fitted_medians(var_name, self.sim_doys)
                devs_sim = self.sim_sea[var_i] - medians
                sign_changes = np.where(np.diff(np.sign(devs_sim)) != 0)[0]
                durations_sim = np.diff(sign_changes)
                amplitudes_sim = [np.mean(epi) for epi in
                                  np.split(devs_sim, sign_changes)
                                  if len(epi) > 0]
                durations += [durations_sim]
                amplitudes += [amplitudes_sim]

            fig, axes = plt.subplots(2, 1)
            axes[0].hist(durations, 30, normed=True)
            axes[0].set_xlabel("Duration [days]")
            axes[1].hist(amplitudes, 30, normed=True)
            axes[1].set_xlabel("Deviation %s" % conf.units[var_name])

            legend1 = (r"observed $\overline{x}=%.2f$" %
                       np.mean(durations[0] - 1))
            legend2 = (r"observed $\sigma=%.2f$" %
                       np.std(amplitudes[0][np.isfinite(amplitudes[0])]))
            if self.sim_sea is not None:
                legend1 = [legend1, r"sim $\overline{x}=%.2f$" %
                           np.mean(durations_sim - 1)]
                legend2 = [legend2, r"sim $\sigma=%.2f$" %
                           np.std(amplitudes_sim)]
            else:
                legend1, legend2 = (legend1,), (legend2,)
            axes[0].legend(legend1, loc="best")
            axes[1].legend(legend2, loc="best")

            title = "Episode statistics %s" % var_name
            fig.suptitle(title)
            fig.canvas.set_window_title(title)
            figs += [fig]
        return figs

    def plot_daily_cycles(self, var_names=None, rain_thresh=None, **fig_kw):
        """Climbing Mt. Allon.

        Parameter
        ---------
        var_names : sequence of str, optional
            Which variables to plot. None means plot all.
        rain_thresh : None, True or float, optional
            Which elemts to include based on rain threshold. If True, it is
            inferred from the configuration file (dist_kwds["R"]["threshold"])
        """
        if self.sim_sea_dis is None:
            print "Call disaggregate first"
            return
        if var_names is None:
            var_names = self.var_names
        elif type(var_names) == str:
            var_names = var_names,

        if not rain_thresh:
            # rain_mask_obs = np.ones_like(self.met["R"], dtype=bool)
            # rain_mask_sim = np.ones(self.sim_sea_dis.shape[1], dtype=bool)
            rain_comparison = False
        if rain_thresh is True or type(rain_thresh) is float:
            if rain_thresh is True:
                rain_thresh = conf.dists_kwds["R"]["threshold"]
            rain_mask_obs = (self.data_raw[self.var_names.index("R")] >=
                             rain_thresh)
            rain_mask_obs = np.repeat(rain_mask_obs,
                                      self.sum_interval_dict["R"][0])
            rain_mask_sim = (self.sim_sea[self.var_names.index("R")] >=
                             (rain_thresh / self.sum_interval_dict["R"][0]))
            rain_mask_sim = np.repeat(rain_mask_sim,
                                      self.sum_interval_dict["R"][0])
            rain_comparison = True

        figs, axes = [], []
        month_iis = [times.time_part(self.times_orig, "%m") == month
                     for month in range(1, 13)]
        hour_iis = [times.time_part(self.times_orig, "%H") == hour
                    for hour in range(24)]
        month_sim_iis = [times.time_part(self.dis_times, "%m") == month
                         for month in range(1, 13)]
        hour_sim_iis = [times.time_part(self.dis_times, "%H") == hour
                        for hour in range(24)]

        for var_name in var_names:
            if rain_comparison:
                fig, axs = plt.subplots(2, 2, **fig_kw)
            else:
                fig, axs = plt.subplots(1, 2, **fig_kw)
                axs = np.array((axs,))

            var_i_dis = self.var_names.index(var_name)

            for mask_i, mask_mod in enumerate((lambda x: x, lambda x: ~x)):
                if not rain_comparison and mask_i == 1:
                    break

                means_obs = np.zeros((12, 24))
                means_sim = np.zeros_like(means_obs)
                for month in range(1, 13):
                    for hour in range(24):
                        month_ii = month_iis[month - 1]
                        hour_ii = hour_iis[hour]
                        mask = month_ii & hour_ii
                        if rain_comparison:
                            mask &= mask_mod(rain_mask_obs)
                        data_obs = self.met[var_name][mask]
                        month_ii = month_sim_iis[month - 1]
                        hour_ii = hour_sim_iis[hour]
                        data_sim = self.sim_sea_dis[var_i_dis]
                        mask = month_ii & hour_ii
                        if rain_comparison:
                            mask &= mask_mod(rain_mask_sim)
                        data_sim = data_sim[mask[:len(data_sim)]]
                        means_obs[month - 1, hour] = my.nanavg(data_obs)
                        means_sim[month - 1, hour] = my.nanavg(data_sim)

                # the following code tries to achieve that the colorbar for the
                # two contour plots spans from vmin to vmax (this would be the
                # case for calls to imshow instead of contourf).
                # see
                # http://matplotlib.1069221.n5.nabble.com/
                # question-about-contours-and-clim-td21111.html
                # and
                # https://github.com/matplotlib/matplotlib/pull/2176
                # to understand why this is so annoying
                vmin = min(means_obs.min(), means_sim.min())
                vmax = max(means_obs.max(), means_sim.max())
                locator = ticker.MaxNLocator(8)
                locator.create_dummy_axis()
                locator.set_bounds(vmin, vmax)
                levs = locator()
                axs[mask_i, 0].contourf(range(24), range(1, 13), means_obs,
                                        levs)
                co = axs[mask_i, 1].contourf(range(24), range(1, 13),
                                             means_sim, levs)

                axs[mask_i, 0].set_title("Observed")
                axs[mask_i, 0].set_xlabel("hour")
                axs[mask_i, 0].set_ylabel("month")
                axs[mask_i, 0].grid()
                axs[mask_i, 1].set_title("Simulated")
                axs[mask_i, 1].set_xlabel("hour")
                axs[mask_i, 1].grid()

            title = "Seasonal daily cycles for %s" % conf.ygreek[var_name]
            fig.suptitle(title)
            fig.canvas.set_window_title(title)
            fig.tight_layout()
            fig.subplots_adjust(right=0.8)
            cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
            cbar = fig.colorbar(co, cax=cbar_ax)
            cbar.set_label(conf.units[var_name])

            figs += [fig]
            axes += [axs]
        return figs, axes

    # def plot_torus(self, var_names=None, **fig_kw):
    #     """Provides a 2d visualization with doys as horizontal and hours as
    #     vertical axis.

    #     Parameter
    #     ---------
    #     var_names : sequence of str, optional
    #         Which variables to plot. None means plot all.
    #     """
    #     if var_names is None:
    #         var_names = self.var_names_dis
    #     elif type(var_names) == str:
    #         var_names = var_names,

    #     figs, axes = [], []
    #     n_years = self.times_orig[-1].year - self.times_orig[0].year + 1
    #     nrows = int(np.ceil(float(n_years) / 10))
    #     ncols = n_years / nrows
    #     for var_name in var_names:
    #         fig, axs = plt.subplots(nrows, ncols, **fig_kw)
    #         # torus = 
            
    #         figs += [fig]
    #         axes += [axs]
