'''
Helper functions for working with DYRESM files.
'''
import numpy as np
import matplotlib.pyplot as plt
from scipy import interp, io
from vg import times


def copfill(nc_var):
    """Fill missing values in a netcdf variable with np.nans and return a copy
    of the whole data (so that the netcdf file can be closed)."""
    values = nc_var[:]
    if hasattr(nc_var, "_FillValue"):
        mask = np.isclose(values, nc_var._FillValue)
        return np.copy(np.where(mask, np.nan, values))
    try:
        # we might have strings in values
        return np.array(["".join(name).strip() for name in values])
    except TypeError:
        return np.copy(values)


# as of scipy '0.13.0.dev-38faf7c' (possibly a bit earlier),
# scipy.io.netcdf_file has a context manager
# here we assure we can use it with a with statement also for older scipies
if hasattr(io.netcdf_file, "__exit__"):
    netcdf_file = io.netcdf_file
else:
    from contextlib import closing
    netcdf_file = lambda x, *arg, **kw: closing(io.netcdf_file(x, *arg, **kw))


class Dyresm(object):
    _layer_str = "dyresmLAYER_HTS_Var"
    _time_str = "dyresmTime"
    _temperature_str = "dyresmTEMPTURE_Var"

    def __init__(self, nc_file, depth=None):
        """A wrapper for netCDF files produced by DYRESM that helps to
        interpolate DYRESM output to specified depths below sea level.

        netCDF variables are accessible by attributes. However, no retrieval
        or interpolation is done, before the attribute is executed (by
        prepending "()" to the variable name).
        Attributes for variables that do not have the shape of
        "dyresmLAYER_HTS_Var", are prepended with "_" for easier access to the
        "important" variables.

        Parameters
        ----------
        nc_file : str
            Filename of the netCDF file.
        depth : float, int, None or 1d ndarray, optional
            Depth or depths to interpolate all variables to. If None, no
            interpolation will be done.

        Examples
        --------
        (don't doctest this, the nc-file isn't there)
        >>> dy = Dyresm("DYsim_base_with_biology.nc")
        Set depths to interpolate to (depth can also be a scalar):
        >>> dy.depth = np.linspace(0, 10, 100)
        Call (!) an attribute to get the interpolated values:
        >>> dy.dyresmTEMPTURE_Var()
        array([[ 17.69033241,  17.69033241,  17.69033241, ...,  17.70616715,
                 17.70616604,  17.70616524],
               [ 17.65152931,  17.65152931,  17.65152931, ...,  17.67119485,
                 17.67119422,  17.6711936 ],
               [ 17.58423996,  17.58423996,  17.58423996, ...,  17.58423996,
                 17.58423996,  17.58423996],
               ...,
               [ 16.54315567,  16.54315567,  16.54315567, ...,  16.54315567,
                 16.54315567,  16.54315567],
               [ 16.49787521,  16.49787521,  16.49787521, ...,  16.49787521,
                 16.49787521,  16.49787521],
               [ 16.44964218,  16.44964218,  16.44964218, ...,  16.44964218,
                 16.44964218,  16.44964218]])
        Nc variable attributes are preserved:
        >>> dy.dyresmTEMPTURE_Var.long_name, dy.dyresmTEMPTURE_Var.units
        ('Temperature', 'Celsius')
        """
        self.nc_file = nc_file
        self.depth = depth
        self.var_names_all = None
        # only the ones that have the right shape
        self.var_names = None
        # eposes variable names and values in the same fashion as usual for
        # nc files
        self.variables = {}

        # scipy.io.netcdf keeps an mmap descriptor per nc variable open,
        # and we quickly reach the limit of how many files are allowed to
        # be opened by the os.
        # so here we just extract some variable information from the nc file
        # and close it again.
        with netcdf_file(nc_file) as nc:
            # expose variable names for outer pleasure
            self.var_names_all = nc.variables.keys()
            nc_time = nc.variables[self._time_str][:].copy()
            try:
                self.dtimes = times.dy2datetime(nc_time)
            except OverflowError:
                print "Could not parse time stamps."

            layer_heights = copfill(nc.variables[self._layer_str])
            self.layer_depths = \
                np.abs(layer_heights -
                       np.nanmax(layer_heights, axis=1)[:, None])
            self.shape = layer_heights.shape

            for var_name in nc.variables.keys():
                if nc.variables[var_name].shape == self.shape:
                    # this is a variable that can be interpolatet to a given
                    # depth
                    set_func = self._var_depth_gen
                    attr_name = var_name
                else:
                    # this is something else
                    set_func = self._var_gen
                    attr_name = "_" + var_name
                # exposes the dyresm variables as functions that later
                # return values when called
                # the fact that the functions are attributes instead of
                # keys in a dictionary makes this more useful in
                # interactive sessions (TAB-key is happy)
                setattr(self, attr_name, set_func(var_name))
                # duplicate this functionality for the usual nc api
                self.variables[attr_name] = getattr(self, attr_name)
                # dyresm variables come with informative attributes like
                # "description", "long_name" and so on. expose them also.
                # this is basically a shallow copy without the actual
                # data, which will be accessed via the _var_depth method
                var_dict = {key: val for key, val in
                            nc.variables[var_name].__dict__.items()
                            if key != "data"}
                getattr(self, attr_name).__dict__.update(var_dict)
                self.variables[attr_name].__dict__.update(var_dict)

            self.var_names = self.variables.keys()

    def _var_depth(self, var_name, depth=True):
        with netcdf_file(self.nc_file) as nc:
            var_ = copfill(nc.variables[var_name])

        if not depth or self.depth is None:
            # returns the whole array. no interpolation or regularization
            return var_
        else:
            # returns the variable interpolated at a certain depth.
            # this also works if self.depth is a sequence!
            return np.array(
                    [interp(self.depth,
                            layer_depth[np.isfinite(layer_depth)],
                            var_sub[np.isfinite(var_sub)])
                     for layer_depth, var_sub in zip(self.layer_depths, var_)])

    def _var_depth_gen(self, var_name):
        return lambda: self._var_depth(var_name)

    def _var_gen(self, var_name):
        return lambda: self._var_depth(var_name, depth=False)

    def plot_contour(self, var_name=None):
        """Quick and dirty contour plot.

        Parameters
        ----------
        var_name : str or None, optional
            nc variable name, will be substituted with water temperature name
            if None.
        """
        if var_name is None:
            var_name = self._temperature_str
        max_depth = np.nanmax(self.layer_depths)
        depth_old = self.depth
        zz = np.linspace(0, max_depth, 100)
        self.depth = zz
        data = getattr(self, var_name)().T

        fig = plt.figure()
        plt.contourf(self.dtimes, zz[::-1], data)
        plt.colorbar()

        self.depth = depth_old
        return fig


class MultiDyresm(object):
    def __init__(self, nc_filenames, depth=None):
        """A wrapper for Dyresm objects. MultiDyresm has the same api as
        Dyresm, but returns arrays that have the additional first dimension
        "number of nc_filenames".

        Parameters
        ----------
        nc_files : sequence of str
            Filenames of netCDF files.
        depth : float, int, None or 1d ndarray, optional
            Depth or depths to interpolate all variables to. If None, no
            interpolation will be done.
        """
        self.nc_files = [Dyresm(nc_filename, depth)
                         for nc_filename in nc_filenames]
        self.depth = depth
        self.var_names = self.nc_files[0].var_names
        self.shape = tuple([len(self.nc_files)] + list(self.nc_files[0].shape))
        self.dtimes = self.nc_files[0].dtimes
        self.variables = {}

        for var_name in self.var_names:
            # slightly adapted from Dysresm.__init__
            setattr(self, var_name, self._var_depth_gen(var_name))
            self.variables[var_name] = getattr(self, var_name)
            var_dict = {key: val for key, val in
                        self.nc_files[0].variables[var_name].__dict__.items()
                        if key != "data"}
            getattr(self, var_name).__dict__.update(var_dict)
            self.variables[var_name].__dict__.update(var_dict)

    @property
    def depth(self):
        return self._depth

    @depth.setter
    def depth(self, value):
        self._depth = value
        for nc_file in self.nc_files:
            nc_file.depth = value

    def _var_depth(self, var_name):
        return np.array([nc_file.variables[var_name]()
                         for nc_file in self.nc_files])

    def _var_depth_gen(self, var_name):
        return lambda: self._var_depth(var_name)


if __name__ == '__main__':
    pass
