import warnings
import os
from abc import ABCMeta, abstractmethod
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats, special
import scipy.optimize as sp_optimize

try:
    import numexpr as ne
    NE = True
except ImportError:
    NE = False

from vg import helpers as my

try:
    import owens
except ImportError:
    import urllib2
    from numpy import f2py
    warn = False
    url = "http://people.sc.fsu.edu/~jburkardt/f_src/owens/owens.f90"
    src_dir = os.path.dirname(__file__)
    with my.chdir(src_dir):
        with open("owens.f90", "w") as owens_file:
            try:
                owens_file.write(urllib2.urlopen(url).read())
            except urllib2.URLError:
                warn = True
        # hack to let f2py think it was called from the command-line
        f2py.sys.argv += ["-c", "-m", "owens", "owens.f90"]
        try:
            f2py.main()
            import owens
        except:
            warn = True
    if warn:
        warnings.warn("""Could not import owens t function.
            Try (if on linux):
            wget {url}
            f2py -c -m owens owens.f90
            cp owens.so {src_dir}/
            """.format(url=url, src_dir=src_dir))
        owens = False
if owens:
    owens_t = np.vectorize(lambda h, a: owens.t(h, a))


# some special functions vectorized to be able to handle arrays as input
gamma_func = np.vectorize(lambda x: special.gamma(x))
gammaln = np.vectorize(lambda x: special.gammaln(x))
gammainc = np.vectorize(lambda a, x: special.gammainc(a, x))
gammaincinv = np.vectorize(lambda a, q: special.gammaincinv(a, q))
digamma = np.vectorize(lambda a: special.digamma(a))
hyp1f1 = np.vectorize(lambda a, b, x: special.hyp1f1(a, b, x))
nctdtr = np.vectorize(lambda x1, x2, x3: special.nctdtr(x1, x2, x3))
nctdtrit = np.vectorize(lambda x1, x2, x3: special.nctdtrit(x1, x2, x3))
stdtr = np.vectorize(lambda df, x: special.stdtr(df, x))
stdtrit = np.vectorize(lambda df, q: special.stdtrit(df, q))


def max_likelihood(density_func, x0, values=None, opt_func=None, disp=False,
                   weights=None, method="BFGS", *args, **kwds):
    """Fits parameters of a density function to data, using the log-maximum-
    likelihood approach."""
    if opt_func is None:
        from scipy import optimize
        opt_func = optimize.minimize
    if weights is None and values is not None:
        weights = np.ones_like(values)
    # i hereby define the unlikelihood function as the negative log-likelihood
    # function, so that i can maximize the likelihood by minimizing the
    # unlikelihood
    if values is None:
        args = ()
        unlikelihood = lambda params: \
            -np.sum(np.log(weights * density_func(params) + 1e-12))
    else:
        args = (values,)

        def unlikelihood(params, values):
            # we have to express fixed values in terms of kwds, because there
            # could be collisions otherwise...

            # first come, first serve: interpret the first n elements of params
            # as the first n variables in dist.parameter_names
            p_kwds = {par_name: value for par_name, value in
                      zip(density_func.im_self.parameter_names, params)
                      if par_name not in kwds}
            p_kwds.update(kwds)
            inside = weights * density_func(values, **p_kwds)
            inside[(inside <= 0) | np.isnan(inside)] = 1e-12
            return -np.sum(np.log(inside))
    return opt_func(unlikelihood, x0, args=args, method=method,
                    options={"maxiter": 1e4 * len(x0),
                             "disp": disp})


def min_ks(cdf, x0, values, opt_func=None, disp=False, *args, **kwds):
    """Fits parameters of a cumulative distribution function to data,
    minimizing the Kolmogorov-Smirnof test statistic."""
    if opt_func is None:
        opt_func = sp_optimize.fmin
    n = len(values)
    values_sorted = np.sort(values)
    ranks_plus = np.arange(0., n) / n
    ranks_minus = np.arange(1., n + 1) / n

    def ks(params):
        # we have to express fixed values in terms of kwds, because there
        # could be collisions otherwise...

        # first come, first serve: interpret the first n elements of params
        # as the first n variables in dist.parameter_names
        p_kwds = {par_name: value for par_name, value in
                  zip(cdf.im_self.parameter_names, params)
                  if par_name not in kwds}
        p_kwds.update(kwds)

        cdf_values = cdf(values_sorted, **p_kwds)
        cdf_values[np.isnan(cdf_values)] = np.inf
        dmin_plus = np.abs(cdf_values - ranks_plus).max()
        dmin_minus = np.abs(cdf_values - ranks_minus).max()
        return max(dmin_plus, dmin_minus)

    return opt_func(ks, x0, args=args, maxiter=1e4 * len(x0),
                    maxfun=1e4 * len(x0), disp=disp)


def min_fsum(cdf, x0, values, opt_func=None, disp=False, *args, **kwds):
    if opt_func is None:
        opt_func = sp_optimize.fmin
    n = len(values)
    values_sorted = np.sort(values)
    ranks_plus = np.arange(0., n) / n
    ranks_minus = np.arange(1., n + 1) / n

    def ks(params):
        # we have to express fixed values in terms of kwds, because there
        # could be collisions otherwise...

        # first come, first serve: interpret the first n elements of params
        # as the first n variables in dist.parameter_names
        p_kwds = {par_name: value for par_name, value in
                  zip(cdf.im_self.parameter_names, params)
                  if par_name not in kwds}
        p_kwds.update(kwds)

        cdf_values = cdf(values_sorted, **p_kwds)
        cdf_values[np.isnan(cdf_values)] = np.inf
        dmin_plus = ((cdf_values - ranks_plus) ** 2).sum()
        dmin_minus = ((cdf_values - ranks_minus) ** 2).sum()
        return dmin_plus + dmin_minus

    return opt_func(ks, x0, args=args, maxiter=1e4 * len(x0),
                    maxfun=1e4 * len(x0), disp=disp)


class DistMeta(ABCMeta):
    """Assure some class attributes are set, so that life is easier later on.
    """
    # pylint does not let me call the first parameter meta :(
    def __new__(cls, name, bases, cls_dict):
        new_cls = super(DistMeta, cls).__new__(cls, name, bases, cls_dict)
        new_cls.name = name.lower()
        # store the distribution parameter names as a class attribute
        # take the pdf as an example
        # could do some signature checks on cdf and ppf with that...
        code = cls_dict["_pdf"].func_code
        # first varname is self, second is x
        new_cls.parameter_names = code.co_varnames[2: code.co_argcount]
        return new_cls


class MyDist(object):
    """Mimics part of the interface of stats.rv_continuous. Comes with a few
    extra goodies."""
    __metaclass__ = DistMeta

    @abstractmethod
    def _pdf(self):
        pass

    @abstractmethod
    def _cdf(self):
        pass

    @abstractmethod
    def _ppf(self):
        pass

    @abstractmethod
    def _fit(self):
        pass

    @property
    def scipy_(self):
        return False

    def _clean_kwds(self, kwds):
        """Return a copy with only the keywords that are also in
        self.parameter_names."""
        return {key: value for key, value in kwds.items()
                if key in self.parameter_names}

    def fit(self, *args, **kwds):
        return self._fit(*args, **self._clean_kwds(kwds))

    @my.asscalar
    def pdf(self, *args, **kwds):
        densities = np.atleast_1d(self._pdf(*args, **self._clean_kwds(kwds)))
        if "x" in kwds:
            x = kwds.pop("x")
            invalid_x = self._invalid_x(x, *args, **kwds)
        else:
            invalid_x = self._invalid_x(args[0], *args[1:], **kwds)
        densities[invalid_x] = np.nan
        return densities

    @my.asscalar
    def cdf(self, *args, **kwds):
        quantiles = np.atleast_1d(self._cdf(*args, **self._clean_kwds(kwds)))
        if "x" in kwds:
            x = kwds.pop("x")
            invalid_x = self._invalid_x(x, *args, **kwds)
        else:
            invalid_x = self._invalid_x(args[0], *args[1:], **kwds)
        quantiles[invalid_x] = np.nan
        return quantiles

    @my.asscalar
    def ppf(self, *args, **kwds):
        quantiles = np.atleast_1d(args[0])
        x = np.atleast_1d(self._ppf(*args, **self._clean_kwds(kwds)))
        x[(quantiles < 0) | (quantiles > 1)] = np.nan
        return x

    def __call__(self, *params):
        return Frozen(self, *params)

    def fit_ml(self, values, x0=None, *args, **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return max_likelihood(self.pdf, x0, values, *args, **kwds)

    def fit_ks(self, values, opt_func=sp_optimize.fmin, x0=None, *args,
               **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return min_ks(self.cdf, x0, values, opt_func, *args, **kwds)

    def fit_fsum(self, values, opt_func=sp_optimize.fmin, x0=None, *args,
                 **kwds):
        if x0 is None:
            x0 = (1,) * len(self.parameter_names)
        return min_fsum(self.cdf, x0, values, opt_func, *args, **kwds)

    def _constraints(self, *args, **kwds):
        return True

    def _invalid_x(self, x, *args, **kwds):
        """Returns a mask indicating where x < lower bound or x > upper bound.
        """
        x = np.atleast_1d(x)
        mask = np.atleast_1d(np.full_like(x, False, dtype=bool))
        lower_key = ("lc" if "lc" in self.parameter_names else
                     "l" if "l" in self.parameter_names else
                     None)
        if lower_key:
            lower = kwds.get(lower_key)
            if lower is None:
                lower = args[self.parameter_names.index(lower_key)]
            mask[x < lower] = True

        upper_key = ("uc" if "uc" in self.parameter_names else
                     "u" if "u" in self.parameter_names else
                     None)
        if upper_key:
            upper = kwds.get(upper_key)
            if upper is None:
                upper = args[self.parameter_names.index(upper_key)]
            mask[x > upper] = True

        return mask


class Frozen(object):
    # this idea is mainly stolen from the scipy.stats-module
    def __init__(self, dist, *params):
        self.dist = dist
        self.params = params
        self.parameter_names = dist.parameter_names
        self.name = dist.name

    def pdf(self, x):
        return self.dist.pdf(x, *self.params)

    def cdf(self, y):
        return self.dist.cdf(y, *self.params)

    def ppf(self, x):
        return self.dist.ppf(x, *self.params)

    @property
    def parameter_dict(self):
        return {name: value
                for name, value in zip(self.parameter_names, self.params)}

    def plot_fit(self, values, n_classes=30):
        """Display a combined plot of a histogram, fitted pdf, empirical cdf
        and fitted cdf."""
        fig = plt.figure()
        ax1 = fig.add_subplot(111)

        # the histogram of the data
        bins = ax1.hist(values, n_classes, normed=True, facecolor='green',
                        alpha=0.75)[1]

        class_middles = 0.5 * (bins[1:] + bins[:-1])
        density = self.pdf(class_middles)
        ax1.plot(class_middles, density, 'r--')

        # the quantile part
        ax2 = ax1.twinx()
        # empirical cdf
        values_sort = np.sort(values)
        ranks_emp = (.5 + np.arange(len(values))) / len(values)
        ax2.plot(values_sort, ranks_emp)
        # theoretical cdf
        xx = np.linspace(values.min(), values.max(), 5e2)
        ranks_theory = self.cdf(xx)
        ax2.plot(xx, ranks_theory, 'r--')
        ax2.grid()
        plt.title(", ".join(" %s: %.3f" % (par_name, par)
                            for par_name, par in self.parameter_dict.items()))
        return fig

    def plot_qq(self, values, *args, **kwds):
        """A qq-plot. Scatters theoretic over empirical ranks."""
        ranks_emp = (stats.stats.rankdata(values) - .5) / len(values)
        ranks_the = self.cdf(values)
        fig = plt.figure()
        plt.scatter(ranks_emp, ranks_the, marker="o", *args, **kwds)
        plt.plot([0, 1], [0, 1])
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.xlabel("Empirical ranks")
        plt.ylabel("Fitted ranks")
        plt.grid()
        return fig


class Censored(MyDist):
    def __init__(self, distribution, lc=-np.inf, uc=np.inf):
        self.dist = distribution
        self.lc = lc
        self.uc = uc
        # read "lc" as "lower cut", "uc" as "upper cut"
        # mind you: these are different than lower and upper bound in i.e. the
        # beta distribution. we can also truncate the beta distribution and
        # have upper bounds above the upper cut.
        if hasattr(distribution, "parameter_names"):
            self.parameter_names = tuple(list(distribution.parameter_names) +
                                         ["lc", "uc"])
        self.name = "censored " + distribution.name

    def _pdf(self, x, *args, **kwds):
        x = np.atleast_1d(x)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds.pop("uc")
        density = self.dist.pdf(x, *args, **kwds)
        quantiles = self.dist.cdf(x, *args, **kwds)
        lc_above_ii = np.atleast_1d(x > lc)
        if np.any(lc_above_ii):
            try:
                lc_index = np.argmin(x[lc_above_ii])
                density[lc_index] += quantiles[lc_index]
            except IndexError:
                for jj, lc_above in enumerate(lc_above_ii):
                    lc_index = np.argmin(x[0, lc_above])
                    density[jj, lc_index] += quantiles[jj, lc_index]
        uc_below_ii = np.atleast_1d(x < uc)
        if np.any(uc_below_ii):
            try:
                uc_index = np.argmax(x[uc_below_ii])
                density[uc_index] += 1 - quantiles[uc_index]
            except IndexError:
                for jj, uc_below in enumerate(uc_below_ii):
                    uc_index = np.argmax(x[0, uc_below])
                    density[jj, uc_index] += 1 - quantiles[jj, uc_index]
        density[(x < lc) | (x > uc)] = 0
        return density

    def _cdf(self, x, *args, **kwds):
        x = np.asarray(x)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds.pop("uc")
        quantiles = np.atleast_1d(self.dist.cdf(x, *args, **kwds))
        below_lc_ii = np.atleast_1d(x <= lc)
        if np.any(below_lc_ii):
            lc_quantile = self.dist.cdf(lc, *args, **kwds)
            if np.isscalar(lc_quantile):
                lc_quantile = np.full_like(x, lc_quantile)
            quantiles[below_lc_ii] = lc_quantile[below_lc_ii]
        above_uc_ii = np.atleast_1d(x >= uc)
        if np.any(above_uc_ii):
            uc_quantile = self.dist.cdf(uc, *args, **kwds)
            if np.isscalar(uc_quantile):
                uc_quantile = np.full_like(x, uc_quantile)
            quantiles[above_uc_ii] = 1 - uc_quantile[above_uc_ii]
        quantiles[quantiles < 0] = 0
        quantiles[quantiles > 1] = 1
        return np.squeeze(quantiles)

    def _ppf(self, quantiles, *args, **kwds):
        quantiles = np.asarray(quantiles)
        args = [np.asarray(arg) for arg in args]
        lc = self.lc if "lc" not in kwds else kwds["lc"]  # kwds.pop("lc")
        uc = self.uc if "uc" not in kwds else kwds["uc"]  # kwds.pop("uc")
        x = self.dist.ppf(quantiles, *args, **kwds)
        lower_ii = x < lc
        if np.any(lower_ii):
            x[lower_ii] = (np.full_like(x, lc) if np.isscalar(lc)
                           else lc
                           )[lower_ii]
        upper_ii = x > uc
        if np.any(upper_ii):
            x[upper_ii] = (np.full_like(x, uc) if np.isscalar(uc)
                           else uc
                           )[upper_ii]
        return x

    def _fit(self, x, *args, **kwds):
        # try:
        #     x0 = self.dist._feasible_start
        # except AttributeError:
        #     x0 = self.dist.fit(x, *args, **kwds)
        # using the cdf to fit, because of the insanity inherent in the pdf
        # (i.e. no maximum likelihood)
        # return min_ks(self.cdf, x0, x, lc=self.lc, uc=self.uc, *args)
        # fit on valid x
        valid_x = x[~self._invalid_x(x, *args, **kwds)]
        return self.dist.fit(valid_x, *args,
                             **{key: val for key, val in kwds.items()
                                if key != "lc" and key != "uc"})

    def constraints(self, x, *args, **kwds):
        return self.dist.constraints(x, *args, **kwds)


class Truncated(MyDist):
    """Wraps around any other distribution found here to truncate its upper
    and/or lower tail."""
    def __init__(self, distribution, lc=-np.inf, uc=np.inf):
        self.dist = distribution
        self.lc = lc
        self.uc = uc
        # read "lc" as "lower cut", "uc" as "upper cut"
        # mind you: these are different than lower and upper bound in i.e. the
        # beta distribution. we can also truncate the beta distribution and
        # have upper bounds above the upper cut.
        self.parameter_names = tuple(list(distribution.parameter_names) +
                                     ["lc", "uc"])
        self.name = "truncated " + distribution.name

    def _pdf(self, x, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        # still sober
        un_trunc = (self.dist.pdf(x, *args, **kwds) /
                    (self.dist.cdf(uc, *args, **kwds) -
                     self.dist.cdf(lc, *args, **kwds)))
        un_trunc[(x < lc) | (x > uc)] = 1e-9
        # now totally drunkated
        return un_trunc

    def _cdf(self, x, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        quantiles = ((self.dist.cdf(x, *args, **kwds) -
                      self.dist.cdf(lc, *args, **kwds)) /
                     (self.dist.cdf(uc, *args, **kwds) -
                      self.dist.cdf(lc, *args, **kwds)))
        quantiles[quantiles < 0] = 0
        quantiles[quantiles > 1] = 1
        return quantiles

    def _ppf(self, quantiles, *args, **kwds):
        args = [np.asarray(arg) for arg in args]
        if "uc" in kwds:
            uc = kwds.pop("uc")
            if len(args) > 0:
                del args[-1]
        else:
            uc = self.uc
        if "lc" in kwds:
            lc = kwds.pop("lc")
            if len(args) > 0:
                del args[-1]
        else:
            lc = self.lc
        if len(args) > len(self.dist.parameter_names):
            args = args[:len(self.dist.parameter_names)]
        x = self.dist.ppf(quantiles * self.dist.cdf(uc, *args, **kwds) +
                          (1 - quantiles) * self.dist.cdf(lc, *args, **kwds),
                          *args, **kwds)
        lower_ii = x < lc
        if np.sum(lower_ii) > 0:
            x[lower_ii] = lc[lower_ii]
        upper_ii = x > uc
        if np.sum(upper_ii) > 1:
            x[upper_ii] = uc[upper_ii]
        return x

    def _fit(self, x, *args, **kwds):
        try:
            x0 = self.dist._feasible_start
        except AttributeError:
            x0 = self.dist.fit(x, *args)
        # using the cdf to fit, because of the insanity inherent in the pdf
        # (i.e. no maximum likelihood)
        if self.lc is None:
            x0 = tuple(list(x0) + [np.min(x)])
            kwds["lc"] = x0[-1]
        else:
            kwds["lc"] = self.lc
        if self.uc is None:
            x0 = tuple(list(x0) + [np.max(x)])
            kwds["uc"] = x0[-1]
        else:
            kwds["uc"] = self.uc
        solution = list(min_ks(self.cdf, x0, x, *args, **kwds))
        if self.uc is None:
            self.uc = solution.pop(-1)
        if self.lc is None:
            self.lc = solution.pop(-1)
        if "lc" in kwds:
            solution += [kwds["lc"]]
        if "uc" in kwds:
            solution += [kwds["uc"]]
        return solution

    def constraints(self, x, *args, **kwds):
        return self.dist.constraints(x, *args, **kwds)


class Rain(MyDist):
    def __init__(self, distribution, threshold=.0001):
        self.dist = distribution
        self.thresh = threshold
        self.parameter_names = tuple(["rain_prob"] +
                                     list(distribution.parameter_names))
        self.name = "rain " + distribution.name

    def _mask_kwds(self, mask, kwds):
        return {name: (vals[mask] if vals.shape == mask.shape else vals)
                for name, vals in kwds.items()}

    def _pdf(self, x, rain_prob, *args, **kwds):
        x, rain_prob = np.atleast_1d(x, rain_prob)
        rain_mask = x >= self.thresh
        if len(rain_prob) == 1:
            rain_prob = np.full_like(x, rain_prob)
        # the following tries to catch broadcast ambiguities, but might not
        # be the most parsimonious formulation in terms of memory
        dens = np.empty_like(rain_prob + rain_mask)
        dens[~rain_mask] = rain_prob[~rain_mask]
        kwds_rain = self._mask_kwds(rain_mask, kwds)
        dens_rain = self.dist.pdf(x[rain_mask] - self.thresh,
                                  *args, **kwds_rain)
        if dens.shape == rain_mask.shape:
            dens[rain_mask] = dens_rain
        else:
            dens[:, rain_mask[0]] = dens_rain
        return dens

    def _cdf(self, x, rain_prob, *args, **kwds):
        x, rain_prob = np.atleast_1d(x, rain_prob)
        qq = np.empty_like(x)
        if len(rain_prob) == 1:
            rain_prob = np.full_like(x, rain_prob)
        rain_mask = x >= self.thresh
        # subtracting self.thresh from x has the unwanted side-effect of
        # shifting x to possibly out-of-bound values. so this should not be
        # done with values where ~rain_mask.
        # as we mask x, we also have to mask the parameters given in **kwds
        kwds_rain = self._mask_kwds(rain_mask, kwds)
        qq[rain_mask] = (1 + rain_prob[rain_mask] *
                         (self.dist.cdf(x[rain_mask] - self.thresh,
                                        *args, **kwds_rain)
                          - 1))
        qq[~rain_mask] = 1 - rain_prob[~rain_mask]
        return qq

    def _ppf(self, q, rain_prob, *args, **kwds):
        q, rain_prob = np.atleast_1d(q, rain_prob)
        if len(rain_prob) == 1:
            rain_prob = np.full_like(q, rain_prob)
        # we want to have zero rain, where probability of rain is lower than
        # rain_prob
        x = np.zeros_like(q)
        rain_mask = q > 1 - rain_prob
        # subtracting self.thresh from x has the unwanted side-effect of
        # shifting x to possibly out-of-bound values. so this should not be
        # done with values where ~rain_mask.
        # as we mask x, we also have to mask the parameters given in **kwds
        kwds_rain = self._mask_kwds(rain_mask, kwds)
        x[rain_mask] = self.thresh + self.dist.ppf((q[rain_mask] - 1) /
                                                   rain_prob[rain_mask] + 1,
                                                   *args, **kwds_rain)
        return x

    def _fit(self, x, **kwds):
        rain_mask = x >= self.thresh
        rain_prob = np.mean(rain_mask)
        solution = self.dist.fit(x[rain_mask] - self.thresh, **kwds)
        return tuple([rain_prob] + list(solution))


class Normal(MyDist):
    _feasible_start = (0, 2)

    def _pdf(self, x, mu=0, sigma=1):
        x, mu, sigma = np.atleast_1d(x, mu, sigma)
        if NE:
            _pi = np.pi
            dens = ne.evaluate(
                "(1. / (2 * _pi * sigma ** 2) ** .5 *" +
                "exp(-(x - mu) ** 2 / (2 * sigma ** 2)))")
        else:
            dens = (1. / (2 * np.pi * sigma ** 2) ** .5 *
                    np.exp(-(x - mu) ** 2 / (2 * sigma ** 2)))
        return dens

    def _cdf(self, x, mu=0, sigma=1):
        x, mu, sigma = np.atleast_1d(x, mu, sigma)
        q = .5 * (1 + special.erf((x - mu) / (sigma * 2 ** .5)))
        return q

    def _ppf(self, quantiles, mu=0, sigma=1):
        quantiles, mu, sigma = np.atleast_1d(quantiles, mu, sigma)
        x = special.ndtri(quantiles) * sigma + mu
        return x

    def _fit(self, x, mu=None, sigma=None):
        x_noninf = x[np.isfinite(x)]
        if mu is None:
            mu = x_noninf.mean()
        if sigma is None:
            sigma = x_noninf.std()
        return mu, sigma
norm = Normal()


# class SkewNormal(MyDist):
#     """This is taken and adapted from
#     http://promanthan.com/randomstuff/skewt-0.0.1.tgz
#     which might end up in scipy.stats"""
#     @staticmethod
#     def _fui(h, i):
#         return (h ** (2 * i)) / ((2 ** i) * gamma(i + 1))
#
#     @staticmethod
#     def _tInt(h, a, jmax, cutPoint):
#         seriesL = np.empty(0)
#         seriesH = np.empty(0)
#         i = np.arange(0, jmax + 1)
#         low = h <= cutPoint
#         hL = h[low]
#         hH = h[np.logical_not(low)]
#         L = hL.size
#         if L > 0:
#             b = SkewNormal._fui(hL[:, np.newaxis], i)
#             cumb = b.cumsum(axis=1)  # transposed compared to R code
#             b1 = np.exp(-0.5 * hL ** 2)[:, np.newaxis] * cumb
#             matr = np.ones((jmax + 1, L)) - b1.transpose()
#             jk = ([1.0, -1.0] * jmax)[0:jmax + 1] / (2 * i + 1)
#             matr = np.inner((jk[:, np.newaxis] * matr).transpose(),
#                             a ** (2 * i + 1.0))
#             seriesL = (np.arctan(a) - matr.flatten(1)) / (2 * np.pi)
#         if hH.size > 0:
#             seriesH = (np.arctan(a) *
#                        np.exp(-0.5 * (hH ** 2.0) * a / np.arctan(a))
#                         * (1 + 0.00868 * (hH ** 4.0) * a ** 4.0) /
#                         (2.0 * np.pi))
#         series = np.empty(h.size)
#         series[low] = seriesL
#         series[np.logical_not(low)] = seriesH
#         return series
#
#     @staticmethod
#     def _tOwen(h, a, jmax=50, cutPoint=6):
#         aa = np.abs(a)
#         ah = np.abs(h)
#         if np.isnan(aa):
#             raise ValueError("a is NaN")
#         if np.isposinf(aa):
#             return 0.5 * norm().cdf(-ah)
#         if aa == 0.0:
#             return np.zeros(h.size)
#         na = np.isnan(h)
#         inf = np.isposinf(ah)
#         ah[np.logical_or(na, inf)] = 0
#         if aa <= 1:
#             owen = SkewNormal._tInt(ah, aa, jmax, cutPoint)
#         else:
#             owen = (0.5 * norm().cdf(ah) + norm().cdf(aa * ah)
#                     * (0.5 - norm().cdf(ah)) -
#                     SkewNormal._tInt(aa * ah, (1.0 / aa), jmax, cutPoint))
#         owen[np.isposinf(owen)] = 0
#         return owen * np.sign(a)
#
#     def _pdf(self, x, zeta, omega, alpha):
#         x, zeta, omega, alpha = np.atleast_1d(x, zeta, omega, alpha)
#         return (1 / (omega * np.pi) *
#                 np.exp(-(x - zeta) ** 2 / (2 * omega ** 2))
#                 * norm.cdf(alpha * (x - zeta) / omega))
#
#     def _cdf(self, x, zeta, omega, alpha):
#         x, zeta, omega, alpha = np.atleast_1d(x, zeta, omega, alpha)
#         return (norm.cdf((x - zeta) / omega) -
#                 2 * SkewNormal._tOwen((x - zeta) / omega, alpha))
#
#     def _ppfInternal(self, q, shape):
#         maxQ = np.sqrt(chi2.ppf(q, 1))
#         minQ = -np.sqrt(chi2.ppf(1 - q, 1))
#         if shape > 1e+5:
#             return maxQ
#         if shape < -1e+5:
#             return minQ
#         nan = np.isnan(q) | (q > 1) | (q < 0)
#         zero = q == 0
#         one = q == 1
#         q[nan | zero | one] = 0.5
#         cum = SkewNormal._cumulants(shape, 4)
#         g1 = cum[2] / cum[1] ** (3 / 2.0)
#         g2 = cum[3] / cum[1] ** 2
#         x = norm().ppf(q)
#         x = (x + (x ** 2 - 1) * g1 / 6 + x * (x ** 2 - 3) * g2 / 24 -
#              x * (2 * x ** 2 - 5) * g1 ** 2 / 36)
#         x = cum[0] + np.sqrt(cum[1]) * x
#         tol = 1e-8
#         maxErr = 1
#         while maxErr > tol:
#             sn = skewnorm(shape)
#             x1 = x - (sn.cdf(x) - q) / (sn.pdf(x))
#             x1 = np.minimum(x1, maxQ)
#             x1 = np.maximum(x1, minQ)
#             maxErr = np.amax(np.abs(x1 - x) / (1 + np.abs(x)))
#             x = x1
#         x[nan] = np.NaN
#         x[zero] = -np.Infinity
#         x[one] = np.Infinity
#         return x
#
#     def _ppf(self, q, shape):
#         if np.all(shape == shape[0]):
#             return self._ppfInternal(q, shape[0])
#         else:
#             vec = np.vectorize(lambda q, shape:
#                                 self._ppfInternal(np.array([q]), shape))
#             return vec(q, shape)
#
#     def _ppf(self, quantiles, zeta, omega, alpha):
#         quantiles, zeta, omega, alpha = \
#             np.atleast_1d(quantiles, zeta, omega, alpha)
#         return self._ppf(quantiles, alpha) * omega + zeta

class SkewNormal(MyDist):
    _feasible_start = (0, 1, .5)

    def _pdf(self, x, zeta=0, omega=1, alpha=0):
        x, zeta, omega, alpha = np.atleast_1d(x, zeta, omega, alpha)
        t = (x - zeta) / omega
        dens = 2. / omega * norm.pdf(t) * norm.cdf(t * alpha)
        return dens

    def _cdf(self, x, zeta=0, omega=1, alpha=0):
        x, zeta, omega, alpha = np.atleast_1d(x, zeta, omega, alpha)
        t = (x - zeta) / omega
        q = norm.cdf(t) - 2. * owens_t(t, alpha)
        return q

    def _ppf(self, quantiles, zeta=0, omega=1, alpha=0):
        quantiles, zeta, omega, alpha = \
            np.atleast_1d(quantiles, zeta, omega, alpha)
        if len(zeta) == 1:
            zeta = zeta * np.ones_like(quantiles)
        if len(omega) == 1:
            omega = omega * np.ones_like(quantiles)
        if len(alpha) == 1:
            alpha = alpha * np.ones_like(quantiles)

        xx0 = norm.cdf(quantiles) * omega + zeta
        xx = np.empty_like(quantiles)
        for i in xrange(len(xx)):
            quant = quantiles[i]
            error = lambda x: (self.cdf(x, zeta[i], omega[i], alpha[i]) -
                               quant) ** 2
            xx[i] = \
                sp_optimize.minimize(error, xx0[i], method="Nelder-Mead")["x"]
        return xx

    def _fit(self, x, skew_max=.9):
        skew = stats.skew(x)
        skew_23 = np.abs(skew) ** (2. / 3)
        if not (-skew_max < skew < skew_max):
            delta = np.sqrt(np.pi / 2 * skew_max ** (2. / 3) /
                            (skew_max ** (2. / 3) +
                             ((4 - np.pi) / 2) ** (2. / 3)))
            delta = np.copysign(delta, skew)
#             warnings.warn("Sample skew %.2f not in feasible range ~(-1,1)!" %
#                           skew)
        else:
            delta = np.sqrt(np.pi / 2 * skew_23 /
                            (skew_23 + ((4 - np.pi) / 2) ** (2. / 3)))
        delta = np.copysign(delta, skew)
        alpha = delta / np.sqrt(1 - delta ** 2)
        omega = np.sqrt(np.var(x) / (1 - 2 * delta ** 2 / np.pi))
        zeta = np.mean(x) - omega * np.sqrt(2 / np.pi) * delta
        return self.fit_ml(x, x0=(zeta, omega, alpha)).x
        if not (-skew_max < skew < skew_max):
            zeta, omega, alpha = self.fit_ml(x, x0=(zeta, omega, alpha / 2))
            delta = alpha / np.sqrt(1 + alpha ** 2)
            skew_ = ((4 - np.pi) / 2 * (np.sqrt(2 / np.pi) * delta) ** 3 /
                     (1 - 2 * delta ** 2 / np.pi) ** (3. / 2))
            print alpha, skew, skew_
        return zeta, omega, alpha
skewnorm = SkewNormal()


class TruncatedNormal(MyDist):
    _feasible_start = (0, 1, -1, 1)
    _additional_kwds = {"lc": -1, "uc": 1}

    def _pdf(self, x, mu=0, sigma=1, lc=-np.inf, uc=np.inf):
        x, mu, sigma, lc, uc = np.atleast_1d(x, mu, sigma, lc, uc)
        un_trunc = np.atleast_1d(norm.pdf(x, mu, sigma) /
                                 (norm.cdf(uc, mu, sigma) -
                                  norm.cdf(lc, mu, sigma)))
        un_trunc[(x < lc) | (x > uc)] = 0
        sigma_neg_mask = sigma <= 0
        if np.any(sigma_neg_mask):
            un_trunc[sigma_neg_mask] += sigma[sigma_neg_mask] ** 2
        return un_trunc

    def _cdf(self, x, mu=0, sigma=1, lc=-np.inf, uc=np.inf):
        x, mu, sigma, lc, uc = np.atleast_1d(x, mu, sigma, lc, uc)
        q = ((norm.cdf(x, mu, sigma) - norm.cdf(lc, mu, sigma)) /
             (norm.cdf(uc, mu, sigma) - norm.cdf(lc, mu, sigma)))
        q = np.atleast_1d(q)
        q[q < 0] = 0
        q[q > 1] = 1
        return q

    def _ppf(self, quantiles, mu=0, sigma=1, lc=-np.inf, uc=np.inf):
        quantiles, mu, sigma, lc, uc = \
            np.atleast_1d(quantiles, mu, sigma, lc, uc)
        x = norm.ppf(quantiles * norm.cdf(uc, mu, sigma) +
                     (1 - quantiles) * norm.cdf(lc, mu, sigma), mu, sigma)
        x = np.atleast_1d(x)
        lower_ii = x < lc
        if np.sum(lower_ii) > 0:
            x[lower_ii] = lc[lower_ii]
        upper_ii = x > uc
        if np.sum(upper_ii) > 1:
            x[upper_ii] = uc[upper_ii]
        return x

    def _fit(self, x, lc=-np.inf, uc=np.inf):
        x0 = [x.mean(), x.std()]
        kwds = {}
        if lc is None:  # or np.isneginf(lc):
            x0 += [x.min()]
        else:
            kwds["lc"] = lc
        if uc is None:  # or np.isinf(uc):
            x0 += [x.max()]
        else:
            kwds["uc"] = uc
        return self.fit_ml(x, x0=x0, method="Nelder-Mead", **kwds).x
#        return self.fit_ml(x, sp_optimize.fmin, l, u)
#        xmean = x.mean()
#        xstd = x.std()
#        fac1 = ((norm.pdf(l, xmean, xstd) - norm.pdf(u, xmean, xstd)) /
#                (norm.cdf(l, xmean, xstd) - norm.cdf(u, xmean, xstd)))
#        fac2 = (((l - xmean) / xstd * norm.pdf(l, xmean, xstd) -
#                 (u - xmean) / xstd * norm.pdf(u, xmean, xstd)) /
#                (norm.cdf(l, xmean, xstd) - norm.cdf(u, xmean, xstd)))
#        mu = xmean + fac1 * xstd
#        sigma = xstd ** 2 * (1 + fac2 - fac1 ** 2)
#        return mu, sigma, l, u

    def _constraints(self, x, mu, sigma, lc=-np.inf, uc=np.inf):
        if np.any(x > uc):
            return False
        if np.any(x < lc):
            return False
        if np.any(sigma <= 0):
            return False
        return True
truncnorm = TruncatedNormal()


class LogNormal(MyDist):
    _feasible_start = (0, 1)

    def _pdf(self, x, mu, sigma):
        return norm.pdf(np.log(x + 1e-9), mu, sigma)

    def _cdf(self, x, mu, sigma):
        return norm.cdf(np.log(x + 1e-9), mu, sigma)

    def _ppf(self, quantiles, mu, sigma):
        return np.exp(norm.ppf(quantiles, mu, sigma))

    def _fit(self, x):
        x_noninf = x[(x > 0) & np.isfinite(x)]
        mu = np.mean(np.log(x_noninf))
        sigma = np.std(np.log(x_noninf))
        return mu, sigma

    def _constraints(self, x, mu, sigma):
        if np.any(x <= 0):
            return False
        if np.any(sigma <= 0):
            return False
        return True
lognormal = LogNormal()


class JohnsonSU(MyDist):
    _feasible_start = (1, 1, 0, 1)

    def _pdf(self, x, a, b, loc, scale):
        x, a, b, loc, scale = np.atleast_1d(x, a, b, loc, scale)
        x = (x - loc) / scale
        x2 = x * x
        trm = norm.pdf(a + b * np.log(x + np.sqrt(x2 + 1)))
        dens = b * 1.0 / np.sqrt(x2 + 1.0) * trm / scale
        return dens

    def _cdf(self, x, a, b, loc, scale):
        x = np.atleast_1d(x - loc) / scale
        q = np.atleast_1d(norm.cdf(a + b * np.log(x + np.sqrt(x * x + 1))))
        q[np.isneginf(x)] = 0
        return q

    def _ppf(self, q, a, b, loc, scale):
        q, a, b, loc, scale = np.atleast_1d(q, a, b, loc, scale)
        z = np.sinh((norm.ppf(q) - a) / b)
        x = z * scale + loc
        return x

    def _fit(self, x):
        x_noinf = x[np.isfinite(x)]
        loc, scale = x_noinf.mean(), x_noinf.std()
        return self.fit_fsum(x_noinf, x0=(1, 1, loc, scale))

    def _constraints(self, a, b, loc, scale):
        if np.any(b <= 0):
            return False
        return True
johnsonsu = JohnsonSU()


class Cauchy(MyDist):
    _feasible_start = (0, 1)

    def _pdf(self, x, x0, gamma):
        return 1 / (np.pi * gamma * (1 + ((x - x0) / gamma) ** 2))

    def _cdf(self, x, x0, gamma):
        return (1 / np.pi) * np.arctan2(x - x0, gamma) + .5

    def _ppf(self, q, x0, gamma):
        return x0 + gamma * np.tan(np.pi * (q - .5))

    def _fit(self, x):
        x0 = np.median(stats.trimboth(x, .38))
        inner_quart = stats.trimboth(x, .25)
        inner_range = inner_quart.max() - inner_quart.min()
        return self.fit_ml(x, x0=(x0, .5 * inner_range)).x

    def _constraints(self, x, x0, gamma):
        if np.any(gamma < 0):
            return False
        return True
cauchy = Cauchy()


class StudentT(MyDist):
    _feasible_start = (0, 1)

    def _pdf(self, x, mu=0, df=1):
        x, mu, df = np.atleast_1d(x, mu, df)
        Px = (np.exp(gammaln((df + 1) / 2) - gammaln(df / 2)) /
              np.sqrt(df * np.pi) *
              (1 + (x - mu) ** 2 / df) ** (-(df + 1) / 2))
        return Px

    def _cdf(self, x, mu=0, df=1):
        q = stdtr(df, np.atleast_1d(x) - mu)
        return q

    def _ppf(self, q, mu=0, df=1):
        x = stdtrit(df, np.atleast_1d(q)) + mu
        x[q == 0] = -np.inf
        x[q == 1] = np.inf
        return x

    def _fit(self, x):
        x_noninf = x[np.isfinite(x)]
        sigma = np.var(x_noninf)
        df = 2 * sigma / (sigma - 1)
        if df > 6:
            df = 6
        return self.fit_ml(x_noninf,
                           x0=(np.mean(x_noninf), df if df > 0 else .1),
                           method="Nelder-Mead").x

    def _constraints(self, x, mu, df):
        x, mu, df = (np.asarray(var) for var in (x, mu, df))
        if np.any((df <= 0) | (df > 8)):
            return False
        return True
student_t = StudentT()


class NoncentralT(MyDist):
    _feasible_start = (2, 3, 0)

    # stolen from scipy.stats.distributions.nct
    def _pdf(self, x, df, nc, mu=0):
        x, df, nc = (np.atleast_1d(var).astype(float) for var in (x, df, nc))
        pdf_x = lambda x: ((df / x) *
                           (self.cdf(x * (1 + 2 / df) ** .5, df + 2, nc) -
                            self.cdf(x, df, nc)))
        dens = np.where(np.abs(x - mu) <= .5,
                        self._pdf_old(x, df, nc, mu),
                        pdf_x(x - mu))
        dens[np.isinf(x)] = 0
        return dens

    def _pdf_old(self, x, df, nc, mu=0):
        n = df * 1.0
        nc = nc * 1.0
        x2 = (x - mu) ** 2
        ncx2 = nc * nc * x2
        fac1 = n + x2
        trm1 = (n / 2. * np.log(n) + gammaln(n + 1) -
                (n * np.log(2) + nc * nc / 2. + (n / 2.) * np.log(fac1) +
                 gammaln(n / 2.)))
        Px = np.exp(trm1)
        valF = ncx2 / (2 * fac1)
        trm1 = (np.sqrt(2) * nc * (x - mu) * hyp1f1(n / 2 + 1, 1.5, valF) /
                (np.asarray(fac1 * gamma_func((n + 1) / 2))))
        trm2 = (hyp1f1((n + 1) / 2, 0.5, valF) /
                (np.asarray(np.sqrt(fac1) * gamma_func(n / 2 + 1))))
        Px *= trm1 + trm2
        return Px

    def _cdf(self, x, df, nc, mu=0):
        x, df, nc, mu = (np.atleast_1d(var).astype(float)
                         for var in (x, df, nc, mu))
        q = nctdtr(df, nc, x - mu)
        return q

    def _ppf(self, q, df, nc, mu=0):
        q, df, nc = (np.atleast_1d(var).astype(float) for var in (q, df, nc))
        x = nctdtrit(df, nc, q) + mu
        x[q == 0] = -np.inf
        x[q == 1] = np.inf
        return x

    def _fit(self, x):
        x_noninf = x[np.isfinite(x)]
        mode = stats.mode(np.round(x_noninf, 1))[0][0]
        x0 = (1, 2 * stats.skew(x_noninf), mode)
        df, nc, mu = self.fit_ml(x_noninf, x0=x0, method="Nelder-Mead").x
#        if df >= 30:
#            # then we are normal anyway
#            df = 30
        return df, nc, mu

    def _constraints(self, x, df, nc, mu):
        df = np.asarray(df).astype(float)
        if np.any((df < .1) | (df > 8)):
            return False
        return True
noncentral_t = NoncentralT()


# TODO: implement ppf
# class LogitNormal(MyDist):
#    def logit(self, x):
#        return np.log(x / (1 - x))
#
#    def _pdf(self, x, mu, sigma, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        return norm_pdf(logit, mu, sigma) / (x_normed * (1 - x_normed))
#
#    def _cdf(self, x, mu, sigma, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        return norm_cdf(logit, mu, sigma)
#
#    def _ppf(self, q, mu, sigma, l=0, u=1):
#        raise NotImplementedError
#
#    def _fit(self, x, l=None, u=None):
#        l = x.min() - 1e-9 if l is None else l
#        u = x.max() + 1e-9 if u is None else u
#        x_normed = (x - l) / (u - l)
#        logit = self.logit(x_normed)
#        mu, sigma = norm_fit(logit)
#        return mu, sigma, l, u
#
#    def _constraints(self, x, mu, sigma, l=0, u=1):
#        x, mu, sigma, l, u = (np.asarray(var) for var in (x, mu, sigma, l, u))
#        if np.any(x < l):
#            return False
#        if np.any(x > u):
#            return False
#        return True
# logitnormal = LogitNormal()


# class Rayleigh(MyDist):
#    def _pdf(self, x, sigma):
#        return (x / sigma ** 2) * np.exp(-x ** 2 / (2 * sigma ** 2))
#
#    def _cdf(self, x, sigma):
#        return 1 - np.exp(-x ** 2 / (2 * sigma ** 2))
#
#    def _fit(self, x):
#        return ((.5 * np.average(x ** 2)) ** .5,)
#
#    @abstractmethod
#    def _ppf(self):
#        raise NotImplementedError
#
#    def _constraints(self, x, sigma):
#        if np.any(sigma <= 0):
#            return False
#        return True
# rayleigh = Rayleigh()


# TODO: implement ppf
# class RayleighLU(MyDist):
#    def _pdf(self, x, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return (x_normed / sigma ** 2) * \
#            np.exp(-x_normed ** 2 / (2 * sigma ** 2))
#
#    def _cdf(self, x, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return 1 - np.exp(-x_normed ** 2 / (2 * sigma ** 2))
#
#    def _fit(self, x, l=None, u=None):
#        l = x.min() if l is None else l
#        u = x.max() if u is None else u
#        x_normed = (x - l) / (u - l)
#        return (.5 * np.average(x_normed ** 2)) ** .5, l, u
#
#    def _ppf(self):
#        raise NotImplementedError
#
#    def _constraints(self, x, sigma, l, u):
#        if np.any(sigma <= 0):
#            return False
#        if np.any(l >= u):
#            return False
#        return True
# rayleighlu = RayleighLU()


# TODO: fit does not work well
# class LogNormalLU(MyDist):
#    def _pdf(self, x, mu, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return norm_pdf(np.log(x_normed + 1e-9), mu, sigma) / (u - l)
#
#    def _cdf(self, x, mu, sigma, l, u):
#        x_normed = (x - l) / (u - l)
#        return norm_cdf(np.log(x_normed + 1e-9), mu, sigma)
#
#    def _ppf(self, quantiles, mu, sigma, l, u):
#        return np.exp(norm_ppf(quantiles, mu, sigma)) * (u - l) + l
#
#    def _fit(self, x, l=None, u=None):
#        x_noninf = x[np.isfinite(x)]
#        l = x_noninf.min() if l is None else l
#        u = x_noninf.max() if u is None else u
#        x_normed = (x_noninf - l) / (u - l)
#        mu, sigma = norm_fit(np.log(x_normed + 1e-9))
#        return self.fit_ml(x_noninf, x0=(mu, sigma, l, u))
#
#    def _constraints(self, x, mu, sigma, l, u):
#        if np.any(l >= u):
#            return False
#        return True
# lognormallu = LogNormalLU()
# lognormallu._feasible_start = (0, 1, -1, 1)


# class WeibullLU(MyDist):
#    def _pdf(self, x, alpha, beta, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        # alpha and beta should be positive
#        densities = alpha * beta * x_normed ** (beta - 1) * \
#            np.exp(-alpha * x_normed ** beta)
#        return densities / (u - l)
#
#    def _cdf(self, x, alpha, beta, l=0, u=1):
#        x_normed = (x - l) / (u - l)
#        return 1 - np.exp(-alpha * x_normed ** beta)
#
#    def _ppf(self, quantiles, alpha, beta, l=0, u=1):
#        x = (-np.log(1 - quantiles) / alpha) ** (1 / beta)
#        return x * (u - l) + l
#
#    def _fit(self, values, beta_start=10, *args, **kwds):
#        """Implements a semi-analytical method of moments. beta is found by
#        minimizing errors. After that, alpha can be estimated with the help of
#        beta.
#        See
#        http://interstat.statjournals.net/YEAR/2000/articles/0010001.norm_pdf
#        """
#        values_noninf = values[np.isfinite(values)]
#        l = values_noninf.min()
#        u = values_noninf.max()
#        values_noninf = (values_noninf - l) / (u - l)
#        # we need the coefficient of variation now and the mean later. so do
#        # not use stats.variation in order to not calculate the mean twice.
#        mean = np.mean(values_noninf)
#        cv = np.std(values_noninf) / mean
#
#        def cv_error(beta):
#            """Squared error between the theoretical and empirical coefficient
#            of variation."""
#            gamma1 = special.gamma(1 + 2 / beta)
#            gamma2 = special.gamma(1 + 1 / beta)
#            return (cv - (gamma1 - gamma2 ** 2) ** .5 / gamma2) ** 2
#        beta = sp_optimize.fmin(cv_error, [beta_start], disp=False)[0]
#        alpha = (mean / special.gamma(1 / beta + 1)) ** (-beta)
#        return alpha, beta, l, u
#
#    def _constraints(self, x, alpha, beta, l, u):
#        x, alpha, beta = (np.asarray(var) for var in (x, alpha, beta))
#        if np.any(alpha <= 0):
#            return False
#        if np.any(beta <= 0):
#            return False
#        if np.any(x < l):
#            return False
#        if np.any(x > u):
#            return False
#        return True
# weibulllu = WeibullLU()
# weibulllu._feasible_start = (1, 1, -1, 1)


class Weibull(MyDist):
    _feasible_start = (1, 1)

    def _pdf(self, x, alpha, beta):
        # alpha and beta should be positive
        densities = alpha * beta * x ** (beta - 1) * \
            np.exp(-alpha * x ** beta)
        return densities

    def _cdf(self, x, alpha, beta):
        return 1 - np.exp(-alpha * x ** beta)

    def _ppf(self, quantiles, alpha, beta):
        return (-np.log(1 - quantiles) / alpha) ** (1 / beta)

    def _fit(self, values, beta_start=10, *args, **kwds):
        """Implements a semi-analytical method of moments. beta is found by
        minimizing errors. After that, alpha can be estimated with the help of
        beta.
        See
        http://interstat.statjournals.net/YEAR/2000/articles/0010001.norm_pdf
        """
        # we need the coefficient of variation now and the mean later. so do
        # not use stats.variation in order to not calculate the mean twice.
        mean = np.nanmean(values)
        cv = np.nanstd(values) / mean

        def cv_error(beta):
            """Squared error between the theoretical and empirical coefficient
            of variation."""
            gamma1 = special.gamma(1 + 2 / beta)
            gamma2 = special.gamma(1 + 1 / beta)
            return (cv - (gamma1 - gamma2 ** 2) ** .5 / gamma2) ** 2
        beta = sp_optimize.fminbound(cv_error, -1, 1e6, disp=False)
        alpha = (mean / special.gamma(1 / beta + 1)) ** (-beta)
        return alpha, beta

    def _constraints(self, x, alpha, beta):
        x, alpha, beta = (np.asarray(var) for var in (x, alpha, beta))
        if np.any(x < 0):
            return False
        if np.any(alpha <= 0):
            return False
        if np.any(beta <= 0):
            return False
        return True
weibull = Weibull()


class Kumaraswamy(MyDist):
    """Resembles the Beta distribution, but does not need a transcendental
    function."""
    _feasible_start = (1, 1, 0, 1)

    def _pdf(self, x, a, b, l=0, u=1):
        x = np.atleast_1d(x).astype(float)
        if NE:
            x = ne.evaluate("(x - l) / (u - l)")
            dens = ne.evaluate("a * b * x ** (a - 1) * (1 - x ** a) ** " +
                               "(b - 1) / (u - l)")
        else:
            x = (x - l) / (u - l)
            dens = a * b * x ** (a - 1) * (1 - x ** a) ** (b - 1) / (u - l)
        return dens

    def _cdf(self, x, a, b, l=0, u=1):
        x = np.atleast_1d(x).astype(float)
        if np.any((x < l) | (x > u)):
            warnings.warn("Some values below lower or above upper bounds.")
        return 1 - (1 - ((x - l) / (u - l)) ** a) ** b

    def _ppf(self, quantiles, a, b, l=0, u=1):
        quantiles, a, b, l, u = np.atleast_1d(quantiles, a, b, l, u)
        if NE:
            x = ne.evaluate("((1 - (1 - quantiles) ** (1 / b)) ** " +
                            "(1 / a)) * (u - l) + l")
        else:
            x = ((1 - (1 - quantiles) ** (1 / b)) ** (1 / a)) * (u - l) + l
        return x

    def _fit(self, x, l=None, u=None):
        return beta.fit(x, l, u)

    def _constraints(self, x, a, b, l=0, u=1):
        x, a, b, l, u = (np.asarray(var) for var in (x, a, b, l, u))
        if np.any(a < 0):
            return False
        if np.any(b < 0):
            return False
        if np.any(x < l):
            return False
        if np.any(x > u):
            return False
        return True
kumaraswamy = Kumaraswamy()


class Beta(MyDist):
    """Beta distribution on the interval [l, u]."""
    _feasible_start = (1, 1, 0, 1)

    def _pdf(self, x, alpha, beta, l=0, u=1):
        x, alpha, beta, l, u = map(np.asarray, (x, alpha, beta, l, u))
        # putting the values into [0, 1] according to [l, u]
        # we avoid the value of exactly 0 or 1 to not get numerical problems
#        x_normed = (x - l) / (u - l)
#        xx = np.copy(x_normed)
#        ll, uu = sys.float_info.min, 1 - sys.float_info.min
#        x_normed = np.where(x_normed >= uu, uu, x_normed)
#        x_normed = np.where(x_normed <= ll, ll, x_normed)
        if NE:
            x = ne.evaluate("(x - l) / (u - l)")
            return 1 / special.beta(alpha, beta) * \
                ne.evaluate("x ** (alpha - 1) * (1 - x) ** (beta - 1) / " +
                            "(u - l)")
        else:
            x = (u - l) / (u - l)
            return (1 / special.beta(alpha, beta) * x ** (alpha - 1) * (1 - x)
                    ** (beta - 1) / (u - l))

    def _cdf(self, x, alpha, beta, l=0, u=1):
        if np.any((x < l) | (x > u)):
            warnings.warn("Some values below lower or above upper bounds.")
        # note: the betainc function below really delivers the regularized
        # incomplete beta function
        return special.betainc(alpha, beta, (x - l) / (u - l))

    def _ppf(self, q, alpha, beta, l=0, u=1):
        return (u - l) * special.betaincinv(alpha, beta, q) + l

    def _fit(self, x, l=None, u=None):
        """stolen from
        http://en.wikipedia.org/wiki/Beta_distribution#Parameter_estimation."""
        # remember whether l and u where given. if they were, we do not return
        # them
        if l is None:
            l = x.min()
            return_l = True
        else:
            return_l = False
        if u is None:
            u = x.max()
            return_u = True
        else:
            return_u = False

        # have to make sure that the parameters all have the same lengths
        try:
            ll = np.empty_like(u)
            ll[:] = l
            l = ll
        except (TypeError, ValueError):
            pass
        try:
            uu = np.empty_like(l)
            uu[:] = u
            u = uu
        except (TypeError, ValueError):
            pass

        xmean = (x.mean() - l) / (u - l)
        xvar = x.var() / (u - l) ** 2
        alpha = xmean * (xmean * (1 - xmean) / xvar - 1)
        beta = (1 - xmean) * alpha / xmean

        return_list = [abs(alpha), abs(beta)]
        if return_l:
            return_list += [l]
        if return_u:
            return_list += [u]
        return return_list
#        return self.fit_ml(x, x0=(alpha, beta, l, u))

    def _constraints(self, x, alpha, beta, l=0, u=1):
        x, alpha, beta, l, u = (np.asarray(var)
                                for var in (x, alpha, beta, l, u))
        if np.any(alpha <= 0):
            return False
        if np.any(beta <= 0):
            return False
        if np.any(x < l):
            return False
        if np.any(x > u):
            return False
        return True
beta = Beta()


class Gamma(MyDist):
    """Gamma distribution."""
    _feasible_start = (2, 2)

    def _pdf(self, x, k, theta):
        x = np.atleast_1d(x)
        dens = (x ** (k - 1) * np.exp(-x / theta) /
                (special.gamma(k) * theta ** k))
        return dens

    def _cdf(self, x, k, theta):
        x = np.atleast_1d(x)
        q = gammainc(k, x / theta) / gamma_func(k)
        q[np.isinf(x)] = 1
        return q

    def _ppf(self, q, k, theta):
        q = np.atleast_1d(q)
        x = theta * gammaincinv(k, q * gamma_func(k))
        x[q == 1] = np.inf
        return x

    def _fit(self, x, rel_change=1e-6):
        """A maximum-likelihood estimator. See
        http://en.wikipedia.org/wiki/Gamma_distribution#Parameter_estimation
        """

        def psi(k):
            if k >= 8:
                return np.log(k) - (1 + (1 - (.1 - 1 / (21 * k ** 2)) /
                                         k ** 2) / (6 * k)) / (2 * k)
            else:
                return psi(k + 1) - 1. / k

        def psi_(k):
            if k >= 8:
                return (1 + (1 + (1 - (1. / 5 - 1 / (7 * k ** 2)) / k ** 2) /
                        (3 * k)) / (2 * k)) / k
            else:
                return psi_(k + 1) + 1 / k ** 2

        xmean = x[np.isfinite(x)].mean()
        # we get problems here if there are 0's in x (log(x) = -inf, which
        # happens for wind)
        s = np.log(xmean) - np.mean(np.log(x[np.isfinite(x) & (x > 0)]))
        k_old = (3 - s + ((s - 3) ** 2 + 24 * s) ** .5) / (12 * s)
        k_rec = lambda k: k - (np.log(k) - psi(k) - s) / (1. / k - psi_(k))
        k_new = k_rec(k_old)
        while abs(k_old - k_new) / k_old > rel_change:
            k_old = k_new
            k_new = k_rec(k_old)
        theta = xmean / k_new
        return self.fit_fsum(x, x0=(k_new, theta))

    def _constraints(self, x, k, theta):
        x, k, theta = (np.asarray(var) for var in (x, k, theta))
        if np.any(k <= 0):
            return False
        if np.any(theta <= 0):
            return False
        if np.any(x < 0):
            return False
        return True
gamma = Gamma()


class Gamma1(MyDist):
    _feasible_start = (1, .5, 2)

    def _pdf(self, x, a, loc, scale):
        x, a, loc, scale = np.atleast_1d(x, a, loc, scale)
        z = (x - loc) / scale
        dens = (np.exp((a - 1) * np.log(z) - z - gammaln(a))) / scale

        return dens

    def _cdf(self, x, a, loc, scale):
        x, a, loc, scale = np.atleast_1d(x, a, loc, scale)
        z = (x - loc) / scale
        q = gammainc(a, z)
        q[np.isinf(x)] = 1
        return q

    def _ppf(self, q, a, loc, scale):
        q, a, loc, scale = np.atleast_1d(q, a, loc, scale)
        x = gammaincinv(a, q) * scale + loc
        x[q == 1] = np.inf
        return x

    def _fit(self, x):
        return stats.gamma.fit(x[np.isfinite(x)])
        # this is stolen from sp.stats.gamma-gen and not understood
#        x_noninf = x[np.isfinite(x)]
#        a = 4 / stats.skew(x_noninf) ** 2
#        muhat = np.array(x_noninf).mean()
#        mu2hat = np.array(x_noninf).var()
#        Shat = (mu2hat / a) ** .5
#        Lhat = muhat - Shat * a
#        return self.fit_ml(x_noninf, x0=(a, Lhat, Shat))
gamma1 = Gamma1()


class Expon(MyDist):
    _feasible_start = (0, 1)

    def _pdf(self, x, x0, lambd):
        return lambd * np.exp(-lambd * (x - x0))

    def _cdf(self, x, x0, lambd):
        return 1 - np.exp(-lambd * (x - x0))

    def _ppf(self, q, x0, lambd):
        return -np.log(1 - q) / lambd + x0

    def _fit(self, x):
        x_noninf = x[np.isfinite(x)]
        x0 = np.min(x_noninf)
        lambd = 1 / np.mean(x_noninf)
        return x0, lambd

    def _constraints(self, x, x0, lambd):
        if np.any(x <= x0):
            return False
        return True
expon = Expon()


class ExponTwo(MyDist):
    _feasible_start = (0, .5, 1, 1)

    def _pdf(self, x, x0, q0, lambda1, lambda2):
        x = np.atleast_1d(x)
        dens = np.where(x <= x0,
                        q0 * expon.pdf(x0 - x, 0, lambda1),
                        (1 - q0) * expon.pdf(x, x0, lambda2))
        return dens

    def _cdf(self, x, x0, q0, lambda1, lambda2):
        x = np.atleast_1d(x)
        q = np.where(x <= x0,
                     q0 * (1 - expon.cdf(x0 - x, 0, lambda1)),
                     q0 + (1 - q0) * expon.cdf(x, x0, lambda2))
        return q

    def _ppf(self, q, x0, q0, lambda1, lambda2):
        q = np.atleast_1d(q)
        x = np.where(q <= q0,
                     x0 + np.log(q / q0) / lambda1,
                     x0 - np.log(1 - (q - q0) / (1 - q0)) / lambda2)
        return x

    def _fit(self, x):
        x_noninf = x[np.isfinite(x)]
        x0 = stats.mode(np.round(x_noninf, 1))[0][0]
        lambda1 = -1 / np.mean(x_noninf[x_noninf < x0] - x0)
        lambda2 = 1 / np.mean(x_noninf[x_noninf > x0] - x0)
        q0 = (np.argmin(np.abs(np.sort(x_noninf) - x0)) + .5) / len(x)
        return self.fit_ml(x_noninf, x0=(x0, q0, lambda1, lambda2),
                           method="Nelder-Mead").x

    def _constraints(self, x, x0, q0, lambda1, lambda2):
        if np.any(lambda1 <= 0):
            return False
        if np.any(lambda2 <= 0):
            return False
        if np.any((q0 < 0) | (q0 > 1)):
            return False
        return True
expon_two = ExponTwo()


class NoncentralLaplace(MyDist):
    _feasible_start = (1, .5, .5)

    def _pdf(self, x, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        x = np.atleast_1d(x)
        dens = np.where(x <= x0,
                        q0 * lambd * np.exp(-lambd * (x0 - x)),
                        (1 - q0) * lambda2 * np.exp(-lambda2 * (x - x0)),
                        )
        dens[np.isinf(x)] = 0
        return dens

    def _cdf(self, x, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        x = np.atleast_1d(x)
        q = np.where(x <= x0,
                     q0 * (1 - expon.cdf(x0 - x, 0, lambd)),
                     q0 + (1 - q0) * expon.cdf(x, x0, lambda2))
        return q

    def _ppf(self, q, x0, q0, lambd):
        lambda2 = lambd * q0 / (1 - q0)
        q = np.atleast_1d(q)
        x = np.where(q <= q0,
                     x0 + np.log(q / q0) / lambd,
                     x0 - np.log(1 - (q - q0) / (1 - q0)) / lambda2)
        return x

    def _fit(self, x):
        x0 = stats.mode(np.round(x, 1))[0][0]
        q0 = (np.argmin(np.abs(np.sort(x) - x0)) + .5) / len(x)
        lambd = (1 - 2 * q0) / (q0 * np.mean(x[np.isfinite(x)]))
        if lambd < 0:
            # this actually means that x0 was not right, so provide a feasible
            # lambd and let the ml estimation figure the rest out
            lambd = -1. / np.mean(x[(x < x0) & np.isfinite(x)] - x0)
        return self.fit_ks(x, x0=(x0, q0, lambd))

    def _constraints(self, x, x0, q0, lambd):
        if np.any(lambd <= 0):
            return False
        if np.any((q0 < 0) | (q0 > 1)):
            return False
        return True
noncentral_laplace = NoncentralLaplace()


if __name__ == "__main__":
    import vg
    import config_konstanz
    vg.conf = config_konstanz
    times_hourly, met = vg.read_met(vg.conf.met_file)
    rain, times = vg.my.sumup(met["R"], 24, times_hourly)
#     met_vg = vg.VG(("R", "theta"), verbose=True, refit="R")
#     rain = met_vg.data_raw[0]
    trunc_gamma = vg.distributions.Rain(vg.distributions.weibull,
                                        threshold=.0002)
    rain_dist = vg.sd.SlidingDist(trunc_gamma, rain, times,
                                  verbose=True)
    solution = rain_dist.fit()
    rain_dist.plot_monthly_fit(solution)
    rain_dist.scatter_pdf(solution)
    plt.show()
#     import vg
#     from vg.core import config_shira
#     vg.conf = config_shira
#     dt_hourly, met = vg.read_met(os.path.join(vg.conf.data_dir,
#                                               vg.conf.met_file))
#     cloud, dt = vg.my.sumup(met["Cloudness"], 8, times_=dt_hourly,
#                             sum_to_nan=False)
#     cloud_dist = vg.sd.SlidingDist((vg.distributions.Censored,
#                                     vg.distributions.norm),
#                                cloud / 80, dt,
#                                fixed_pars=vg.conf.par_known["Cloudness"],
#                              verbose=True)
#     sol = cloud_dist.fit()
#     fig, ax = cloud_dist.scatter_pdf()
#     plt.show()


#    R, dtimes = np.squeeze(vg.my.sumup(met["R"], 24, dtimes_))
#    R = R.astype(float)
#    R_dist = stats.gamma(*stats.gamma.fit(R))
#    R_dist.pdf(0)
#    cen_exp = vg.distributions.Censored(vg.distributions.expon, lc=0)
#    r_cen = cen_exp(*cen_exp.fit(R))
