import numpy as np
import matplotlib.pyplot as plt
from vg import helpers as my


def trans_prob(rain, thresh=.0001):
    """Estimates transition probabilities.

    Parameter
    --------
    rain : 1d array
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m]

    Returns
    -------
    trans : (2, 2) ndarray
        Transition probabilities.
        [[P(no rain -> no rain), P(no rain -> rain],
         [P(rain -> no rain),    P(rain -> rain)]]
    """
    rain_mask = rain >= thresh
    n = float(len(rain))
    diffs = np.diff(rain_mask.astype(int))
    no_change_mask = diffs == 0
    p11 = np.sum(rain_mask[:-1] & no_change_mask)
    p00 = np.sum(~rain_mask[:-1] & no_change_mask)
    p10 = np.sum(diffs == -1)
    p01 = np.sum(diffs == 1)
    return np.array([[p00, p01],
                     [p10, p11]]) / n


def spell_lengths(rain, thresh=.001):
    """Calculate dry and wet spell lenghts.

    Parameter
    ---------
    rain : 1d array
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m]

    Returns
    -------
    dry : 1d int array
    wet : 1d int array
    """
    rain_mask = rain >= thresh
    dry = np.array([stop_i - start_i + 1
                    for start_i, stop_i in my.gaps(~rain_mask)])
    wet = np.array([stop_i - start_i + 1
                    for start_i, stop_i in my.gaps(rain_mask)])
    return dry, wet


def plot_exceedance(obs, sim=None, kind="depth", thresh=.001):
    """Plot exceedance probability.

    Parameter
    --------
    obs : 1-dim ndarray
        observations
    sim : 1 or 2-dim ndarray, optional
        simulated values. If 2-dimensional, median, and range of those values
        will be plotted.
    kind : "depth", "dry", "wet" or "all", optional
        What to plot:
        - depth: rain intensity
        - dry: lengths of dry spells
        - wet: lengths of wet spells
        - all
    thresh : float, optional
        threshold a value must exceed to be interpreted as rain [m].
        only needed if kind != "depth".
    """
    if kind == "all":
        fig, axs = plt.subplots(ncols=3, subplot_kw=dict(aspect="equal"))
        kinds = "depth", "dry", "wet"
    else:
        fig, axs = plt.subplots(subplot_kw=dict(aspect="equal"))
        axs = axs,
        kinds = kind,

    rel_ranks = lambda n: (np.arange(float(n)) - .5) / n

    def plot_sim_single(ax, sim):
        sim_conv = conv(sim)
        ranks = rel_ranks(len(sim_conv))
        sim_sorted = np.sort(sim_conv)[::-1]
        ax.loglog(sim_sorted, ranks, color="grey")

    for ax_i, kind in enumerate(kinds):
        ax = axs[ax_i]

        if kind == "depth":
            conv = lambda x: x[x >= thresh]
            ax.set_xlabel("depth [mm]")
        elif kind == "dry":
            conv = lambda x: spell_lengths(x, thresh)[0]
            ax.set_xlabel("dry spell length")
        elif kind == "wet":
            conv = lambda x: spell_lengths(x, thresh)[1]
            ax.set_xlabel("wet spell length")

        if np.ndim(sim) == 1:
            plot_sim_single(ax, sim)

        if np.ndim(sim) == 2:
            sim = np.asarray(sim)
            for sim_ in sim:
                plot_sim_single(ax, sim_)

        obs_conv = conv(obs)
        ranks = rel_ranks(len(obs_conv))
        obs_sorted = np.sort(obs_conv)[::-1]
        ax.loglog(obs_sorted, ranks, " ", marker="o", color="black",
                  fillstyle="none")

        ax.set_ylabel("F(X)")
        ax.grid()
    fig.tight_layout()
    return fig, axs


def richardson_model(T, trans_pp):
    """Simulate rain occurrence with Richardson's (1981) model (2 state, first
    order Markov Chain).

    Parameter
    ---------
    T : int
        Length of simulated time series
    trans_pp : (2, 2) ndarray
        Transition probabilities (see trans_prob).
        [[P(no rain -> no rain), P(no rain -> rain],
         [P(rain -> no rain),    P(rain -> rain)]]

    Returns
    -------
    occs : (T,) bool ndarray
        Rain occurrences.

    References
    ----------
    Richardson, C. W. (1981). Stochastic simulation of daily precipitation,
    temperature, and solar radiation. Water Resources Research, 17(1),
    182-190. doi:10.1029/WR017i001p00182

    See also
    --------
    trans_prob : Estimates transition probabilities given a time series.
    """
    pp = trans_pp
    T = int(round(T))
    occs = np.full(T, False, dtype=bool)
    rr = np.random.rand(T)
    # decide on the first occurrence with unconditional probability
    occs[0] = rr[0] < pp[0].sum
    for t in xrange(1, T):
        occs[t] = rr[t] < (pp[occs[t - 1], 1] / pp[:, occs[t - 1]].sum())
    return occs

if __name__ == "__main__":

    import vg
    import config_konstanz
    vg.conf = config_konstanz
    met_vg = vg.VG(("R", "theta"))
    rain = met_vg.data_raw[0] / 24
    plot_exceedance(rain, kind="all")
    plt.show()
