# -*- coding: utf-8 -*-
import os
import shelve
from contextlib import closing
import numpy as np
from scipy import optimize
from vg import times
from vg import helpers as my


shelve_filename = "resampler_cache.sh"


def _calibrate(**res_kwds):
    del res_kwds["calibrate"]
    data = res_kwds["data"]
    theta_i = res_kwds["theta_i"]
    theta = data[theta_i]
    # in order to have a roughly equally spaced mean array
    # means_incrs = np.linspace(0.5, 5, 12, endpoint=False)
    # theta_incrs = (-16.583 * np.log(1 - means_incrs /
    #                                 (1.1 * max(means_incrs)))) ** (1 / .513)
    theta_incrs = np.concatenate((np.linspace(0, 1, 10),
                                  np.linspace(2, 8, 5)))

    means = []
    for incr in theta_incrs:
        res, _ = resample(theta_incr=incr, calibrate=True, **res_kwds)
        theta_res = res[theta_i]
        means += [np.mean(theta_res)]
    means = np.array(means) - np.mean(theta)

    def error(args):
        a, b, c, d = args
        means_est = a * (1 - np.exp(-theta_incrs ** b / c)) + d
        return np.sum((means - means_est) ** 2)

    result = optimize.minimize(error,
                               (max(means) + min(means), .5, 1., min(means)),
                               options=dict(disp=True))

    # import matplotlib.pyplot as plt
    # plt.figure()
    # plt.scatter(theta_incrs, means)
    # a, b, c, d = result.x
    # plt.plot(theta_incrs, a * (1 - np.exp(-theta_incrs ** b / c)) + d)
    # plt.title("a: %.3f, b: %.3f, c: %.3f, d: %.3f" % (a, b, c, d))
    # plt.show()

    return result.x


def _transform_theta_incr(theta_incr, cache_dir=None, verbose=False,
                          recalibrate=False, **res_kwds):
    """Calibrate theta_incr in order to get a specific mean increase in
    temperature.

    Fits a curve of the following form:
      x = (- d \ln |1 - a / b|)^{1/c}
    """
    if cache_dir:
        shelve_filepath = os.path.join(cache_dir, shelve_filename)
    else:
        shelve_filepath = shelve_filename
    with closing(shelve.open(shelve_filepath, "c")) as sh:
        par_key = "{p}_{n_candidates}_{doy_tolerance}".format(**res_kwds)
        par_key_pos = par_key + "pos"
        if not recalibrate and par_key_pos in sh:
            a_pos, b_pos, c_pos, d_pos = sh[par_key_pos]
        else:
            if verbose:
                print "Calibrating resampler bias (positive leg)!"
            a_pos, b_pos, c_pos, d_pos = _calibrate(**res_kwds)
            sh[par_key_pos] = a_pos, b_pos, c_pos, d_pos

        par_key_neg = par_key + "neg"
        if not recalibrate and par_key_neg in sh:
            a_neg, b_neg, c_neg, d_neg = sh[par_key_neg]
        else:
            if verbose:
                print "Calibrating resampler bias (negative leg)!"
            # calibration for negative mean changes
            a_neg, b_neg, c_neg, d_neg = _calibrate(
                **{key: -val if key == "data" else val
                   for key, val in res_kwds.items()})
            sh[par_key_neg] = a_neg, b_neg, c_neg, d_neg

    def calc_d(x, a, b, c, d):
        return np.where(x < a,
                        (-c * np.log(1 - abs(x - d) / a)) ** (1 / b),
                        (-c * np.log(1e-4)) ** (1 / b))

    bias_pos = calc_d(theta_incr, a_pos, b_pos, c_pos, d_pos)
    bias_neg = calc_d(-theta_incr, a_neg, b_neg, c_neg, d_neg)
    bias = np.where(theta_incr >= d_pos, bias_pos, -bias_neg)

    # if verbose:
    #     import matplotlib.pyplot as plt
    #     plt.figure()
    #     plt.plot(theta_incr[0], "-x")
    #     plt.plot(bias[0], "-x")

    #     plt.figure()
    #     th = np.linspace(-8, 8, 500)
    #     _bias_pos = calc_d(th, a_pos, b_pos, c_pos, d_pos)
    #     _bias_neg = calc_d(-th, a_neg, b_neg, c_neg, d_neg)
    #     _bias = np.where(th >= d_pos, _bias_pos, -_bias_neg)
    #     plt.plot(th, _bias)
    #     plt.show()

    # # do not fiddle with the data, where no change is requested
    # return np.where(np.isclose(theta_incr, 0), 0, bias)
    # do fiddle with the data, resampler can have inherent bias
    return bias


def resample(data, dtimes, p=3, n_sim_steps=None, theta_incr=0.0,
             theta_i=0, n_candidates=None, doy_tolerance=60,
             cache_dir=None, calibrate=False, verbose=False,
             return_candidates=False, recalibrate=False, **kwds):
    """A simple multivariate resampler.
    Assumes that all variables are in the same dimension as similarity is
    evaluated in terms of summed absolute differences.

    Parameters
    ----------
    data : (K, T) float ndarray
        input data to mimic
        K - number of variables, T - number of time steps
    dtimes : (T,) ndarray of datetime objects
        dates of data.
    p : int, optional
        number of previous time steps to consider when searching for candidates
    n_sim_steps : int, optional
        number of time steps to simulate.
        If None, T is used (same as in `data`)
    theta_incr : scalar or None
        increase in average of temperature [deg C]
    theta_i : int
        row-index of temperature in data.  needed if theta_incr is not None.
    n_candidates : int or None, optional
        number of candidates to consider when choosing a similar time step.
        If None, it will be chosen as the square root of `T`.
    cache_dir : str or None, optional
        directory where calibration result is stored.
    calibrate : bool, optional
        if True, do not transform theta_incr.  This is needed for calibration.
    return_candidates : bool, optional
        return the resampling candidates for every time step.

    Returns
    -------
    sim : (K, n_sim_steps) float ndarray
        Resampled time series.
    chosen_indices : (n_sim_steps,) int ndarray
        Resampled indices from `data`
    candidate_series : (K, n_sim_steps, n_candidates) double ndarray
        Candidates at every time step.
    """
    if recalibrate or (theta_incr is not None and not calibrate):
        cal_kwds = my.ADict(locals()) - "return_candidates"
        theta_incr = _transform_theta_incr(**cal_kwds)
    if n_sim_steps is None:
        n_sim_steps = data.shape[1]
    K, T = data.shape
    if n_candidates is None:
        n_candidates = int(np.sqrt(T))
    candidate_series = np.empty((K, n_sim_steps, n_candidates))
    chosen_indices = np.empty(n_sim_steps, dtype=np.int)
    doys = times.datetime2doy(dtimes)
    # this is the hardest part
    # want to have T-1 chunks of (K, p) - shape
    # time domain indices, roughly in the form:
    # [(0,..,p), (1,...,p+1), ... (T-2*p,...,T-p)], but transposed!
    ii = np.arange(p)[:, None] + np.arange(data.shape[1] - p)[None, :]
    # this is now (K, p, T-p) - shape
    data_chunked = data[:, ii]

    if theta_incr is not None:
        bias_arr = np.zeros((K, n_sim_steps))
        bias_arr[theta_i, :] = np.atleast_2d(theta_incr)
        ii = np.arange(p)[:, None] + np.arange(n_sim_steps - p)[None, :]
        bias_chunked = bias_arr[:, ii]

    sim = np.empty((data.shape[0], n_sim_steps))
    # start with something present in the data
    doy_neighbors = np.where(times.doy_distance(doys[0], doys[:-p]) <=
                             doy_tolerance)[0]
    start_i = np.random.choice(doy_neighbors)
    sim[:, :p] = data[:, start_i:start_i + p]
    candidate_series[:, :p] = sim[:, :p, None]
    chosen_indices[:p] = range(start_i, start_i + p)

    # probability for choosing candidates, giving close candidates a
    # higher chance
    # see
    # Lall, U., & Sharma, A. (1996). A Nearest Neighbor Bootstrap For
    # Resampling Hydrologic Time Series. Water Resources Research,
    # 32(3), 679–693. http://doi.org/10.1029/95WR02966
    k_ji = 1. / np.arange(1, n_candidates + 1)
    k_ji /= np.sum(k_ji)

    for t in xrange(p, n_sim_steps):
        now = sim[:, t - p:t, None]
        # find time steps not far in terms of doy
        doy_distance = times.doy_distance(doys[t % len(doys)], doys[:-p])
        doy_neighbors = np.where(doy_distance <= doy_tolerance)[0]
        # which bias do we use here?
        if theta_incr is not None:
            bias = bias_chunked[..., t - p, None]
        else:
            bias = 0
        # what looks most like now?
        diffs = np.sum((data_chunked[..., doy_neighbors] -
                        now - bias) ** 2,
                       axis=(0, 1))
        candidates_i = np.argpartition(diffs, n_candidates)[:n_candidates]
        candidates_i = candidates_i[np.argsort(diffs[candidates_i])]
        if return_candidates:
            neighbors = doy_neighbors[candidates_i]
            candidate_series[:, t, :] = data[:, neighbors + p]
        doy_neighbor_i = np.random.choice(candidates_i, p=k_ji)
        neighbor_i = doy_neighbors[doy_neighbor_i]
        # save the origin for outside analysis
        chosen_indices[t] = neighbor_i
        # i points at the beginning of a chunk, we want the values one time
        # step further
        sim[:, t] = data[:, neighbor_i + p]

    if return_candidates:
        return sim, chosen_indices, candidate_series
    return sim, chosen_indices


if __name__ == "__main__":
    import vg
    import config_konstanz
    vg.conf = vg.vg_base.conf = config_konstanz
    met_vg = vg.VG((
        # "R",
        "theta", "Qsw",
        "ILWR", "rh", "u", "v"
                    ),
                   verbose=True)
    met_vg.fit(p=3)
    simt, sim = met_vg.simulate(res_kwds=dict(recalibrate=True,
                                              cy=False,
                                              n_candidates=20),
                                theta_incr=0.)

    theta_incrs = np.linspace(-10, 10, 20)
    data_mean = np.mean(met_vg.data_raw[0]) / 24
    delta_means = []
    for theta_incr in theta_incrs:
        simt, sim = met_vg.simulate(res_kwds=dict(cy=False,
                                                  n_candidates=20),
                                    theta_incr=theta_incr)
        delta_means += [np.mean(sim[0]) - data_mean]

    import matplotlib.pyplot as plt
    plt.plot(theta_incrs, delta_means, "-x")
    plt.plot(theta_incrs, theta_incrs, color="gray")
    plt.show()
