import warnings
import numpy as np
from scipy import optimize, interpolate, stats, sparse
import matplotlib.pyplot as plt
from vg import times
from vg.smoothing import smooth
from vg.time_series_analysis import seasonal

from vg import helpers as my


def apply_kernel_ne(kernel_width, data, eval_points=None):
    if eval_points is None:
        _distances = distance_array(data, data)
    else:
        _distances = distance_array(data, eval_points)
    _pi = np.pi
    densities = ne.evaluate(
        "1.0 / (sqrt(2 * _pi) * kernel_width) * " +
        "exp(-_distances ** 2 / (2 * kernel_width ** 2))")
    return densities


def apply_kernel_np(kernel_width, data, eval_points=None, mask=None,
                    recalc=False):
    if recalc or not hasattr(apply_kernel_np, "distances"):
        if eval_points is None:
            distances = distance_array(data, data)
            apply_kernel_np.distances = distances
        else:
            distances = distance_array(data, eval_points)
    else:
        distances = apply_kernel_np.distances
    return (1.0 / (np.sqrt(2 * np.pi) * kernel_width) *
            np.exp(-distances ** 2 / (2 * kernel_width ** 2)))


try:
    import numexpr as ne
    apply_kernel = apply_kernel_ne
    NE = True
except ImportError:
    warnings.warn("Could not import numexpr. Using numpy for KDE instead.")
    apply_kernel = apply_kernel_np
    NE = False


# @profile
def apply_2d_kernel_(kernel_width, data, doys, circ, doy_width,
                     eval_points=None, eval_doys=None):
    if eval_points is None:
        eval_points = data
    if eval_doys is None:
        eval_doys = doys
    if NE:
        def kernel(dist, width):
            _pi = np.pi
            return ne.evaluate("(sqrt(2 * _pi) * width) ** -1 *"
                               "exp(-dist ** 2 / (2 * width ** 2))")
    else:
        kernel = lambda dist, width: \
            ((np.sqrt(2 * np.pi) * width) ** -1 *
             np.exp(-dist ** 2 / (2 * width ** 2)))
    distances = distance_array(data, eval_points)
    densities = kernel(distances, abs(kernel_width))
    temp_distances = doyhour_distance(doys, eval_doys, circ=circ)
    time_dens = kernel(temp_distances, abs(doy_width))
    if NE:
        densities = ne.evaluate("densities * time_dens")
    else:
        densities *= time_dens
    return densities  # / (kernel_width * circ)


def apply_2d_kernel(kernels_width, data, hour_width, eval_points=None,
                    eval_doys=None):
    # data_width, hour_width = kernels_width
    data_width = kernels_width
    if eval_points is None:
        eval_points = data
    if NE:
        def kernel(dist, width):
            _pi = np.pi
            return ne.evaluate("(sqrt(2 * _pi) * width) ** -1 *"
                               "exp(-dist ** 2 / (2 * width ** 2))")
    else:
        kernel = lambda dist, width: \
            ((np.sqrt(2 * np.pi) * width) ** -1 *
             np.exp(-dist ** 2 / (2 * width ** 2)))
    # distance in data dimension
    distances = distance_array(data.ravel(), eval_points.ravel())
    densities = kernel(distances, abs(data_width))
    # distance in the two temporal dimensions
    hour_dist, doy_dist = np.meshgrid(range(data.shape[1]),
                                      range(data.shape[0]))
    hour_middle = data.shape[0] // 2
    doy_middle = data.shape[1] // 2
    time_distances = np.empty_like(data)
    temp = ne.evaluate("sqrt((hour_dist - hour_middle) ** 2 + "
                       "(doy_dist - doy_middle) ** 2)")
    time_distances[:] = temp[..., None]
    time_dens = kernel(time_distances.ravel(), abs(hour_width))
    if NE:
        densities = ne.evaluate("densities * time_dens")
    else:
        densities *= time_dens
    return densities


def gaussian_kernel(x, width):
    return (1. / (np.sqrt(2 * np.pi) * width) *
            np.exp(-x ** 2 / (2 * width ** 2)))


def gaussian_kernel_ne(x, width):
    _pi = np.pi
    return ne.evaluate("(1. / (sqrt(2 * _pi) * width) *" +
                       "exp(-x ** 2 / (2 * width ** 2)))")


@my.cache("dists", "rows", "cols")
# # @profile
def apply_sparse_kernel(kernel, widths, data, mask, eval_points=None,
                        **k_kwds):
    """Apply a given kernel on parts of data indicated by a mask.

    Parameter
    ---------
    kernel : callable
    widths : float or 1d array
    data : 1d array
    mask : 2d boolean array
        Indicates which pairs in data are close enough to each other in order
        to influence each others density.
    eval_points : None or 1d array, optional
        Where to evaluate kernel densities.
    **k_kwds : optional
        Extra key-word arguments are passed into the supplied kernen callable.

    Returns
    -------
    dens : sparse.csr matrix
    """
    self = apply_sparse_kernel
    x_vec = data
    if eval_points is None:
        y_vec = data
    else:
        y_vec = eval_points
    # this is a poor man's cache invalidation!
    if self.dists is None or self.dists.size != data.size ** 2:
        # warning: this is a bit hacky! cache invalidation is a bitch
        if self.rows is None or np.sum(mask) != len(self.rows):
            self.cols, self.rows = np.where(mask)
        self.dists = x_vec[self.rows] - y_vec[self.cols]
    try:
        widths = widths[self.rows]
    except TypeError:
        pass
    dens = sparse.coo_matrix((kernel(self.dists, widths, **k_kwds),
                              (self.rows, self.cols)),
                             shape=(len(x_vec), len(y_vec)))
    return dens.T.tocsr()


def array_gen(scalar):
    return lambda tt: scalar * np.ones_like(np.atleast_1d(tt))


def distance_array(x_vec, y_vec, mask=None):
    """creating a matrix with residuals (x:horizontal,y:vertical)"""
    x_vec, y_vec = np.atleast_1d(x_vec, y_vec)
    if mask is None:
        return x_vec[None, :] - y_vec[:, None]
    else:
        return np.where(mask, x_vec[None, :] - y_vec[:, None], np.nan)


def distance_array_sparse(x_vec, y_vec, mask):
    """Creates a sparse residual matrix."""
    rows, cols = np.where(mask)
    dist_sparse = sparse.coo_matrix(((x_vec[rows] - y_vec[cols]),
                                     (rows, cols)),
                                    shape=(len(x_vec), len(y_vec)))
    return dist_sparse.T.tocsr()


# @profile
def doy_distance_int(doys1, doys2, doy_width=1):
    """Distance between two times in terms of integer doys.

    doys1 : decimal doy or ndarray of decimal doys
    doys2 : decimal doy or ndarray of decimal doys
    doy_width : float
        scaling factor for distance in the doy dimension
    """
    doys1, doys2 = map(np.atleast_1d, (doys1, doys2))
    doy_distance = np.squeeze(times.doy_distance(doys1[None, :],
                                                 doys2[:, None]))
    return doy_width * doy_distance


# @profile
def hour_distance_using_doys(doys1, doys2, circ=1):
    """Distance between two times in terms of hours.

    doys1 : decimal doy or ndarray of decimal doys
    doys2 : decimal doy or ndarray of decimal doys
    circ : float
        minor circumference of torus. scales the hour distance in relation to
        the doy distance.
    """
    doys1, doys2 = map(np.atleast_1d, (doys1, doys2))
    hours1, hours2 = map(lambda doy: (doy % 1) * 24, (doys1, doys2))
    hour_distance = np.squeeze(times.hour_distance(hours1[None, :],
                                                   hours2[:, None]))
    return circ * hour_distance


# @profile
def doyhour_distance(doys1, doys2, doy_width=1, circ=1):
    """Distance between two times treating the hour as a second dimension.
    Similar but not exactly like the distance on the surface of a torus.

    Parameter
    ---------
    doys1 : decimal doy or ndarray of decimal doys
    doys2 : decimal doy or ndarray of decimal doys
    doy_width : float
        scaling factor for distance in the doy dimension
    circ : float
        minor circumference of torus. scales the hour distance in relation to
        the doy distance.
    """
    # doys1, doys2 = map(np.atleast_1d, (doys1, doys2))
    # hours1, hours2 = map(lambda doy: (doy % 1) * 24, (doys1, doys2))
    # doy_distance = np.squeeze(times.doy_distance(doys1[None, :],
    #                                              doys2[:, None]))
    # hour_distance = np.squeeze(times.hour_distance(hours1[None, :],
    #                                                hours2[:, None]))
    # return np.sqrt((doy_width * doy_distance) ** 2 +
    #                (circ * hour_distance / 24.) ** 2)
    return np.sqrt(doy_distance_int(doys1, doys2, doy_width=doy_width) ** 2 +
                   (hour_distance_using_doys(doys1, doys2, circ=circ)
                    / 24) ** 2)


def doyhour_distance_dt(dtimes1, dtimes2, circ):
    """Distance between two times treating the hour as a second dimension.
    Similar but not exactly like the distance on the surface of a torus.

    Parameter
    ---------
    dtimes1 : datetime object or ndarray of datetime objects
    dtimes2 : datetime object or ndarray of datetime objects
    circ : float
        minor circumference of torus. scales the hour distance in relation to
        the doy distance.
    """
    doys1, doys2 = map(times.datetime2doy, (dtimes1, dtimes2))
    # TODO: for speed: calculate the hours from the doys without calling
    # datetime2hour
    hours1, hours2 = map(times.datetime2hour, (dtimes1, dtimes2))
    doy_distance = times.doy_distance(doys1, doys2)
    hour_distance = times.hour_distance(hours1, hours2)
    return np.sqrt(doy_distance ** 2 + (circ * hour_distance / 24.) ** 2)


class SeasonalKDE(seasonal.Seasonal):

    def __init__(self, data, datetimes, solution=None, doy_width=15, nx=1e3,
                 fixed_pars=None, verbose=True, freibord=0, kill_leap=False):
        """
        nx is the number of data points used to estimate the quantiles.
        lower- and upper bound must be functions that accept a doy. If None
        smoothed seasonal minima and maxima of data will be used.

        Parameters
        ----------
        data : 1dim ndarray
            may contain nans.
        datetimes : sequence of datetime objects
        solution : sequence
            contains kernel widths and quantile_grid as returned from
            SeasonalKDE.fit
        doy_width :  int
            kernel width in time direction.
        nx : int
            number of datapoints for each doy to base the pdf, cdf and ppf
            estimates on
        fixed_pars : dict
            with 'l' and or 'u' mapping to a function that accepts doys and
            returns lower and upper bounds.
        verbose : boolean
        """
        # super(SeasonalKDE, self).__init__(data, datetimes, kill_leap)
        seasonal.Seasonal.__init__(self, data, datetimes, kill_leap)
        self.doy_width = doy_width
        self.nx = nx
        self.verbose = verbose
        self.freibord = freibord
        self.smooth_len = self.doy_width / self.timestep

        # those will be populated by cached properties
        self._doy_mask = self._doy_mask_dict = self._kernel_widths = None
        self._density_grid = self._x_grid = self._quantile_grid = None
        self._cdf_interp_per_day = self._ppf_interp_per_day = None
        self._data_bounds = self._lower_bound = self._upper_bound = None

        if solution is not None:
            self.solution = solution

        if fixed_pars is None:
            fixed_pars = {}

        try:
            fixed_pars["l"] = fixed_pars["lc"]
        except KeyError:
            pass
        if "l" in fixed_pars:
            self.lower_bound = fixed_pars["l"]
        try:
            fixed_pars["u"] = fixed_pars["uc"]
        except KeyError:
            pass
        if "u" in fixed_pars:
            self.upper_bound = fixed_pars["u"]

        upper = self.upper_bound(self.doys)
        upper_ii = self.data >= upper
        if np.any(upper_ii):
            self.data[upper_ii] = (np.sign(upper[upper_ii]) *
                                   np.abs(upper[upper_ii]) * (1 - 1e-3))

        lower = self.lower_bound(self.doys)
        lower_ii = self.data <= lower
        if np.any(lower_ii):
            self.data[lower_ii] = (np.sign(lower[lower_ii]) *
                                   np.abs(lower[lower_ii]) * (1 + 1e-3))

    @property
    def data_bounds(self):
        """
        Here for optimization: We do not let lower and upper bound make
        a long loop over the data for min and max resp. We do it here once and
        cache it.
        """
        if self._data_bounds is None:
            exact_doy_mask = self.doys == self.doys_unique[:, np.newaxis]
            self._data_bounds = \
                np.array([(np.nanmin(self.data[doy_ii]),
                           np.nanmax(self.data[doy_ii]))
                          for doy_ii in exact_doy_mask])
        return self._data_bounds

    def lower_bound(self, doy):
        if self._lower_bound is None:
            data_mins, data_maxs = self.data_bounds.T
            data_mins_smooth = \
                smooth(data_mins, self.smooth_len, periodic=True)
            # we have to lower the smoothed mins below the actual data mins
            diffs = data_mins_smooth - data_mins
            # ideal would be a kernel width, but we do not have those yet
            diff_thresh = ((data_maxs.mean() - data_mins.mean()) /
                           (len(self.data) / len(self.doys_unique)))
            while np.any(diffs > -diff_thresh):
                sub = np.zeros_like(diffs)
                argmax = np.argmax(diffs)
                diff_max = diffs[argmax]
                sub[argmax] = diff_max
                sub = smooth(sub, self.smooth_len, periodic=True)
                sub *= (1 + diff_thresh) * self.smooth_len / diff_max
                data_mins_smooth -= sub
                diffs = data_mins_smooth - data_mins
            self._lower_bound = data_mins_smooth
        doy = np.rint((doy - 1) / self.timestep).astype(int)
        return self._lower_bound[doy]

    def upper_bound(self, doy):
        if self._upper_bound is None:
            data_mins, data_maxs = self.data_bounds.T
            data_maxs_smooth = \
                smooth(data_maxs, self.smooth_len, periodic=True)
            # we have to lift the smoothed maxs above the actual data maxs
            diffs = data_maxs - data_maxs_smooth
            # ideal would be a kernel width, but we do not have those yet
            diff_thresh = ((data_maxs.mean() - data_mins.mean()) /
                           (len(self.data) / len(self.doys_unique)))
            while np.any(diffs > -diff_thresh):
                add = np.zeros_like(diffs)
                argmax = np.argmax(diffs)
                diff_max = diffs[argmax]
                add[argmax] = diff_max
                add = smooth(add, self.smooth_len, periodic=True)
                add *= (1 + diff_thresh) * self.smooth_len / diff_max
                data_maxs_smooth += add
                diffs = data_maxs - data_maxs_smooth
            self._upper_bound = data_maxs_smooth + self.freibord
        doy = np.rint((doy - 1) / self.timestep).astype(int)
        return self._upper_bound[doy]

    @property
    def doy_mask(self):
        """Returns a (n_unique_doys, len(data)) ndarray"""
        if self._doy_mask is None:
            doy_width, doys = self.doy_width, self.doys
            self._doy_mask = \
                np.empty((len(self.doys_unique), len(self.data)), dtype=bool)
            for doy_i, doy in enumerate(self.doys_unique):
                ii = (doys > doy - doy_width) & (doys <= doy + doy_width)
                if (doy - doy_width) < 0:
                    ii |= doys > (365. - doy_width + doy)
                if (doy + doy_width) > 365:
                    ii |= doys < (doy + doy_width - 365.)
                self._doy_mask[doy_i] = ii
        return self._doy_mask

    @property
    def doy_mask_dict(self):
        if self._doy_mask_dict is None:
            self._doy_mask_dict = {}
            for row_i, row in enumerate(self.doy_mask):
                self._doy_mask_dict[self.doys_unique[row_i]] = row
        return self._doy_mask_dict

    @property
    def x_grid(self):
        if self._x_grid is None:
            xx = np.empty((len(self.doys_unique), self.nx))
            for doy_i, doy in enumerate(self.doys_unique):
                xx[doy_i] = np.linspace(self.lower_bound(doy),
                                        self.upper_bound(doy), self.nx)
            self._x_grid = xx
        return self._x_grid

    @property
    def density_grid(self):
        if self._density_grid is None:
            densities = np.empty_like(self.x_grid)
            for doy_i, doy in enumerate(self.doys_unique):
                kernel_width = self.kernel_widths[doy_i]
                ii = self.doy_mask_dict[doy]
                x = self.x_grid[doy_i]
                densities[doy_i] = \
                    self.density_per_doy(kernel_width, self.data[ii],
                                         self.doys[ii], doy, x)
            self._density_grid = densities
            # HACK the following will normalize the densities
            self.quantile_grid
        return self._density_grid

    @property
    def quantile_grid(self):
        if self._quantile_grid is None:
            quantiles = my.cumtrapz(y=self.density_grid, x=self.x_grid,
                                    axis=1, initial=0)
            # this normalizes the quantiles
            quantiles *= (quantiles[:, -1] ** -1)[:, np.newaxis]
            # just making it sure...
            quantiles[:, 0] = 0
            quantiles[:, -1] = 1.
            self._quantile_grid = quantiles
        return self._quantile_grid

    @quantile_grid.setter
    def quantile_grid(self, grid):
        self._quantile_grid = grid

    @property
    def cdf_interp_per_day(self):
        if self._cdf_interp_per_day is None:
            self._cdf_interp_per_day = \
                [interpolate.interp1d(self.x_grid[doy_i], quantiles)
                 for doy_i, quantiles in enumerate(self.quantile_grid)]
        return self._cdf_interp_per_day

    @property
    def ppf_interp_per_day(self):
        if self._ppf_interp_per_day is None:
            self._ppf_interp_per_day = \
                [interpolate.interp1d(quantiles, self.x_grid[doy_i],
                                      kind="nearest",
                                      assume_sorted=True)
                 for doy_i, quantiles in enumerate(self.quantile_grid)]
        return self._ppf_interp_per_day

    def density_per_doy(self, kernel_width, data, doys, doy_middle,
                        eval_data=None, leave_one_out=False):
        densities = apply_kernel(kernel_width, data, eval_data)
        doy_scale = (1 - times.doy_distance(doy_middle, doys) /
                     (self.doy_width + 1)) / (self.doy_width - 1)
        densities *= doy_scale
        # we want the densities at a specific doy to integrate to one
        densities *= len(doys) / self.doy_width
        if leave_one_out:
            # sets main diagonal to 0 (exploiting the fact that ravel returns
            # a view)
            densities.ravel()[::len(data) + 1] = 0
        return np.sum(densities, axis=1) / float(len(densities) - 1)

    def sum_log_density(self, kernel_width, data, doys, doy_middle):
        """objective function to optimize kernel width with MLM leave one
        out"""
        densities = self.density_per_doy(np.abs(kernel_width), data, doys,
                                         doy_middle, leave_one_out=True)
        return -np.nansum(np.log(densities))

    def fit(self):
        """This estimates the kernel widths per doy AND returns the
        interpolating functions for cdf and ppf."""
        kernel_widths = np.empty(len(self.doys_unique))
        if self.verbose:
            progress_bar = my.ProgressBar(len(self.doys_unique))
        # if the data is very coarse, add a little noise
        data = np.copy(self.data)
        data_unique = np.unique(data)
        if len(data) / len(data_unique) > 100:
            mean_discretization = np.mean(np.diff(data_unique))
            data += np.random.randn(len(data)) / mean_discretization
            # but honour the bounds of the original data
            data[data < data_unique[0]] = data_unique[0]
            data[data > data_unique[-1]] = data_unique[-1]

        for doy_i, doy in enumerate(self.doys_unique):
            ii = self.doy_mask[doy_i]
            if doy_i == 0:
                x0 = (data[ii].max() - data[ii].min()) / 10.
            else:
                x0 = kernel_widths[doy_i - 1]
            kernel_width = \
                optimize.fmin(self.sum_log_density, x0,
                              args=(data[ii], self.doys[ii], doy), disp=False)
            kernel_widths[doy_i] = kernel_width

            if self.verbose:
                progress_bar.animate(int(doy_i))
        if self.verbose:
            print
        kernel_widths = np.abs(kernel_widths)
        kernel_widths = smooth(kernel_widths, self.doy_width, periodic=True)
        self._kernel_widths = kernel_widths
        return kernel_widths, self.quantile_grid

    def chi2_test(self, k=5):
        """Chi-square goodness-of-fit test.
        H0: The given data **x** follows **distribution** with parameters
            aquired by ``func::SeasonalDist.fit``
        To side-step complications arising from having a different
        distribution for every doy, we test whether the quantiles (which are
        deseasonalized) are uniformly distributed.

        Parameters
        ----------
        k : int
            Number of classes (bins)

        Returns
        -------
        p_value : float
        """
        quantiles = self.cdf(self.solution)
        n = len(quantiles)
        observed = np.histogram(quantiles, k)[0]
        expected = float(n) / k
        chi_test = np.sum((observed - expected) ** 2 / expected)
        # degrees of freedom:
        dof = k - 1
        return stats.chisqprob(chi_test, dof)

    @property
    def kernel_widths(self):
        if self._kernel_widths is None:
            self.fit()
        return self._kernel_widths

    @kernel_widths.setter
    def kernel_widths(self, widths):
        self._kernel_widths = widths

    @property
    def solution(self):
        return self.kernel_widths, self.quantile_grid

    @solution.setter
    def solution(self, solution):
        self.kernel_widths, self.quantile_grid = solution

    def pdf(self, solution, x, doys):
        x, doys = np.atleast_1d(x, doys)
        densities = np.empty((len(doys), len(x)))
        for doy in doys:
            doy_i = my.val2ind(self.doys_unique, doy)
            kernel_width = self.kernel_widths[doy_i]
            ii = self.doy_mask_dict[doy]
            densities[doy_i] = self.density_per_doy(kernel_width,
                                                    self.data[ii],
                                                    self.doys[ii], doy, x)
        return densities

    def cdf(self, solution=None, x=None, doys=None):
        if solution is not None:
            self.solution = solution
        if x is None:
            x = self.data
        if doys is None:
            doys = self.doys

        x, doys = np.atleast_1d(x, doys)
        quantiles = np.full_like(x, np.nan)
        for ii, (doy, x_single) in enumerate(zip(doys, x)):
            doy_i = my.val2ind(self.doys_unique, doy)
            try:
                quantiles[ii] = self.cdf_interp_per_day[doy_i](x_single)
            except ValueError, e:
                if x_single < self.lower_bound(self.doys_unique[doy_i]):
                    quantiles[ii] = np.nan
                    continue
                elif x_single > self.upper_bound(self.doys_unique[doy_i]):
                    quantiles[ii] = np.nan
                    continue
                raise e
        return quantiles

    def ppf(self, solution, quantiles, doys):
        if solution is not None:
            self.solution = solution
        quantiles, doys = np.atleast_1d(quantiles, doys)
        x = np.empty_like(quantiles)
        for ii, (doy, quantile_single) in enumerate(zip(doys, quantiles)):
            # doy_i = my.val2ind(self.doys_unique, doy)
            doy_i = int((doy - 1) / self.dt)
            x[ii] = self.ppf_interp_per_day[doy_i](quantile_single)
        return x

    def scatter_pdf(self, solution, n_sumup=24, figsize=None, title=None,
                    opacity=.25, plot_kernel_width=False, s_kwds=None):
        if s_kwds is None:
            s_kwds = dict(marker="o")
        doys = \
            self.doys_unique.repeat(self.nx).reshape(len(self.doys_unique), -1)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        ax1.contourf(doys, self.x_grid / n_sumup, self.density_grid, 15)
        plt.set_cmap("coolwarm")
#        plt.colorbar(co)
        plt.scatter(self.doys, self.data / n_sumup, facecolors=(0, 0, 0, 0),
                    edgecolors=(0, 0, 0, opacity), **s_kwds)
#        plt.plot(self.doys_unique, self.lower_bound(self.doys_unique), "r",
#                 label="lower bound")
#        plt.plot(self.doys_unique, self.upper_bound(self.doys_unique), "r",
#                 label="upper bound")
        ax1.set_ylim(self.x_grid.min() / n_sumup, self.x_grid.max() / n_sumup)

        if plot_kernel_width:
            ax2 = plt.gca().twinx()
            ax2.plot(self.doys_unique, self.kernel_widths.T.ravel(), "r-",
                     label="Kernel width")
            ax2.set_ylim(self.kernel_widths.min(), self.kernel_widths.max())

        plt.xlim(0, 366)
        plt.xticks((1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335),
                   ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug',
                    'Sep', 'Oct', 'Nov', 'Dec'), rotation=45)

#        plt.xlabel("Day of Year")
        plt.grid()
        # plt.legend()
        if title is not None:
            plt.title(title)
        # plt.tight_layout()
        return fig, ax1

    def scatter_cdf(self, solution, n_sumup=1, figsize=None, title=None):
        doys = self.doys_unique.repeat(1e3).reshape(len(self.doys_unique), 1e3)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.gca()
        co = ax1.contourf(doys, self.x_grid, self.quantile_grid, 15)
        plt.colorbar(co)
        plt.scatter(self.doys, self.data / n_sumup, marker="o",
                    facecolors=(0, 0, 0, 0), edgecolors=(0, 0, 0, .5))
#        plt.plot(self.doys_unique, self.lower_bound(self.doys_unique), "r",
#                 label="lower bound")
#        plt.plot(self.doys_unique, self.upper_bound(self.doys_unique), "r",
#                 label="upper bound")
        ax1.set_ylim(self.x_grid.min(), self.x_grid.max())
        ax2 = plt.gca().twinx()
        ax2.plot(self.doys_unique, self.kernel_widths, label="Kernel width")
        ax2.set_ylim(self.kernel_widths.min(), self.kernel_widths.max())

        plt.xlim(0, len(doys))
        plt.xlabel("Day of Year")
        plt.grid()
        plt.legend()
        if title is not None:
            plt.title(title)
        return fig


def doy_hour_fft(data, dtimes, order=4):
    """Remove yearly and daily cycle.

    Parameter
    ---------
    data : 1d array
    dtimes : 1d array of datetime objects
    order : int, optional
        number of frequencies to use

    Returns
    -------
    xx : 1d array
    fft_pars : 24d array
        fast fourier parameters for each hour.
    """
    xx = np.copy(data)
#    doys_int = times.time_part(dtimes, "%Y")
    hours_int = np.array([dtime.hour for dtime in dtimes])
    fft_pars = []
    # approximate cycles of values for a given hour
    for hour in range(24):
        ii = hours_int == hour
        values = data[ii]
        fft_par = np.fft.rfft(values)
        pars_below = np.argsort(np.abs(fft_par))[:len(fft_par) - order - 1]
        par = np.copy(fft_par)
        par[pars_below] = 0
#        par = fft_par
        fft_pars += [par]
        xx[ii] -= np.fft.irfft(par, len(values))
    return xx, np.array(fft_pars)


def fft2par(fft_pars, doys, ifft_func=None, period_length=None,
            lower_bound=None):
    """Converts from the frequency into the time domain.
    The output is repeated for as many periods underlying the doys (think of
    years).

    Parameter
    ---------
    fft_pars : ndarray
        As returned from numpy.fft.fft (or its siblings).
    doys : float 1d array
        Doys associated with the output.
    ifft_func : callable or None, optional
        Function to use for frequency-time-domain transformation.
        If None, np.fft.irfft will be used.
    period_length : int or None, optional
        Length of period.
    lower_bound : float or None, optional
        Smallest desired value of trans.

    Returns
    -------
    trans : 1d array with len = len(doys)

    """
    if ifft_func is None:
        ifft_func = np.fft.irfft
    if period_length is None:
        period_length = int(round(365. / (doys[1] - doys[0])))

    # the parameters do not only be as big as doys, they also have to
    # have the right number of periods!
    # fft_pars consist of the first order important parameters, the rest have
    # to be zero-padded
    fft_pars_pad = np.zeros(len(doys), dtype=complex)
    if fft_pars.dtype == complex:
        order = fft_pars.size
        fft_pars_pad[:order] = fft_pars
    else:
        order = len(fft_pars) / 2
        fft_pars_pad[:order].real = fft_pars[:order]
        fft_pars_pad[:order].imag = fft_pars[order:]

    n_periods = np.ceil((len(doys) + doys[0]) / 365.)
    trans = np.array(n_periods *
                     [ifft_func(fft_pars_pad, period_length)]).ravel()
    start_i = int(doys[0] / period_length)
    trans = trans[start_i:start_i + len(doys)]
    if lower_bound is not None:
        trans[trans < lower_bound] = lower_bound
    return trans


class SeasonalHourlyKDE(SeasonalKDE, seasonal.Torus):

    """Estimates kernel densities for time series exhibiting seasonalities
    in daily cycles."""

    def __init__(self, data, dtimes, doy_width=5, hour_neighbors=4, *args,
                 **kwds):
        """see SeasonalKDE.__init__"""
        seasonal.Torus.__init__(self, hour_neighbors)
        super(SeasonalHourlyKDE, self).__init__(data, dtimes,
                                                doy_width=doy_width,
                                                kill_leap=True,
                                                *args, **kwds)

        self.smooth_len = hour_neighbors * 5
        # 1. make a 3d array with (hours, doys, years)-shape
        # 2. stack three of those 3d arrays with a daily
        #    time-shift on top of each other so that we have three
        #    subsequent days in the vertical dimension:
        #    (hours, doys, n_years) shape
        # 3. pass subslices for each doy to a kde evaluation function
        # self.torus = self._fill_torus()

    def sum_log_density(self, kernel_width, data):
        densities = self.density_per_doy(kernel_width, data,
                                         self.hour_neighbors,
                                         leave_one_out=True)
        if NE:
            return -np.nansum(ne.evaluate("log(densities)"))
        else:
            return -np.nansum(np.log(densities))

    def density_per_doy(self, kernel_width, data, hour_width,
                        eval_data=None, leave_one_out=False):
        densities = apply_2d_kernel(kernel_width, data, hour_width,
                                    eval_points=eval_data)
        if leave_one_out:
            # sets main diagonal to 0 (exploiting the fact that ravel returns
            # a view)
            densities.ravel()[::len(data) + 1] = 0
        dens = np.nansum(densities, axis=1) / float(len(densities) - 1)
        return dens

    @property
    def density_grid(self):
        if self._density_grid is None:
            densities = np.empty_like(self.x_grid)
            for doy_i, doy in enumerate(self.doys_unique):
                hour_slice, doy_slice = self._torus_slice(doy)
                data = self.torus[hour_slice, doy_slice]
                kernel_hour, kernel_doy = self._unpadded_index(doy)
                kernel_width = self.kernel_widths[kernel_hour, kernel_doy]
                x = self.x_grid[doy_i]
                densities[doy_i] = \
                    self.density_per_doy(kernel_width, data,
                                         self.hour_neighbors, eval_data=x)
            self._density_grid = densities
            # HACK the following will normalize the densities
            self.quantile_grid
        return self._density_grid

    @property
    def data_bounds(self):
        # hack to get daily cycles in lower_bound and upper_bound
        # self.smooth_len = 1
        data_unpad = self._unpad_torus(self.torus)
        return np.array((np.nanmin(data_unpad, axis=2).T.ravel(),
                         np.nanmax(data_unpad, axis=2).T.ravel())).T

    def lower_bound(self, doy):
        if self._lower_bound is None:
            data_unpad = self._unpad_torus(self.torus)
            mins = np.nanmin(data_unpad, axis=2)
            # smooth on an hourly basis
            mins_smooth = \
                np.array([smooth(row, self.smooth_len, periodic=True)
                          for row in mins])
            # we have to lower the smoothed mins below the actual data mins
            max_diff = np.max(mins_smooth - mins)
            diff_thresh = self.scotts_rule()
            self._lower_bound = mins_smooth - (max_diff + 5 * diff_thresh)
        hour_index, doy_index = self._unpadded_index(doy)
        return self._lower_bound[hour_index, doy_index]

    def upper_bound(self, doy):
        if self._upper_bound is None:
            data_unpad = self._unpad_torus(self.torus)
            maxs = np.nanmax(data_unpad, axis=2)
            # smooth on an hourly basis
            maxs_smooth = \
                np.array([smooth(row, self.smooth_len, periodic=True)
                          for row in maxs])
            # we have to upper the smoothed maxs below the actual data maxs
            max_diff = np.max(maxs_smooth - maxs)
            diff_thresh = self.scotts_rule()
            self._upper_bound = maxs_smooth + max_diff + 5 * diff_thresh
        hour_index, doy_index = self._unpadded_index(doy)
        return self._upper_bound[hour_index, doy_index]

    def fit(self, thumb=False, order=4):
        """
        Parameter
        ---------
        thumb : boolean or callable, optional
            Use rule of thumb instead of leave-one-out maximum
            likelihood estimation.
            Callable is executed with (n_data, n_dim).
        """
        hpd = self.hours_per_day

        if thumb:
            if not callable(thumb):
                thumb = self.scotts_rule
            kernel_width = thumb()
            self.kernel_widths = np.full((hpd, 365), kernel_width)
            return self.kernel_widths, self.quantile_grid

        # evaluate kernel widths in the doy-hour domain
        kernel_widths = np.empty((hpd, 365))
        x0 = self.scotts_rule()
        if self.verbose:
            progress_bar = my.ProgressBar(hpd * 365)
        for hour in range(hpd):
            hour_torus = hour + self.hour_neighbors
            hour_slice = slice(hour_torus - self.hour_neighbors,
                               hour_torus + self.hour_neighbors + 1)
            for doy in range(365):
                doy_torus = doy + self.doy_width
                doy_slice = slice(doy_torus - self.doy_width,
                                  doy_torus + self.doy_width + 1)
                data = self.torus[hour_slice, doy_slice]
                kernel_width = \
                    optimize.minimize(
                        self.sum_log_density, x0,
                        args=(data,),
                        # options={"disp": True}
                    ).x
                kernel_widths[hour, doy] = kernel_width
                x0 = kernel_width
                if self.verbose:
                    progress_bar.animate(doy + hour * 365)
        self.kernel_width = self._kernel_widths = kernel_widths
        # self.kernel_widths = self.smooth_torus(kernel_widths, order=order)
        return self.kernel_widths, self.quantile_grid

    @property
    def solution(self):
        return self.kernel_widths, self.quantile_grid

    @solution.setter
    def solution(self, solution):
        self.kernel_widths, self.quantile_grid = solution

    def scotts_rule(self, n_data=None, n_dim=3):
        """Scott's rule of thumb for kernel bandwidth."""
        if n_data is None:
            # for getting a first guess we need the number of data in
            # each individual kde
            n_data = ((self.hour_neighbors * 2 + 1) *
                      (self.doy_width * 2 + 1) *
                      self.n_years)
        return n_data ** (-1. / (n_dim + 4))

    def pdf(self, solution, x, doys):
        if solution is not None:
            self.solution = solution
        if x is None:
            x = self.data
        if doys is None:
            doys = self.doys
        x, doys = np.atleast_1d(x, doys)
        densities = np.empty((len(doys), len(x)))
        for ii, doy in enumerate(doys):
            hour_sl, doy_sl = self._torus_index(doy)
            hour_i, doy_i = self._torus_index(doy)
            kernel_width = self.kernel_widths[hour_i, doy_i]
            densities[ii] = self.density_per_doy(kernel_width,
                                                 self.data[hour_sl, doy_sl],
                                                 eval_data=x)
        return densities


if __name__ == "__main__":
    import vg
    import config_konstanz
    vg.vg_base.conf = config_konstanz
    met_vg = vg.VG(("theta", "ILWR", "rh", "u", "v"))
    # from vg.meteo import avrwind
    # data = avrwind.component2angle(met_vg.met["u"], met_vg.met["v"])[1]
    data = met_vg.met["theta"][:3 * 8760]
    dtimes = met_vg.times_orig[:3 * 8760]
    kde = SeasonalHourlyKDE(data, dtimes, doy_width=10, verbose=True)
    solution = kernel_widths, quantile_grid = kde.fit()

    # import vg
    import config_konstanz
    # vg.conf = config_konstanz
    # dtimes_, met = vg.read_met()
    # dist = SeasonalHourlyKDE(met["Qsw"], dtimes_, verbose=True)
    # dist.fit()
    # dist.scatter_pdf(None)
    # plt.show()
#    quantiles = dist.cdf(solution, x=u, doys=vg.times.datetime2doy(dtimes))
