import sys
import os
import copy
import random
import importlib
import pprint
import multiprocessing
from collections import namedtuple, deque, defaultdict
from abc import ABCMeta, abstractmethod, abstractproperty
import numpy as np
from scipy import stats, optimize
import numexpr as ne
import matplotlib.pyplot as plt
from matplotlib import animation, gridspec
from vg import smoothing
from vg import helpers as my

copy_rules = lambda rules: [rule.__class__(**{key: getattr(rule, key).copy()
                                              for key in rule._fields})
                            for rule in rules]
# to infinity ... and beyond!
inf_ = 1e16


def rules_dump(rules, filepath="rules.py"):
    content = \
        """# this file was generated by fuzzy.rules_dump and might be
           # overwritten by it later!
           from collections import namedtuple
           from numpy import array
           import vg
           rule = namedtuple("rule", "{args}")
           rules = \\
           {rules}"""
    content = "\n".join(line.lstrip() for line in content.split("\n"))
    content = content.format(args=" ".join(rules[0]._fields),
                             rules=pprint.pformat(rules))
    with open(filepath, "w") as rule_file:
        rule_file.write(content)


def rules_load(filepath="rules.py"):
    if filepath.endswith(".py"):
        filepath = filepath[:-3]
    # unfortunately we have to move to the path to import the module
    with my.chdir(os.path.dirname(filepath)):
        module_name = os.path.basename(filepath)
        rules_module = importlib.import_module(module_name)
    reload(rules_module)
    return rules_module.rules


def plot_rules(rules, var_names_in=None, var_names_out=None, rules_per_plot=5,
               data_in=None):
    K_in = len(rules[0].ifs)
    K_out = len(rules[0].thens)
    K_max = max(K_in, K_out)
    var_names_in = " " * K_in if var_names_in is None else var_names_in
    var_names_out = " " * K_out if var_names_out is None else var_names_out

    # some configuration setting
    fields = rules[0]._fields
    do_smooth = "agg_lens" in fields
    do_shift = "shifts" in fields
    do_thresh = "dof_thresh" in fields
    do_agg = "agg_funcs" in fields

    if data_in is not None:
        # ensure we have access to the membership values
        membs = rule_system_response(data_in, rules, return_membs=True)
        membs_mean = membs.mean(axis=2)
        memb_width = lambda x: 5 * (.1 + (x - membs_mean.min()) /
                                    (membs_mean.max() - membs_mean.min()))

    thens = np.array([rule.thens for rule in rules])
    then_maxs = np.max(thens, axis=0)
    then_mins = np.min(thens, axis=0)
    shift_agg_min = np.min([-(rule.agg_lens if do_agg else 1) -
                            (rule.shifts if do_shift else 1)
                            for rule in rules])
    mins = defaultdict(lambda: np.inf)
    maxs = defaultdict(lambda:-np.inf)
    cols = plt.cm.jet(np.linspace(0, 1, rules_per_plot))
    figs = []
    for rule_i, rule in enumerate(rules):
        if rule_i % rules_per_plot == 0:
            fig, axes = plt.subplots(nrows=K_max, ncols=3, squeeze=True,
                                     figsize=(10, 6))
            plt.subplots_adjust(hspace=.4)
            figs += [fig]
        color = cols[rule_i % rules_per_plot]
        for var_i in range(K_max):
            if var_i < K_in:
                ifs = np.squeeze(rule.ifs[var_i])
                if data_in is None or rule_i == 0:
                    # do not make the first rule fat, because it applies
                    # anywhere and is fat per default
                    linewidth = 1
                else:
                    linewidth = memb_width(membs_mean[rule_i, var_i])
                axes[var_i, 1].plot(ifs, [0, 1, 0],
                                    color=color, linewidth=linewidth)
                if do_thresh:
                    # plot a bar at the height of the threshold
                    thresh = rule.dof_thresh
                    left = (ifs[1] - ifs[0]) * thresh + ifs[0]
                    right = ifs[2] - (ifs[2] - ifs[1]) * thresh
                    axes[var_i, 1].plot((left, right), (thresh, thresh),
                                        color=color, linewidth=linewidth)
                if ifs[0] < mins[var_i] and ifs[0] > -inf_ / 1e3:
                    mins[var_i] = ifs[0]
                if ifs[2] > maxs[var_i] and ifs[2] < inf_ / 1e3:
                    maxs[var_i] = ifs[2]
                axes[var_i, 1].set_yticklabels("")
                axes[var_i, 1].set_title(var_names_in[var_i])
                axes[var_i, 1].set_ylabel("membership")
                axes[var_i, 1].set_ylim(0, 1.1)

                agg_len = 1 if not do_smooth else rule.agg_lens[var_i]
                shift = rule.shifts[var_i] if do_shift else 0
                y_pos = rule_i % rules_per_plot + 1
                axes[var_i, 0].plot([-agg_len - shift, -shift], 2 * [y_pos],
                                    color=color, marker="|")
                text = str(agg_len)
                if do_agg:
                    text += " " + rule.agg_funcs[var_i].func_name
                axes[var_i, 0].text(min(-.5 * agg_len - shift,
                                        - len(text) * 2),
                                    y_pos + .125, text, color=color,
                                    horizontalalignment="center")
                axes[var_i, 0].set_ylim(0, y_pos + 1)
                axes[var_i, 0].set_xlim(shift_agg_min, 0)
                axes[var_i, 0].grid(True)
                axes[var_i, 0].set_ylabel("rule no.")
                axes[var_i, 0].set_yticklabels("")
                if var_i == 0:
                    axes[var_i, 0].set_title("agg. lengths/shifts")
            else:
                axes[var_i, 0].axis("off")
                axes[var_i, 1].axis("off")
            if var_i < K_out:
                axes[var_i, 2].scatter(rule.thens[var_i], .5, facecolor=color,
                                       edgecolor=(0, 0, 0, 0), s=50)
                axes[var_i, 2].set_yticklabels("")
                axes[var_i, 2].set_title(var_names_out[var_i])
                axes[var_i, 2].set_xlim(then_mins[var_i] -
                                        .1 * np.abs(then_mins[var_i]),
                                        1.1 * then_maxs[var_i])
            else:
                axes[var_i, 2].axis("off")

        for var_i in range(K_in):
            axes[var_i, 1].set_xlim(mins[var_i], maxs[var_i])

    return figs


def animate(data_in, data_out, cv_relation=2. / 3, dtimes=None, out_names=None,
            rules_filepath="rules_dump.py", rules=None):
    """Show the current best solution performance.
    Call from another process while learn is running. learn dumps the best
    rule system found so far every M iteration steps.

    data_in : (K_in, T) ndarray
        Input data. K_in number of input variables, T number of time steps.
    data_out : (K_out, T) ndarray
        Output data. K_out number of output variables.
    cv_relation : [0, 1] float, optional
        Percentage of data used as calibration. Set to 1 if you do not have a
        validation period. Calibration values are taken from the beginning of
        the data set.
    dtimes : (T,) ndarray of datetime objects, optional
        datetime of data points in data_in and data_out.
    out_names : sequence of str, optional
        Names of output variables for ylabels.
    rules_filepath : str, optional
        Filepath where learn dumps the current best solution.
    rules : None or list of rules, optional
        Will be used for the first frame.
    """
    cv_relation = float(cv_relation)
    K_out = data_out.shape[0]
    if out_names is None:
        out_names = " " * K_out

    if rules is None:
        response = np.zeros_like(data_out)
    else:
        response = rule_system_response(data_in, rules, cache=False)

    T = data_in.shape[1]
    if dtimes is None:
        dtimes = np.arange(T)
    if cv_relation < 1.:
        validation = True
        cutoff = T * cv_relation
        responses = response[:, :cutoff], response[:, cutoff:]
        data_outs = data_out[:, :cutoff], data_out[:, cutoff:]
        dtimess = dtimes[:cutoff], dtimes[cutoff:]
    else:
        validation = False
        responses, data_outs, dtimess = (response,), (data_out,), (dtimes,)

    def title_text(dat, resp, rs=None):
        if rs is None:
            finite_ii = np.isfinite(dat)
            rs = stats.linregress(dat[finite_ii], resp[finite_ii])[2]
        nash = my.nash_sutcliff(resp, dat)
        return "$R^2$: {:>8.3f}\nnash: {:>6.3f}".format(rs, nash)

    fig = plt.figure(figsize=(12, 8))
    width_ratios = [cv_relation.as_integer_ratio()[0],
                    (1 - cv_relation).as_integer_ratio()[0]]
    gs = gridspec.GridSpec(2 * K_out, len(responses),
                           width_ratios=width_ratios)
    ax = np.empty((2 * K_out, len(responses)), dtype=object)
    for i in range(2 * K_out):
        for j in range(len(responses)):
            ax[i, j] = plt.subplot(gs[i, j])

    if len(responses) == 1 and validation:
        ax = np.array(ax)[:, None]
    lines, scatter, texts, regr = [], [], [], []
    for out_i in range(K_out):
        for col_i, (resps, dats, time) \
            in enumerate(zip(responses, data_outs, dtimess)):

            dat, resp = dats[out_i], resps[out_i]
            # # time-line plots
            row_i = out_i * 2
            ax_cur = ax[row_i, col_i]
            ax_cur.plot(time, dat.T, label="observation", color=(.8, .8, .8))
            lines += [ax_cur.plot(time, resp.T, label="fuzzy",
                                  # sir lancelot's favourite colour
                                  color=(0, 0, 1, .5))[0]]
            ax_cur.set_ylabel(out_names[out_i])
            if validation and row_i == 0:
                if col_i == 0:
                    ax_cur.set_title("Calibration")
                elif col_i == 1:
                    ax_cur.set_title("Validation")

            # # scatter plots
            row_i += 1
            ax_cur = ax[row_i, col_i]
            scatter += [ax_cur.scatter(dat.T, resp.T,
                                       marker="o", facecolor=(0, 0, 0, 0),
                                       edgecolor=(0, 0, 0, .5))]
            # x=y line
            minmax = (0.9 * np.nanmin(data_out[out_i]),
                      1.1 * np.nanmax(data_out[out_i]))
            ax_cur.plot(minmax, minmax, color="black")
            ax_cur.set_ylim(*minmax)
            ax_cur.set_xlim(*minmax)
            # regression line
            finite_ii = np.isfinite(dat)
            b, a, rs = stats.linregress(dat[finite_ii], resp[finite_ii])[:3]
            regr += [ax_cur.plot(minmax, a + b * np.array(minmax),
                                 "--", color="grey")[0]]
            ax_cur.set_xlabel("observation")
            texts += [ax_cur.text(1.05, .5,
                                  title_text(dat, resp, rs),
                                  transform=ax_cur.transAxes)]
            if col_i == 0:
                ax_cur.set_ylabel("fuzzy")
            ax_cur.set_aspect("equal")
    for ax_ in ax.ravel():
        ax_.grid(True)
    if K_out == 1:
        texts += [fig.suptitle(title_text(data_out, response))]

    def update_response(response, lines, scatter, regr, texts):
        if validation:
            responses = response[:, :cutoff], response[:, cutoff:]
        else:
            responses = (response,)
        for out_i in range(K_out):
            for col_i in range(len(responses)):
                i = out_i * len(responses) + col_i
                dat = data_outs[col_i][out_i]
                resp = responses[col_i][out_i]

                lines[i].set_ydata(resp.T)
                scatter[i].set_offsets(np.array([dat.T, resp.T]).T)
                # regression line
                finite_ii = np.isfinite(dat)
                b, a, rs = stats.linregress(dat[finite_ii],
                                            resp[finite_ii])[:3]
                minmax = np.array((0.9 * np.nanmin(data_out[out_i]),
                                   1.1 * np.nanmax(data_out[out_i])))
                regr[i].set_ydata(a + b * minmax)
                texts[i].set_text(title_text(dat, resp, rs))
        if K_out == 1:
            texts[-1].set_text(title_text(data_out, response))
        return list(my.flatten(lines + scatter + texts))

    def data_gen():
        dummy_response = np.zeros_like(data_out)
        while True:
            try:
                rules = rules_load(rules_filepath)
                yield rule_system_response(data_in, rules)
            except (ImportError, TypeError, IndexError):
                yield dummy_response

    ani = animation.FuncAnimation(fig, update_response, data_gen,
                                  fargs=(lines, scatter, regr, texts),
                                  interval=10)
    plt.show()
    return ani


class RulePart(object):
    __metaclass__ = ABCMeta
    # to signify to IncrementalResponse which variable we changed
    var_i = None

    @abstractproperty
    def rule0(self):
        """For the first anything-goes, neutral rule."""
        pass

    @abstractproperty
    def new(self):
        """Will be accessed when generating a new random rule."""
        pass

    @abstractmethod
    def change(self, rule):
        """Will be accessed when changing an existing rule."""
        pass


class If(RulePart):
    def __init__(self, data_in, n_candidate_rules):
        self.data_in = data_in
        self.n_candidate_rules = n_candidate_rules
        self.K_in = data_in.shape[0]
        # we assume stdn distributed time series. therefore we have to
        # construct only one set of triangles that will is applicable to all
        # variables shape (K_in, n_candidate_ triangles, 3)  third dimension
        # holds lower, central, upper limit.
        triangles = np.empty((self.K_in, n_candidate_rules + 1, 3))
        for var_i, var in enumerate(data_in):
            boundaries = np.histogram(var, n_candidate_rules - 1)[1]
            triangles[var_i, 1, 0] = -inf_
            triangles[var_i, 2:, 0] = boundaries[:-1]
            triangles[var_i, 1:, 1] = boundaries
            triangles[var_i, 1:-1, 2] = boundaries[1:]
            triangles[var_i, -1, 2] = inf_
            # ignorance is bliss
            # we need this to be able to ignore a variable
            triangles[var_i, 0] = np.array([-inf_, 0, inf_])
        self.triangles = triangles

        self.in_mins = np.min(data_in, axis=1)
        self.in_maxs = np.max(data_in, axis=1)

    @property
    def rule0(self):
        return self.triangles[:, 0][:, None, :].copy()

    @property
    def new(self):
        ifs = self.triangles[:, 0]
        var_i = np.random.randint(0, self.K_in)
        ifs[var_i] = \
            self.triangles[var_i, np.random.randint(self.n_candidate_rules)]
        return ifs[:, None, :].copy()

    def change(self, rule):
        var_i = self.var_i = np.random.randint(self.K_in)
        rand_range = lambda a, b: (b - a) * np.random.random() + a
        triangle = np.squeeze(rule.ifs[var_i])
        tri_i = random.choice(range(3))
        if tri_i == 0:
            if triangle[0] == -inf_:
                triangle[0] = rand_range(-inf_, triangle[1])
            else:
                triangle[0] = rand_range(self.in_mins[var_i], triangle[1])
        elif tri_i == 1:
            dat = self.data_in[var_i]
            pool = dat[(dat > triangle[0]) & (dat <= triangle[2])]
            if len(pool) > 0:
                triangle[1] = np.random.choice(pool)
            else:
                candidate = np.random.randint(self.n_candidate_rules)
                triangle = self.triangles[var_i, 1 + candidate]
        elif tri_i == 2:
            if triangle[2] == inf_:
                triangle[2] = rand_range(triangle[1], inf_)
            else:
                triangle[2] = rand_range(triangle[1], self.in_maxs[var_i])
        rule.ifs[var_i] = triangle
        return rule


class Then(RulePart):
    def __init__(self, data_out, default_out=None):
        self.data_out = data_out
        self.default_out = default_out
        self.K_out, self.N = data_out.shape
        if default_out is not None:
            assert len(np.atleast_1d(default_out)) == self.K_out, \
                   "default_out must have as many elemtents as there are "\
                   "output variables"
        self.out_means = np.nanmean(data_out, axis=1)
        self.out_stds = np.nanstd(data_out, axis=1)
        # we have to avoid using an output that contains a nan later in
        # self.new!
        self.valid_cols = np.where(np.isfinite(np.sum(data_out, axis=0)))[0]

    @property
    def rule0(self):
        if self.default_out is None:
            return np.nanmean(self.data_out, axis=1)
        else:
            return np.atleast_1d(self.default_out)

    @property
    def new(self):
        # allow thens that are more extreme than the data
        thens = self.data_out[:, random.choice(self.valid_cols)].copy()
        return 2.5 * (thens - self.out_means) + self.out_means

    def change(self, rule):
        var_i = np.random.randint(self.K_out)
        rule.thens[var_i] += .1 * self.out_stds[var_i] * np.random.randn()
        return rule


class AggLen(RulePart):
    def __init__(self, K_in, max_agg_len):
        self.K_in = K_in
        self.max_agg_len = max_agg_len

    @property
    def rule0(self):
        return np.ones(self.K_in, dtype=int)

    @property
    def new(self):
        return np.random.randint(1, self.max_agg_len + 1, self.K_in)

    def change(self, rule):
        var_i = self.var_i = np.random.randint(self.K_in)
        agg_len = rule.agg_lens[var_i]
        if agg_len == 1:
            agg_len = 2
        elif agg_len == self.max_agg_len:
            agg_len -= 1
        else:
            agg_len += random.choice((-1, 1))
        rule.agg_lens[var_i] = agg_len if agg_len >= 1 else 1
        return rule


class Shift(RulePart):
    def __init__(self, K_in, max_shift):
        self.K_in = K_in
        self.max_shift = max_shift

    @property
    def rule0(self):
        return np.zeros(self.K_in, dtype=int)

    @property
    def new(self):
        return np.random.randint(0, self.max_shift, self.K_in)

    def change(self, rule):
        var_i = self.var_i = np.random.randint(self.K_in)
        shift = rule.shifts[var_i]
        if shift == 0:
            shift = 1
        elif shift == self.max_shift:
            shift -= 1
        else:
            shift += random.choice((-1, 1))
        rule.shifts[var_i] = shift if shift >= 0 else 0
        return rule


class DofThresh(RulePart):
    @property
    def rule0(self):
        return np.array([0.])

    @property
    def new(self):
        return np.array([np.random.random()])

    def change(self, rule):
        rule.dof_thresh[:] += .1 * np.random.random()
        return rule


class AggFunc(RulePart):
    def __init__(self, K_in, agg_funcs):
        self.K_in = K_in
        self.agg_funcs = np.asarray(agg_funcs)

    @property
    def rule0(self):
        return np.asarray(self.K_in * [smoothing.smooth])

    @property
    def new(self):
        return np.asarray([random.choice(self.agg_funcs)
                           for _ in  range(self.K_in)])

    def change(self, rule):
        var_i = self.var_i = np.random.randint(self.K_in)
        old_func = rule.agg_funcs[var_i]
        rule.agg_funcs[var_i] = random.choice(list(set(self.agg_funcs) -
                                                   set([old_func])))
        return rule


def clean_rules(rules, data_in):
    # unify rules that are independent of input variables
    dofs = rule_system_response(data_in, rules, return_dofs=True)
    dof_means = dofs.mean(axis=1)
    blind_i = [rule_i for rule_i in range(1, len(rules))
               if np.isclose(dof_means[rule_i], 1., atol=.1)]
    blind_rules = [rule for rule_i, rule in enumerate(rules)
                   if rule_i in blind_i]
    if len(blind_rules) > 0:
        thens = np.sum([rule.thens for rule in blind_rules], axis=0)
        # as this is a named tuple, we cannot change the "attribute", but we
        # can change the content of the attribute. i.e. without the [:], this
        # throws an exception
        rules[0].thens[:] = (rules[0].thens + thens) / (1 + len(blind_rules))
        rules = [rule for rule_i, rule in enumerate(rules)
                 if rule_i not in blind_i]
    return rules


def greedy_then_opt(rules, data_in, data_out):
    """Do a greedy, dumb optimization of the then parts."""
    K_out = len(rules[0].thens)

    def set_thens(params):
        params = np.asarray(params).reshape((-1, K_out))
        rules_new = copy_rules(rules)
        for rule_i in range(len(rules)):
            rules_new[rule_i].thens[:] = params[rule_i]
        return rules_new

    def obj_func(params):
        new_rules = set_thens(params)
        response = rule_system_response(data_in, new_rules, cache=True)
        return np.nansum((data_out - response) ** 2)

    x0 = [rule.thens for rule in rules]
    params = optimize.minimize(obj_func, x0).x
    return set_thens(params)


def unbias(rules, incr, data_out, debug=False):
    """Optimizes the thens of rule 0 which applies for every input and thus
    can be used to set the mean of the output."""
    rules = copy_rules(rules)
    out_means = np.nanmean(data_out, axis=1)
    dofs = incr.state.dofs.sum(axis=0)
    response = np.where(dofs > 0, incr.state.resps.sum(axis=0) / dofs, 0)
    resp_means = np.nanmean(response, axis=1)
    mean_diff = out_means - resp_means

    rules[0].thens[:] += mean_diff / np.mean(dofs ** -1)
    # update the state of incr by calling it with information about the
    # changes
    response = incr.response(rules, rule_i=0, rule_part="thens", var_i=None)
    if debug:
        resp = rule_system_response(incr.data, rules)
        assert np.all(np.isclose(resp, response))
    return rules


def smooth(data, rule, var_i):
    do_agg = "agg_funcs" in rule._fields
    agg_len = rule.agg_lens[var_i]
    smooth_func = rule.agg_funcs[var_i] if do_agg else smoothing.smooth
    if hasattr(smooth, "cache"):
        cache_key = \
            " ".join(str(i) for i in (var_i, agg_len, smooth_func.func_name))
        if cache_key not in smooth.cache:
            smooth.cache[cache_key] = smoothed = \
                smooth_func(data[var_i], agg_len, window_function="flat",
                            l_value=0, r_value=0, no_future=True)
            return smoothed
        else:
            return smooth.cache[cache_key]
    else:
        return smooth_func(data[var_i], agg_len, window_function="flat",
                           l_value=0, r_value=0, no_future=True)


def rule_system_response(data, rules, return_smoothed=False,
                         return_dofs=False, return_membs=False,
                         return_resps=False, last_only=False, cache=False):
    """

    The return-parameters are used to build up the information for the
    incremental version.
    """

    K_in, T = data.shape
    if last_only:
        T = 1
    K_out = len(rules[0].thens)
    n_rules = len(rules)
    responses = np.empty((n_rules, K_out, T))
    dofs = np.zeros(T)

    if cache and not hasattr(smooth, "cache"):
        smooth.cache = {}

    if return_smoothed:
        all_smoothed = np.empty((n_rules, K_in, T))
    if return_dofs:
        all_dofs = np.empty((n_rules, T))
    if return_membs:
        all_membs = np.empty((n_rules, K_in, T))

    # some configuration setting
    fields = rules[0]._fields
    do_smooth = "agg_lens" in fields
    do_shift = "shifts" in fields
    do_thresh = "dof_thresh" in fields

    for rule_i, rule in enumerate(rules):
        # K x T shape
        smoothed = data.copy()
        if do_smooth or do_shift:
            for var_i in range(K_in):
                if do_smooth and rule.agg_lens[var_i] > 1:
                    smoothed[var_i] = smooth(data, rule, var_i)
                else:
                    smoothed[var_i] = data[var_i]
                if do_shift and rule.shifts[var_i] > 0:
                    shift = rule.shifts[var_i]
                    smoothed[var_i, shift:] = smoothed[var_i, :-shift]
                    # move it out of scope to get membership == 0
                    smoothed[var_i, :shift] = inf_ + 1

        # K x T shape
        _ifs_left = rule.ifs[..., 0]
        _ifs_center = rule.ifs[..., 1]
        _ifs_right = rule.ifs[..., 2]
        if last_only:
            smoothed = smoothed[:, -1:]
        memberships = ne.evaluate(
            "where(smoothed < _ifs_center,"
                  "(smoothed - _ifs_left) / (_ifs_center - _ifs_left),"
                  "(_ifs_right - smoothed) / (_ifs_right - _ifs_center))")
        memberships[(memberships < 0) | np.isnan(memberships)] = 0.

        # T shape
        dof = np.product(memberships, axis=0)
        if do_thresh:
            dof = np.where(dof > rule.dof_thresh, dof, 0.)

        # K x T shape
        responses[rule_i] = dof * rule.thens[:, None]
        # T shape
        dofs += dof

        if return_smoothed:
            all_smoothed[rule_i] = smoothed
        if return_dofs:
            all_dofs[rule_i] = dof
        if return_membs:
            all_membs[rule_i] = memberships

    if return_smoothed:
        return all_smoothed
    if return_dofs:
        return all_dofs
    if return_membs:
        return all_membs
    if return_resps:
        return responses
    if do_thresh:
        # looks more complicated than it is, but the extra milage is to avoid
        # RuntimeWarning obstacles.
        ret = responses.sum(axis=0)
        valid_dofs = dofs > 0
        ret[:, ~valid_dofs] = 0
        ret[:, valid_dofs] /= dofs[valid_dofs]
        return ret
    return responses.sum(axis=0) / dofs


class AttrDict(dict):
    """Stolen from
    http://stackoverflow.com/questions/4984647/
    accessing-dict-keys-like-an-attribute-in-python"""
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        return AttrDict(**{key:
                           copy_rules(val)
                           if key == "rules"
                           else copy.copy(val)
                           for key, val in self.items()})


class IncrementalResponse(object):
    """Provides a method to calculate the rule system response incrementally,
    i.e. without re-calculating everything that does not change."""
    def __init__(self, data, rules):
        self.data = data
        self.K_in, self.T = data.shape
        # some configuration setting
        fields = rules[0]._fields
        self.do_smooth = "agg_lens" in fields
        self.do_shift = "shifts" in fields
        self.do_thresh = "dof_thresh" in fields
        self.old_state = self.state = self._reset(rules)
        self.K_out = self.state.resps.shape[1]

    def _reset(self, rules):
        """Give up and call mommy (which is rule_system_response)."""
        state = AttrDict(smoothed=rule_system_response(self.data, rules,
                                                       return_smoothed=True),
                         dofs=rule_system_response(self.data, rules,
                                                   return_dofs=True),
                         membs=rule_system_response(self.data, rules,
                                                    return_membs=True),
                         resps=rule_system_response(self.data, rules,
                                                    return_resps=True),
                         rules=rules)
        # here we look up which rule and which part of it will change
        self.rule_i = self.rule_part = self.var_i = None
        # set the smooth cache (overwriting it is ok)
        smooth.cache = {}
        return state

    def response(self, rules=None, rule_i=None, rule_part=None, var_i=None):
        self.old_state = self.state.copy()
        if rules is None:
            rules = self.state.rules

        if len(rules) > len(self.state.rules):
            # we have to make room for a new rule
            self.state.smoothed = np.vstack((self.state.smoothed,
                                             self.data[None, ...]))
            self.state.dofs = np.vstack((self.state.dofs,
                                         np.empty((1, self.T))))
            self.state.resps = np.vstack((self.state.resps,
                                          np.empty((1, self.K_out, self.T))))
            rule_i = len(rules) - 1
            rule_part = "all"
            self.var_i = None
            self.state.rules = rules

        elif len(rules) < len(self.state.rules):
            # get rid off a rule
            self.state.smoothed = np.delete(self.state.smoothed, self.rule_i,
                                            0)
            self.state.dofs = np.delete(self.state.dofs, self.rule_i, 0)
            self.state.resps = np.delete(self.state.resps, self.rule_i, 0)
            self.state.rules = rules
            self.rule_i = self.rule_part = self.var_i = None

        if rule_i is None:
            rule_i = self.rule_i
        if rule_i is not None:
            rule = rules[rule_i]
        if rule_part is None:
            rule_part = self.rule_part
            # note that rule_part can still be None! this will happen if
            # _reset was called before and it means that we do not have to
            # recalculate anything
        if var_i is None:
            var_i = self.var_i
        if var_i is None:
            # we have to calculate the if part on all variables if a new rule
            # was created
            in_variables = range(self.K_in)
        else:
            in_variables = (var_i,)

        if rule_part in ("ifs", "agg_lens", "agg_funcs", "shifts", "all"):
            # if any of those parts are changed, we have to recalculate the
            # memberships and all that follows
            # difference to version above is that we can ignore all rules but
            # the one that is changed
            smoothed = self.state.smoothed[rule_i]
            for var_i in in_variables:
                if self.do_smooth and rule.agg_lens[var_i] > 1:
                    smoothed[var_i] = smooth(self.data, rule, var_i)
                else:
                    smoothed[var_i] = self.data[var_i]
                if self.do_shift and rule.shifts[var_i] > 0:
                    shift = rule.shifts[var_i]
                    smoothed[var_i, shift:] = smoothed[var_i, :-shift]
                    smoothed[var_i, :shift] = inf_ + 1
            self.state.smoothed[rule_i] = smoothed
            # K x T shape
            _ifs_left = rule.ifs[..., 0]
            _ifs_center = rule.ifs[..., 1]
            _ifs_right = rule.ifs[..., 2]
            memberships = ne.evaluate(
                "where(smoothed < _ifs_center,"
                      "(smoothed - _ifs_left) / (_ifs_center - _ifs_left),"
                      "(_ifs_right - smoothed) / (_ifs_right - _ifs_center))")
            memberships[(memberships < 0) | np.isnan(memberships)] = 0.

            # T shape
            dof = np.product(memberships, axis=0)
            if self.do_thresh:
                dof = np.where(dof > rule.dof_thresh, dof, 0.)
            self.state.dofs[rule_i] = dof

        if rule_i is not None:
            # easy: dofs and membs stay the same, we just have to adjust the
            # responses
            self.state.resps[rule_i] = (self.state.dofs[rule_i] *
                                        rule.thens[:, None])

        dofs = self.state.dofs.sum(axis=0)
        if self.do_thresh:
            return np.where(dofs > 0, self.state.resps.sum(axis=0) / dofs, 0)
        return self.state.resps.sum(axis=0) / dofs


def learn(data_in, data_out=None, max_rules=10, max_agg_len=1, agg_funcs=None,
          max_shift=0, thresholding=False, default_out=None, k=.99, T=100,
          M=100, stop_rej=1., n_candidate_rules=10, debug=False):
    """Fits a fuzzy rule-based system mapping a multivariate input to a
    multivariate output time series using simulated annealing.

    Parameter
    ---------
    data_in : (K_in, T) ndarray
        Input data. K_in is number of input variables. T number of time steps.
    data_out : (K_out, T) ndarray or None, optional
        Output data. K_out is number of output variables, T number of time
        steps.
        If None, data_in will be used (think of autoregressive modelling).
    max_rules : int, optional
        Rule system will consist of <= max_rules.
    max_agg_len : int, optional
        Maximum number of time steps rules are allowed to aggregate over.
    agg_funcs : sequence of callables, optional
        Possible functions to use when aggregating. Must accept a window_len
        parameter and no_future=True.
        Take a look into the functions in vg.smoothing.
    max_shift : int, optional
        Maximum number of time steps rules are allowed to shift individual
        variables. For time-lagged responses.
    thresholding : int, optional
        Each rule attains a degree of fulfillment threshold that must be
        reached to activate it. Very experimental feature.
    default_out : float or None, optional
        Use this value as output if no rule applies or the result is negative.
    k : float, optional
        Cooling rate of the simualated annealing algorithm.
    T : float, optional of the simualated annealing algorithm.
        Starting temperature
    M : int, optional
        Number of time steps until temperature of the simualated annealing
        algorithm is decreased. See also stop_rej.
    stop_rej : [0.,..., 1.] float, optional
        Stop the algorithm if the rejection rate of newly generated candidates
        of the last M steps reaches this value.
    n_candidate_rules : int, optional
        Not very important. Number of rules to choose from when randomly
        generating a new one.
    debug : boolean, optional
        Print debug information
    """
    if os.path.exists("rules_dump.py"):
        os.remove("rules_dump.py")
    if os.path.exists("rules_dump.pyc"):
        os.remove("rules_dump.pyc")

    if data_out is None:
        data_out = data_in

    K_in = data_in.shape[0]
    K_out, N = data_out.shape

    # first rule is just a neutral filler
    # read: rule factory keyword arguments
    rule_parts = {"ifs": If(data_in, n_candidate_rules),
                  "thens": Then(data_out, default_out)}
    if max_agg_len > 1:
        rule_parts["agg_lens"] = AggLen(K_in, max_agg_len)
    if agg_funcs is not None:
        rule_parts["agg_funcs"] = AggFunc(K_in, agg_funcs)
    if max_shift > 0:
        rule_parts["shifts"] = Shift(K_in, max_shift)
    if thresholding:
        rule_parts["dof_thresh"] = DofThresh()
    RuleFactory = namedtuple("rule", " ".join(rule_parts.keys()))

    # dynamically generating a class to get a nicer formatting later on
    # maybe i should look for a more serious hobby
    class MyRuleFactory(RuleFactory):
        def __repr__(self):
            name = RuleFactory.__name__
            # adds a nice indentation
            filler = ",\n" + " " * (len(name) + 2)
            kwd_list = []
            for key, val in self._asdict().items():
                if key == "agg_funcs":
                    # functions do not have a repr that allows for a
                    # resurrection
                    val = ["%s.%s" % (func.__module__, func.func_name)
                           for func in val]
                    val = "[" + ", ".join(val) + "]"
                else:
                    val = val.__repr__()
                kwd_list += ["%s=%s" % (key, val)]
            kwd_str = filler.join(kwd_list)
            return '{name}({kwd_str})'.format(name=name, kwd_str=kwd_str)
    rules = \
        [MyRuleFactory(**{key: val.rule0 for key, val in rule_parts.items()})]

    incr = IncrementalResponse(data_in, rules)

    def gen_new_candidate(rules_):
        rules = copy_rules(rules_)

        # generate a new rule?
        new_prob = np.random.random()
        if len(rules) == 1 or new_prob ** 2 > len(rules) / float(max_rules):
            rules += [MyRuleFactory(**{key: val.new
                                       for key, val in rule_parts.items()})]
            if debug:
                print "new ",
            return rules
        # delete a rule?
        if len(rules) > 1 and new_prob ** .5 < len(rules) / float(max_rules):
            # the first rule of the fuzzy club is not to delete the first rule
            rule_i = incr.rule_i = np.random.randint(1, len(rules))
            del rules[rule_i]
            if debug:
                print "delete ",
            return rules

        # the second rule of the fuzzy club is not to change the first rule
        # we also set the rule to be changed and their parts on the
        # IncrementalResponse object, so that it knows where to save cpu cycles
        rule_i = incr.rule_i = np.random.randint(1, len(rules))
        part = incr.rule_part = random.choice(rules[rule_i]._fields)
        rules[rule_i] = rule_parts[part].change(rules[rule_i])
        incr.var_i = rule_parts[part].var_i
        if debug:
            print "change %s " % part + str(incr.var_i),
        return rules

    def info(best=False):
        response = rule_system_response(data_in,
                                        rules_best if best else rules,
                                        cache=False)
        fitness = obj_func(response)
        rme = lambda x: np.sqrt(x / (N * K_out))
        ns_str = ",".join("%.3f" % my.nash_sutcliff(res, dat)
                          for res, dat in zip(response, data_out))
        print ("\r%d obj: %.0f, nash: (%s), rmse: %.2f, "
               "T: %.2f, rules: %d, rej: %.2f" %
               (iteration, fitness, ns_str, rme(fitness), T, len(rules),
                np.mean(rejections))),
        sys.stdout.flush()

    obj_func = lambda x: np.nansum((data_out - x) ** 2)
    response = rule_system_response(data_in, rules)
    fitness = obj_func(response)

    fitness_best, rules_best = fitness, rules
    m = 0
    iteration = 0
    rejections = deque(M * [0], M)
    while np.mean(rejections) < stop_rej or iteration < 2 * M:
        iteration += 1
        rules_new = gen_new_candidate(rules)
        response = incr.response(rules_new)
        fitness_new = obj_func(response)

        if debug:
            resp = rule_system_response(data_in, rules_new, cache=True)
            assert np.all(np.isclose(resp, response))

        if fitness < fitness_best:
            fitness_best = fitness
            rules_best = copy_rules(rules)

        if (fitness_new < fitness or
            random.random() < np.exp((fitness - fitness_new) / T)):

            rules, fitness = rules_new, fitness_new
            m += 1
            rejections.appendleft(0)
            if debug:
                print "accepted"
        else:
            rejections.appendleft(1)
            incr.state = incr.old_state.copy()
            if debug:
                print "rejected"

        if m == M:
            T *= k
            m = 0
#             rules = clean_rules(rules, data_in)
            rules = greedy_then_opt(rules, data_in, data_out)
            if not thresholding:
                rules = unbias(rules, incr, data_out, debug)
            rules_dump(rules, "rules_dump.py")
            info()

    rules = greedy_then_opt(rules_best, data_in, data_out)

    # clear the cache
    del smooth.cache

    info(best=True)
    rules_dump(rules, "rules_best.py")
    return rules_best


def watch_and_learn(data_in, data_out, cv_relation=2. / 3, a_kwds=None,
                    *args, **kwds):
    """Starts animation and learn in separate processes."""
    cv_relation = float(cv_relation)
    a_kwds = {} if a_kwds is None else a_kwds
    p = multiprocessing.Process(target=animate,
                                args=(data_in, data_out, cv_relation),
                                kwargs=a_kwds)
    p.start()
    if cv_relation < 1:
        T = data_in.shape[1]
        cutoff = T * cv_relation
        data_in, data_out = data_in[:, :cutoff], data_out[:, :cutoff]
    rules = learn(data_in, data_out, *args, **kwds)
    p.join()
    return rules

if __name__ == "__main__":
#     import warnings
#     warnings.simplefilter("error", RuntimeWarning)

    np.random.seed(2)
    random.seed(1)

    import datetime
    import netCDF4 as nc
    from vg import times
    from vg.meteo import avrwind

    data_dir = "/home/dirk/Desktop/Diss/fuzzy/gradient/2013_03_21_uwnd_prec"
    precip_filepath = os.path.join(data_dir,
                                   "stations_precip_daily_2001-2010.csv")
    rr_data = np.loadtxt(precip_filepath, delimiter=';', skiprows=1)

    rr_years = rr_data[:, 0]
    rr_months = rr_data[:, 1]
    rr_daily = rr_data[:, 2:]
    rr_daily[rr_daily < -1] = np.nan

    indexarray = np.array([1, 3, 5, 7, 10])

    # anpassen, da stat_elev indicies anders sind als in rr_daily
    rr_daily5 = rr_data[:, indexarray + 2]

    rr_autisha = rr_daily5[:, 4]
    rr_1, rr_2, rr_3 = rr_daily5[:, 1:4].T
    rr_milloc = rr_daily5[:, 0]

    ds_u = nc.Dataset(os.path.join(data_dir, "uwnd.1981-2010.remap.nc"))
    ds_v = nc.Dataset(os.path.join(data_dir, "vwnd.1981-2010.remap.nc"))

    begin = nc.date2index(datetime.datetime(2001, 1, 1, 0, 0),
                          ds_u.variables['time'])
    end = nc.date2index(datetime.datetime(2009, 1, 1, 0, 0),
                        ds_u.variables['time'])

    uwind500 = ds_u.variables["uwnd"][begin:end, 3, 0, 0]
    uwind400 = ds_u.variables["uwnd"][begin:end, 2, 0, 0]
    uwind300 = ds_u.variables["uwnd"][begin:end, 1, 0, 0]
    # uwind250=ds_u.variables["uwnd"][:,3,0,0]
    uwind200 = ds_u.variables["uwnd"][begin:end, 0, 0, 0]
    uwind_time = ds_u.variables["time"][begin:end]
    vwind500 = ds_v.variables["vwnd"][begin:end, 1, 0, 0]
    vwind400 = ds_v.variables["vwnd"][begin:end, 2, 0, 0]
    vwind300 = ds_v.variables["vwnd"][begin:end, 3, 0, 0]
    # vwind250=ds_v.variables["vwnd"][:,3,0,0]
    vwind200 = ds_v.variables["vwnd"][begin:end, 5, 0, 0]
    vwind_time = ds_v.variables["time"][begin:end]

    dtimes = nc.num2date(vwind_time, ds_v.variables["time"].units)
    doys = times.datetime2doy(dtimes)
    var_names_in = ("u", "v",
                    # "doy",
#                     "direction",
#                     "speed"
                    )
    var_names_out = ("rr_milloc",
#                      "rr_autisha",
#                      "rr_1",
#                      "rr_2",
#                      "rr_3",
                     )
    dirs, speed = avrwind.component2angle(uwind200, vwind200)
    data_in = np.array([
                      uwind200, vwind200,
                      # doys,
#                       dirs,
#                       speed
                      ])

    data_out = np.array([
                         rr_milloc,
#                          rr_autisha,
#                          rr_1,
#                          rr_2,
#                          rr_3,
                         ])

    agg_funcs = (smoothing.smooth, smoothing.std,
                 smoothing.max, smoothing.min,
                 smoothing.maxdiff, smoothing.mindiff)

    rules = watch_and_learn(data_in,
                            data_out,
                            cv_relation=1,
                            # animation kwds
                            a_kwds=dict(dtimes=dtimes,
                                        out_names=var_names_out),
                            # fuzzy kwds
                            max_agg_len=14, max_shift=6,
                            # thresholding=True,
                            default_out=(0),  # , 0),
                            agg_funcs=agg_funcs,
                            # simulated annealing kwds
                            max_rules=5, T=200, M=50, k=.99,
                            # stop_rej=.01
                            )


#     import os
#     import pandas as pd
#     from scipy import stats
#     from vg import dyresm, times, smoothing
#     from vg import helpers as my
#
#     plt.rcParams["figure.figsize"] = (6, 5)
#
#     data_dir = "/home/dirk/Desktop/Diss/dendro/Daten_Jochen_2_Main"
#
#     pd_kwds = dict(index_col=0, parse_dates=0, skiprows=1)
#     temp = pd.read_table(os.path.join(data_dir, "Wuerzburg_Temperatur_neu.txt"), **pd_kwds)
#     nied = pd.read_table(os.path.join(data_dir, "Wuerzburg_Niederschlag_neu.txt"), **pd_kwds)
#     nied[nied < 0] = 0
#
#     dendro = pd.read_table(os.path.join(data_dir, "Main_Chrono_rezent_Arstan_cut.txt"),
#                            index_col=0, parse_dates=0)
#     dendro.Main_rezent = dendro.Main_rezent.astype(float)
#
#     # prepend temp/nied to column names so that we can join the stuff without collisions
#     # nied seems to have the smallest date range
#     items = [("temp %s" % key, temp.ix[nied.index, key].as_matrix())
#              for key in temp.columns]
#     items += [("nied %s" % key, nied.ix[nied.index, key].as_matrix())
#               for key in nied.columns]
#     data = pd.DataFrame.from_items(items)
#     data.index = nied.index
#
#     data = data.join(dendro, how="inner")
#
#     data["nied Mrz-Mai"] = data["nied Mrz"] + data["nied Apr"] + data["nied Mai"]
#     data["nied Apr-Mai"] = data["nied Apr"] + data["nied Mai"]
#     data["nied Mrz-Apr"] = data["nied Mrz"] + data["nied Apr"]
#
#     data_ranks = data  # .apply(my.rel_ranks)
#     # use only the aggregates
#     in_names = [key for key in data.columns if "-" in key]
#     data_in = np.asarray(data_ranks[in_names]).T
#     out_names = ["Main_rezent"]
#     data_out = np.asarray(data_ranks[out_names]).T
#
#     cv_relation = 2. / 3
#     cutoff = data_out.shape[1] * cv_relation
#     data_in_cal, data_in_val = data_in[:, :cutoff], data_in[:, cutoff:]
#     data_out_cal, data_out_val = data_out[:, :cutoff], data_out[:, cutoff:]
#
#     agg_funcs = (smoothing.smooth, smoothing.std,
#              smoothing.min, smoothing.max)
#
#     rules = watch_and_learn(data_in, data_out, data.index.to_datetime(),
#                             # fuzzy rule configuration
#                             max_rules=10, max_agg_len=2, max_shift=1,
#                             agg_funcs=agg_funcs,
#                             # simulated annealing parameters
#                             T=0.01, M=200, k=.995)

